
//*********************************************************************

// 2019.10.25 pjm: Format for subscripts is
 
//    "{" ( "f" or "c" or "w" or none) index ["," length] "}"

// where "f" means index by field, "c" means index by character, and
// "w" means index by word. Negative index indicates from right hand
// end of the string. Default is f = field.
// Only use the split-spec with strings, ignored for lists.
// hmmm... I think I like using a variable property to indicate that
// the string is to be handled as a record if it has a subscriptor.

// 2019.06.11 pjm: Appears to be a bug: The last line of a referenced
// subscript is not getting executed. Ok, found what looks to be the
// problem in CStrList::ListInsert(). It appeared to "intentionally"
// ignore the last line of the list it is inserting into the object list.
//
// 2019.06.1 pjm: maybe using common names, such as "date" and "time"
// for tags variables is not such a good idea. I should use something a
// bit more distinctive. I wrote a script that tried to use $?time$$ not
// realizing that that is a Tags variable. Going forward, I propose to
// identify Tags variables using an underline prefix, as in $?_time$$.
// I would exempt parameter references, since they are defined by the
// user, and environment variables.
//
// 2017.11.17 pjm: so I currently support $! and $? referencer types
// I have not been thinking. How about $| to reference a text file, and
// $/ to reference a website? Later: I have no idea what this would do.
//
// 2016.12.19 pjm: for some reason all syslog::log calls are missing.
// Have no explanation for this. Also, it would be maybe good to have
// several debug levels:
//    0. no debug                                          (noDebugLog)
//    1. proc calls                                      (procCallsLog)
//    2. commands w/parameters before resolve             (commandsLog)
//    3. verbose                                           (verboseLog)
//       if - choice path
//       forEach - forEach vars
//       etc.
//
// 2019.06.05 pjm: I found out why syslog::log calls are missing: I was
// trying to use the X-flag for help. It is supposed to be for the log
// function. If you want help just type tags without parameters. The
// X-flag turns logging on.
//
// 2014.09.05 pjm: implemented extended form of subscriptor. When the
// data type is string, a negative index counts right to left, while a
// positive index counts left to right. Also now support a second
// parameter {index,length}, which, when present and not zero,
// indicates to treat the subscript as a substring request. Neither
// a negative index parameter nor a length parameter are currently
// supported for string lists and node lists. They could, but it is
// very late, and I have run out of inspiration. A negative or zero
// length is treated as if it was absent.
                                                              
// 2014.09.04 pjm: added new variable called $ ?usep$ $. It contains x0ff,
// which is intended to be used as a field separator when you can't be
// sure what printable characters might be in a string.

// 2014.08.26 pjm: regexp ifn doesn't work. Ok, fixed it.

// 2014.08.26 pjm: store the results of a regular expression evaluation
// in a record instead of having a bunch of separate variables. So much
// better. Ok, I coded it in the Resolve::match method but I left it
// commented out until I have a chance to fix the references in the
// scripts.

// 2014.08.22 pjm: need to add a number type. only know if it is a string.
// done. Type name is "number". Note that number implies string. Done.

// 2013.12.18 pjm: forEach needs an $#else option that would apply if
// the forEach itself is empty. Had this situation where I wanted to
// close an output structure when a <menulist> had no <menuitems>. Not
// so easy to implement though.

// 2013.10.27 pjm: support the tab character. The simplest way to do
// this is to store a tab character in a standard variable, called, say,
// $ ?tab$ $. Probably need to do this for other control characters as
// well?

// 2013.04.9 pjm: should consolidate text and cdata sections so that
// the following would work:
// <test script="/test/text()"> commands <elements> commands </test>

// also, $#out var(file) should serialize the var even if it is a DOM.
// (in keeping with the XML operations below, which I haven't implemented
// anyway.)

// 2013.04.01 pjm: add a new command?
//
// $#outlink varname(filename)
//
// would be for outputting directly to a file. Probably has problems?
// Yeah, for one, needs to be closeable. Getting pretty ad-hoc, I think.
//
// Note that the maximum string length is sizeString (8192)
//
// TBD:
//   What XML operations do I need?
//
//       #graft var - append the node/nodelist in var
//       #clone var - deep-copy to var
//       #prune     - discard
//
//       add an attribute node given name=value
//       add an element node given name
//       add text node given text

// TBD: you should check that a set does not copy something to itself.

// TBD: figure out how to switch checking FArgs to checking the
// associated variable. This will allow the script to talk back to the
// interpreter.
//
// add a variable, called type, that contains the type of the last
// lvalue-type variable. (done)
//
// $#forEach xml() should accept HTML as well as XML. (done)
//
// need another odb command:
//
//    * #odbwrite topic(keys)
//
// changes:
//
// * ABC called iVarHandler to support user resolution of variable
//   names. (done)
//
// * Constructor to specify the iVarHandler.
//
// * Remove Defer command in favor of new notation for defering
//   resolution: $ !~xpath$ $ or $ ?~varname$ $
//   * add '~' as deferChar. (done)
//   * support in add and set commands
//
// * API for processing a list for output:
//
//   FB_USE_(bool) Output(RPCStrList _outList,
//                        Pchar      _fileName,
//                        bool       _append);
//
//
//*********************************************************************
// xtext.cpp
//*********************************************************************
//
//                              CXText
//                      A Text Generator Class
//                       based on XML and XPath
//                      (maybe simpler than XSLT,
//                      certainly not as powerful)
//                                 by
//                            Paul Medlock
//                       13036 Long Green Pike
//                          Hydes, MD 21082
//                         paul@medlock.com
//                           410.592.8468
//
//*********************************************************************

#include <cstring>
#include "console.h"
#include "date.h"
#include "filelist.h"
#include "http.h"
#include "misc.h"
#include "odb.h"
#include "odbc.h"
#include "process.h"
#include "regx.h"
#ifndef FV_SOUND
//#include "sound.h"
#endif
#include "time.h"
#include "xpath.h"
#include "xtext.h"

#define noDebugLog  0
#define procCallLog 1
#define commandLog  2
#define verboseLog  3

bool CXText::Resolve::didBlanks = false;
bool CXText::Resolve::didOut    = false;
bool CXText::Resolve::lastLoop  = false;
bool CXText::Resolve::nextLoop  = false;

const char  ftBlankFlag   = 'U';
const char  ftVocalFlag   = 'V';
const char  ftConsoleFlag = 'W';
const char  ftHelpFlag    = 'X';
const char  ftTraceFlag   = 'Y';
const char  ftDebugFlag   = 'Z';

const char  ftOpnChar     = '(';
const char  ftClsChar     = ')';

      char  ftEscChar     = '$';

      char  ftCmdChar     = '#';
      char *ftCmdStr      = Pchar("$#");

      char  ftVarChar     = '?';
      char  ftPopChar     = '^';
      char  ftTypeChar    = '?';
      char  ftSizeChar    = '%';
      char  ftFldsChar    = '#';
      char  ftTrimChar    = '-';
      char  ftCapChar     = '+';

      char *ftVarStr      = Pchar("$?");
      char *ftPopStr      = Pchar("$^");

      char  ftPathChar    = '!';
      char *ftPathStr     = Pchar("$!");

      char  ftCtxtChar    = '@';
      char *ftCtxtStr     = Pchar("  ");

      char  ftDeferChar   = '~';
      char *ftDeferStr    = Pchar("~");

      char  ftEndChar     = '$';
      char *ftEndStr      = Pchar("$$");

      char *ftLineCmd     = Pchar("$\\");

const char  ftUnkChar     = ' ';
      char *ftIncStr      = Pchar("  INCLUDE");

      char *traceName     = Pchar("xtext.trace");

const int longSizeString  = 16 * sizeString;

char errmssg[] = "Error detected while processing <%s>: %s";
char theline[sizeString];
char message[sizeString];

#define THROW(msg) \
   { \
     sprintf(message, errmssg, theline, msg); \
     log(true, message); \
     throw(message); \
   }
     
CXPath xp;

//---------------------------------------------------------------------
//
void log(bool _error, Pchar _msg, ...) {
   if (!_error && !CArgs::TheArgs->FArg(ftTraceFlag))
      return;
   
   char line[sizeString];
   va_list args;
   va_start(args, _msg);
   vsprintf(line, _msg, args);
   va_end(args);

   assert(strlen(line) < sizeString);
   char date[sizeLine];
   GetGMT(date, sizeLine);
   char work[4*sizeString];
   sprintf(work, Pchar("%s %s"), date, line);
   assert(strlen(work) < sizeString);

   ::OutputDebugString(work);
   } // log

//=====================================================================
// aCondition class implementation
//=====================================================================

const Pchar False     = Pchar("false");
const Pchar HardFalse = Pchar("hardFalse");
const Pchar True      = Pchar("true");
const Pchar Mark      = Pchar("mark");
const Pchar Done      = Pchar("done");

//---------------------------------------------------------------------
//
aCondition::aCondition(void) {
   stack = new CStrList();
   } // aCondition

//---------------------------------------------------------------------
//
aCondition::~aCondition() {
   } // ~aCondition

//---------------------------------------------------------------------
//
bool aCondition::IsTrue(void) {
   if (stack->GotoHead())
      return strcmp(stack->StrItem(false), True) == 0;

   return false;
   } // IsTrue

//---------------------------------------------------------------------
//
bool aCondition::IsFalse(void) {
   if (stack->GotoHead())
      return strcmp(stack->StrItem(false), False)     == 0 ||
             strcmp(stack->StrItem(false), HardFalse) == 0;

   return false;
   } // IsFalse

//---------------------------------------------------------------------
//
bool aCondition::IsHardFalse(void) {
   if (stack->GotoHead())
      return strcmp(stack->StrItem(false), HardFalse) == 0;

   return false;
   } // IsHardFalse

//---------------------------------------------------------------------
//
bool aCondition::IsMark(void) {
   if (stack->GotoHead())
      return strcmp(stack->StrItem(false), Mark)       == 0;

   return false;
   } // IsMark

//---------------------------------------------------------------------
//
bool aCondition::IsDone(void) {
   if (stack->GotoHead())
      return strcmp(stack->StrItem(false), Done)       == 0;

   return false;
   } // IsDone

//---------------------------------------------------------------------
//
void aCondition::Push(Pchar _state) {
   stack->StrInsert(_state, atTop);
   } // Push

//---------------------------------------------------------------------
//
void aCondition::Set(Pchar  _state) {
   stack->GotoHead();
   stack->StrReplace(_state);
   } // Set

//---------------------------------------------------------------------
//
void aCondition::Pop(void) {
   Assert(stack->Number());

   if (stack->GotoHead())
      stack->Remove(atTop);
   } // Pop

//---------------------------------------------------------------------
//
long aCondition::Size(void) {
   return stack->Number();
   } // Size

//---------------------------------------------------------------------
//
void aCondition::Flush(void) {
   stack->Flush();
   } // Flush

//=====================================================================
// CTXTType class implementation
//=====================================================================

//---------------------------------------------------------------------
//
CTXTType::CTXTType(void):
   CAtom() {
   } // CTXTType

//---------------------------------------------------------------------
//
CTXTType::~CTXTType() {
   }

//---------------------------------------------------------------------
//
RTTI_CPP(TXTType, Atom)

//---------------------------------------------------------------------
//
TXT_Type CTXTType::Type(Pchar _text) {
   if (StrEmpty(_text))
      return TXT_PlainText;

   // Ignore whitespace (and other control characters!?)
   Pchar p;
   for (p = _text; *p && *p <= ' '; p++);
   if (*p != '<')
      return TXT_PlainText;

   if (strlen(p) <= 5)
      return TXT_PlainText;

   char t = p[5];
   p[5] = 0;
   if (stricmp(p, Pchar("<html")) == 0) {
      p[5] = t;
      return TXT_HTMLText;
      }

   p[5] = t;
   return TXT_XMLText;
   } // Type

//---------------------------------------------------------------------
//
TXT_Type CTXTType::Type(PCBlock _blck) {
   Pchar p    = _blck->Lock();
   TXT_Type t = Type(p);
   _blck->Unlock();
   return t;
   } // Type

//---------------------------------------------------------------------
// Converts a few lines
//
TXT_Type CTXTType::Type(PCStrList _list) {
   int where = _list->Where();

   char work[sizeString];
   work[0] = 0;
   if (_list->GotoHead()) do {
      Strcat(work, _list->StrItem(false), sizeString);
      Trim(work, frontEnd);
      if (strlen(work) > 32)
         break;
      } while (_list->GotoNext());

   _list->GotoItem(where);

   return Type(work);
   } // Type

//---------------------------------------------------------------------
//
TXT_Type CTXTType::Type(PCFile _file) {
   if (!_file->OpenInput())
      return TXT_Error;

   char work[sizeString];
   work[0] = 0;
   while (!_file->atEOF()) {
      char temp[sizeLine];
      _file->ReadString(temp, sizeLine);
      Strcat(work, temp, sizeString);
      Trim(work, frontEnd);
      if (strlen(work) > 32)
         break;
      }
   _file->Close();

   return Type(work);
   } // Type

//---------------------------------------------------------------------
//
TXT_Type CTXTType::Type(Pchar _name, Pchar _path) {
   char ext[sizeString];
   GetExtension(_name, ext, sizeString);
   CapStr(ext);
   if (StrPos(Pchar(".HTM"), ext) > 0)
      return TXT_HTMLText;

   if (StrPos(Pchar(".XML"), ext) > 0)
      return TXT_XMLText;

   PCFile file = new CFile(_name, _path, false);
   return Type(file);
   } // Type

//=====================================================================
// CTXTFile class implementation
//=====================================================================

//---------------------------------------------------------------------
// Constructor
//
CTXTFile::CTXTFile(Pchar _name, Pchar _path):
   CAtom   (),
   sizeBuff(sizeString),
   more    (true),
   avail   (0),
   error(0),
   pbuf    (0) {

   file = new CFile(_name, _path);
   } // CTXTFile

//---------------------------------------------------------------------
// Constructor
//
CTXTFile::CTXTFile(PCFile _file):
   CAtom   (),
   sizeBuff(sizeString),
   more    (true),
   avail   (0),
   error(0),
   pbuf    (0) {

   file = _file;
   }

//---------------------------------------------------------------------
//
CTXTFile::~CTXTFile() {
   } // ~CTXTFile

//---------------------------------------------------------------------
//
RTTI_CPP(TXTFile, Atom)

//---------------------------------------------------------------
// get a line from the file
//
PCBlock CTXTFile::GetLine(void) {
   char line[sizeString];
   line[0] = 0;

   char ch = readChar();
   if (ch == 0)
      return 0;
   while (ch != 0) {
      if      (ch == '\r') {
         ch = checkChar();
         if (ch == 0 || ch == '\n')
            readChar();
         break;
         }
      else if (ch == '\n') {
         ch = checkChar();
         if (ch == 0 || ch == '\r')
            readChar();
         break;
         }
      // maybe I should not be doing this. Messes up records using 0xff
//?   else if (ch < ' ' && ch != '\t')
//?      ch = ' ';
      ChrCat(line, ch, sizeString);
      ch = readChar();
      }
   PCBlock blk = new CBlock(line);

   return blk;
   } // GetLine

//---------------------------------------------------------------------
//
void CTXTFile::resetChar(void) {
   more  = true;
   avail = 0;
   error = 0;
   pbuf  = 0;
   } // resetChar

//---------------------------------------------------------------------
//
bool CTXTFile::moreChar(void) {
   return (avail > 0 || !file->atEOF());
   } // moreChar

//---------------------------------------------------------------------
//
char CTXTFile::readChar(void) {
   if (avail == 0) {
      if (!more)
         return 0;

      memset(buffer, 0, sizeof(buffer));
      more  = file->ReadBlock(buffer+1, sizeof(buffer)-2);
      pbuf  = buffer+1;
      avail = strlen(pbuf);
      }

   char ch = *pbuf;
   pbuf++;
   avail--;

   return ch;
   } // readChar

//---------------------------------------------------------------------
//
char CTXTFile::checkChar(void) {
   if (avail == 0) {
      if (!more)
         return 0;

      memset(buffer, 0, sizeof(buffer));
      more  = file->ReadBlock(buffer+1, sizeof(buffer)-2);
      pbuf  = buffer+1;
      avail = strlen(pbuf);
      }

   return *pbuf;
   } // checkChar

//=====================================================================
// CXMLFile class implementation
//=====================================================================

//---------------------------------------------------------------------
// Constructor
//
CXMLFile::CXMLFile(Pchar _name, Pchar _path):
   CAtom(),
   more (true),
   avail(0),
   error(0),
   pbuf (0) {

   file = new CFile(_name, _path);
   } // CXMLFile

//---------------------------------------------------------------------
// Constructor
//
CXMLFile::CXMLFile(PCFile _file):
   CAtom(),
   more (true),
   avail(0),
   error(0),
   pbuf (0) {
   file = _file;
   } // CXMLFile

//---------------------------------------------------------------------
//
CXMLFile::~CXMLFile() {
   } // ~CXMLFile

//---------------------------------------------------------------------
//
RTTI_CPP(XMLFile, Atom)

//---------------------------------------------------------------------
//
bool CXMLFile::moreChar(void) {
   return (avail > 0 || !file->atEOF());
   } // moreChar

//---------------------------------------------------------------------
//
char CXMLFile::readChar(void) {
   if (avail == 0) {
      if (!more)
         return 0;

      memset(buffer, 0, sizeof(buffer));
      more  = file->ReadBlock(buffer+1, sizeof(buffer)-2);
      pbuf  = buffer+1;
      avail = strlen(pbuf);
      }

   char ch = *pbuf;
   if (ch < ' ' && ch != '\n' &&
                   ch != '\r' &&
                   ch != '\t')
      ch = ' ';
   if ((ch & 0x80) == 0x80)
      ch = ' ';

   pbuf++;
   avail--;

   return ch;
   } // readChar

//---------------------------------------------------------------------
//
char CXMLFile::checkChar(void) {
   if (avail == 0) {
      if (!more)
         return 0;

      memset(buffer, 0, sizeof(buffer));
      more  = file->ReadBlock(buffer+1, sizeof(buffer)-2);
      pbuf  = buffer+1;
      avail = strlen(pbuf);
      }

   return *pbuf;
   } // checkChar

//---------------------------------------------------------------------
//
void CXMLFile::pushChar(char _ch) {
if (pbuf <= buffer)
   Assert(pbuf > buffer);
   pbuf--;
   avail++;
   *pbuf = _ch;
   } // pushChar

//---------------------------------------------------------------------
// Get the remainder of the line from the file
//
void CXMLFile::getLine(Pchar _line, int _size) {
   char ch = readChar();
   while (ch != 0) {
      if      (ch == '\r') {
         ch = readChar();
         if (ch != 0 && ch != '\n')
            pushChar(ch);
         break;
         }
      else if (ch == '\n') {
         ch = readChar();
         if (ch != 0 && ch != '\r')
            pushChar(ch);
         break;
         }
      ChrCat(_line, ch, _size);
      ch = readChar();
      }
   } // getLine

//---------------------------------------------------------------------
// Get the name of the element
//
char CXMLFile::getElementName(Pchar _buf,  int _szBuf,
                              Pchar _name, int _szName) {
   char ch = readChar();
   while (ch != '>' && ch != '/' && ch != ' ') {
      if (ch == 0)
         THROW(Pchar("Unexpected end of input"));
      ChrCat(_name, ch, _szName);
      ch = readChar();
      }
   Strcat(_buf, _name, _szBuf);

   Trim(_name);
   return ch;
   } // getElementName

//---------------------------------------------------------------------
// Get the end tag. Has the form </..>
//
void CXMLFile::getEndTag(PCBlock _blk, Pchar _endName, int _size) {
   char buf[sizeString];
   Strcpy(buf, Pchar("</"), sizeString);

   // Get the element name
   char ename[sizeString];
   ename[0] = 0;
   char ch = getElementName(buf, sizeString, ename, sizeString);
   Strcpy(_endName, ename, _size);

   // Get the remainder of the tag
   while (ch != '>') {
      ch = readChar();
      if (ch == 0)
         THROW(Pchar("Unexpected end of input"));
      ChrCat(buf, ch, sizeString);
      }
   ChrCat(buf, '>', sizeString);

   _blk->AddBuffer(buf, strlen(buf));
   } // getEndTag

//---------------------------------------------------------------------
//
void CXMLFile::getContent(PCBlock _blk, Pchar _endName, int _size) {
   bool inComment1 = false;
   bool inComment2 = false;
   bool inComment3 = false;
   bool inCDATA1   = false;
   bool inCDATA2   = false;
   bool inCDATA3   = false;
   bool inInst1    = false;
   bool inInst2    = false;
   int nr          = 0;

   char buf[sizeString];
   buf[0]          = 0;

   for (;;) {
      char ch = readChar();
      if      (inComment1) {
         if (ch == '-') {
            inComment1 = false;
            inComment2 = true;
            }
         ChrCat(buf, ch, sizeString);
         if (++nr >= 128) {
            _blk->AddBuffer(buf, strlen(buf));
            buf[0] = 0;
            nr     = 0;
            }
         }
      else if (inComment2) {
         inComment2 = false;
         if (ch != '-')
            inComment1 = true;
         else
            inComment3 = true;
         ChrCat(buf, ch, sizeString);
         nr++;
         }
      else if (inComment3) {
         inComment3 = false;
         if (ch != '>')
            inComment1 = true;
         ChrCat(buf, ch, sizeString);
         nr++;
         }
      else if (inCDATA1) {
         if (ch == ']') {
            inCDATA1 = false;
            inCDATA2 = true;
            }
         ChrCat(buf, ch, sizeString);
         if (++nr >= 128) {
            _blk->AddBuffer(buf, strlen(buf));
            buf[0] = 0;
            nr     = 0;
            }
         }
      else if (inCDATA2) {
         inCDATA2 = false;
         if (ch != ']')
            inCDATA1 = true;
         else
            inCDATA3 = true;
         ChrCat(buf, ch, sizeString);
         nr++;
         }
      else if (inCDATA3) {
         inCDATA3 = false;
         if (ch != '>')
            inCDATA1 = true;
         ChrCat(buf, ch, sizeString);
         if (!StrEmpty(buf))
            _blk->AddBuffer(buf, strlen(buf));
         buf[0] = 0;
         nr     = 0;
         }
      else if (inInst1) {
         if (ch == '?') {
            inInst1 = false;
            inInst2 = true;
            }
         ChrCat(buf, ch, sizeString);
         if (++nr >= 128) {
            _blk->AddBuffer(buf, strlen(buf));
            buf[0] = 0;
            nr     = 0;
            }
         }
      else if (inInst2) {
         inInst2 = false;
         if (ch != '>')
            inInst1 = true;
         ChrCat(buf, ch, sizeString);
         nr++;
         }
      else if (ch == '<') {
         _blk->AddBuffer(buf, strlen(buf));
         buf[0] = 0;
         nr     = 0;

         if      (checkChar() == '!') {
            readChar();
            if      (checkChar() == '-') {                   // comment
               readChar();
               if (readChar() != '-')
                  THROW(Pchar("Expected a comment"))
               else {
                  inComment1 = true;
                  Strcat(buf, Pchar("<!--"), sizeString);
                  nr += 4;
                  }
               }
            else if (checkChar() == '[') {                     // CDATA
               readChar();
               if (readChar() != 'C' ||
                   readChar() != 'D' ||
                   readChar() != 'A' ||
                   readChar() != 'T' ||
                   readChar() != 'A' ||
                   readChar() != '[')
                  THROW(Pchar("Expected a CDATA section"))
               else {
                  inCDATA1 = true;
                  Strcat(buf, Pchar("<![CDATA["), sizeString);
                  nr += 9;
                  }
               }
            }
         else if (checkChar() == '?') {
            readChar();
            inInst1 = true;
            Strcat(buf, Pchar("<?"), sizeString);
            nr += 2;
            }
         else if (checkChar() == '/') {
            readChar();
            getEndTag(_blk, _endName, _size);
            return;
            }
         else
            getElement(_blk);
         }
      else if (ch == 0)
         return;
      else {
         ChrCat(buf, ch, sizeString);
         if (++nr >= 128) {
            _blk->AddBuffer(buf, strlen(buf));
            buf[0] = 0;
            nr     = 0;
            }
         }
      }
   } // getContent

//---------------------------------------------------------------------
// Get the begin tag. Note that it could have the form <../>
//
void CXMLFile::getElement(PCBlock _blk) {
   char buf[sizeString];
   ChrCpy(buf, '<');

   // Get the element name
   char begName[sizeString];
   begName[0] = 0;
   char ch = getElementName(buf, sizeString, begName, sizeString);
   ChrCat(buf, ch, sizeString);

   // Get the remainder of the tag
   bool recurse = true;
   if (ch == '/' && checkChar() == '>') {
      recurse = false;
      ch = readChar();
      ChrCat(buf, ch, sizeString);
      }
   else if (ch == ' ') {
      ch = readChar();
      while (ch != '>') {
         if (ch == 0)
            THROW(Pchar("Unexpected end of input"))
         ChrCat(buf, ch, sizeString);
         if (ch == '/' && checkChar() == '>')
            recurse = false;
         ch = readChar();
         }
      ChrCat(buf, '>', sizeString);
      }

   _blk->AddBuffer(buf, strlen(buf));

   if (recurse) {
      char endName[sizeString];
      endName[0] = 0;
      getContent(_blk, endName, sizeString);
      if (strcmp(begName, endName) != 0) {
         THROW(Pchar("Invalid structure in XML input"))
         }
      }
   } // getElement

//---------------------------------------------------------------------
//
bool CXMLFile::getXML(PCBlock _blk) {
   int nr = 0;
   char buf[sizeString];
   buf[0] = 0;

   // bypass whitespace
   while (checkChar() <= ' ') {
      if (checkChar() == 0)
         return false;
      readChar();
      }

   // Verify the first non-whitespace character
   if (checkChar() != '<')
      THROW(Pchar("Unexpected character in XML input"))

   // Bypass the optional prolog(s) <?xml...?>
   do {
      char ch;
      while (checkChar() != '<') {
          ch = readChar();
          ChrCat(buf, ch, sizeString);
          nr++;
          if (++nr >= 128) {
             _blk->AddBuffer(buf, strlen(buf));
             buf[0] = 0;
             nr     = 0;
             }
          }
      ch = readChar();
      if (checkChar() == '?') {
         ChrCat(buf, ch, sizeString);
         nr++;
         if (++nr >= 128) {
            _blk->AddBuffer(buf, strlen(buf));
            buf[0] = 0;
            nr     = 0;
            }
         }
      } while (checkChar() == '?');

   _blk->AddBuffer(buf, strlen(buf));

   if (checkChar() == '!') {
      pushChar('<');
      char endName[sizeString];
      endName[0] = 0;
      getContent(_blk, endName, sizeString);
      }
   else
      getElement(_blk);

   return true;
   } // getXML

//---------------------------------------------------------------------
// Get a chunk from the specified input and return it in the CBlock
// object if it is a valid XML element. Otherwise, return zero.
//
PCBlock CXMLFile::GetXML(void) {
   PCBlock block = new CBlock();
   try {
      if (getXML(block))
         return block;
      else
         return 0;
      }

   catch (Pchar _error) {
      log(true, Pchar("XText Error: %s"), _error);
      return 0;
      }
   } // GetXML

//=====================================================================
// CXText class implementation
//=====================================================================

/*?
//---------------------------------------------------------------------
//
CXText::CXText(void):
   CAtom () {

   // Prepare the Tags mark constants
   ftEscChar      = '$';
   ftCmdChar      = '#';
   ftCmdStr[0]    = '$'; ftCmdStr[1]  = '#';
   ftVarChar      = '?';
   ftPathChar     = '!';
   ftPathStr[0]   = '$'; ftPathStr[1] = '!';
   ftDeferChar    = '~';
   ftDeferStr[0]  = '~';
   ftEndChar      = '$';
   ftEndStr[0]    = '$'; ftEndStr[1]  = '$';
   ftIncStr[0]    = '$'; ftIncStr[1]  = '#';

   tagsPath[0]    = 0;
   ::GetEnvironmentVariable(Pchar("tagsPath"), tagsPath, sizeString);
   scriptPath[0]  = 0;

   // Create a console window on request for tracking activity
   if (CArgs::TheArgs->FArg(ftConsoleFlag))
      new CConsole();

   // Initialize the default output list
   output   = new CStrList();

   // Initialize the forEach input() stack
//?inpStack = new CStrList();
   } // CXText
?*/

//---------------------------------------------------------------------
//
CXText::CXText(IVarHandler* _varHandler):
   CAtom     (),
   resolver  (0),
   varHandler(_varHandler) {

   // Prepare the Tags mark constants
   ftEscChar      = '$';
   ftCmdChar      = '#';
   ftCmdStr[0]    = '$'; ftCmdStr[1]  = '#';
   ftVarChar      = '?';
   ftPathChar     = '!';
   ftPathStr[0]   = '$'; ftPathStr[1] = '!';
   ftDeferChar    = '~';
   ftDeferStr[0]  = '~';
   ftEndChar      = '$';
   ftEndStr[0]    = '$'; ftEndStr[1]  = '$';
   ftIncStr[0]    = '$'; ftIncStr[1]  = '#';

   Assert(varHandler);

   tagsPath[0]    = 0;
   scriptPath[0]  = 0;

   // Initialize the default output list
   output         = new CStrList();

   // Initialize the forEach input() stack
//?inpStack = new CStrList();
   } // CXText

//---------------------------------------------------------------------
//
CXText::~CXText() {
   delete resolver;
   } // ~CXText

//---------------------------------------------------------------------
//
RTTI_CPP(XText, Atom)

//---------------------------------------------------------------------
// Preprocess the includes, and then load the script document into
// the DOM.
//
PCDocument CXText::LoadScript(Pchar _scriptName, bool _preserveWhiteSpace) {
   log(true, Pchar("LoadScript: loading %s"), _scriptName);
   
   initialize();

   char scriptName[sizeString];
   if (!FindFileInPath(_scriptName, scriptPath, scriptName, sizeString)) {
      log(true, Pchar("Could not find the file in %s"), scriptPath);
      return 0;
      }

   PCStrList script = new CStrList();
   if (!script->Load(scriptName)) {
      log(true, Pchar("Could not load the script file %s"), scriptName);
      return 0;
      }

   Resolve::didOut    = false;
   Resolve::lastLoop  = false;
   Resolve::nextLoop  = false;
   Resolve* loader    = new Resolve();

   PCDocument doc;
   try {
      Resolve::Initialize(tagsPath, scriptPath, varHandler);
      doc = loader->LoadScript(script, _preserveWhiteSpace);
      }
   catch (Pchar _error) {
#ifdef _DEBUG
      ::MessageBox(NULL, _error,
                         Pchar("XText Load failed"),
                         MB_OK | MB_ICONERROR);
#endif
      delete loader;

      log(true, Pchar("Script file %s not well formed"), scriptName);
      return 0;
      }

   delete loader;

   return doc;
   } // LoadScript

//---------------------------------------------------------------------
//
void CXText::Initialize(PCNode _scriptNode) {
   scriptNode = _scriptNode;
   script     = scriptNode->OwnerDocument();
   } // Initialize

//---------------------------------------------------------------------
//
void CXText::Initialize(PCNode _scriptNode, PCNode _dataNode) {
   scriptNode = _scriptNode;
   script     = scriptNode->OwnerDocument();

   dataNode   = _dataNode;
   data       = dataNode->OwnerDocument();
   } // Initialize

//---------------------------------------------------------------------
//
bool CXText::Evaluate(Pchar _xpath) {
   log(true, Pchar("Evaluate: evaluating %s"), _xpath);

   initialize();
   getScript(_xpath);

   Resolve::didOut   = false;
   Resolve::lastLoop = false;
   Resolve::nextLoop = false;
   resolver          = new Resolve();

   try {
      Resolve::Initialize(tagsPath, scriptPath, varHandler);
      resolver->resolve(body, output);
      }
   catch (Pchar _error) {
      ::MessageBeep(MB_ICONASTERISK);
#ifdef _DEBUG
      ::MessageBox(NULL, _error,
                         Pchar("XText Evaluation Error"),
                         MB_OK | MB_ICONERROR);
#endif
//?   output->Save(Pchar("output.lst"));
      log(true, Pchar("XText: %s"), _error);

      return false;
      }
   catch (...) {
      ::MessageBeep(MB_ICONASTERISK);
#ifdef _DEBUG
      ::MessageBox(NULL, Pchar("XText excepted"),
                         Pchar("XText Evaluation Error"),
                         MB_OK | MB_ICONERROR);
#endif
//?   output->Save(Pchar("output.lst"));
      log(true, Pchar("XText: Error detected"));

      return false;
      }

//?output->Save(Pchar("output.lst"));

   return true;
   } // Evaluate

//---------------------------------------------------------------------
//
bool CXText::Output(Pchar _fileName, bool _append) {
   if (resolver == 0)
      return false;
      
   return resolver->dumpOutput(output, _fileName, _append);
   } // Output

//---------------------------------------------------------------------
//
void CXText::initialize(void) {
   log(true, Pchar("Initializing"));
   
   char path[sizeString];
   path[0] = 0;
   PCAtom atom;
   PCVar var;

   // Initialize the empty variable
   varHandler->Variable("empty", Pchar(""));

   // Initialize the tab variable
   varHandler->Variable("tabchar", Pchar("\t"));

   // Initialize the universal record separator variable
   path[0] = 0xff;
   path[1] = 0;
   varHandler->Variable(Pchar("usep"), path);

   // Initialize the tagsPath and scriptPath variables
   var = varHandler->Variable(Pchar("tagsPath"));
   if (var) {
      Pchar temp = toString(var);
      Strcpy(tagsPath, temp, sizeString);
      strcpy(scriptPath, tagsPath);
      delete[] temp;
      }

   // Initialize the output variable to point to the output list
   atom = TYPECAST(output, Atom);
   varHandler->Variable(Pchar("output"), atom);

   // Initialize the data variable to point to the data node
   if (dataNode) {
      atom = TYPECAST(dataNode, Atom);
      varHandler->Variable(Pchar("data"), atom);
      }

   // Initialize the script variable to point to the script node
   if (scriptNode) {
      atom = TYPECAST(scriptNode, Atom);
      varHandler->Variable(Pchar("script"), atom);
      }

   // Initialize the contextNode variable to point to the script node
   varHandler->Variable(Pchar("contextNode"), atom);

   // Initialize the currentPath variable to contain the current
   // directory
   GetCurrentDir(path, sizeString);
   varHandler->Variable(Pchar("currentPath"), path);

   // Initialize the columns variables to empty
   varHandler->Variable("columns",     Pchar(""));
   varHandler->Variable("getColumns",  Pchar(""));
   varHandler->Variable("lineColumns", Pchar(""));
   varHandler->Variable("SQLColumns",  Pchar(""));
   varHandler->Variable("regXColumns", Pchar(""));

   // Initialize the position variable to zero
   varHandler->Variable(Pchar("position"), 0L);

   // Initialize the last variable to zero
   varHandler->Variable(Pchar("last"), 0L);

   // Initialize the grep variable to empty
   varHandler->Variable("grep", Pchar(""));

   // Initialize the forEach variable to empty
   varHandler->Variable("forEach", Pchar(""));

   // Initialize the forEach file variables to empty
   varHandler->Variable("fileInfo", Pchar(""));
   varHandler->Variable("fileDrive",Pchar(""));
   varHandler->Variable("filePath", Pchar(""));
   varHandler->Variable("fileName", Pchar(""));

   // Find the output body text
   xp.SetVars(varHandler);
   } // initialize

//---------------------------------------------------------------------
//
void CXText::getScript(Pchar _path) {
   PCNode node;
   char path[sizeString];
   path[0] = 0;
   ulong error;

   // Use the optional setmarks attribute to allow the script
   // writer to control the actual marking codes as in
   //    <myscript script="/myscript/text()" marks=Pchar("$#?!$")>
   // format of marks value is
   //    Pchar("escape,command,variable,xpath,end")          (omit the commas)
   Strcpy(path, Pchar("/*/@marks"), sizeString);
   node = xp.SelectSingleNode(scriptNode, path);
   if (node) {
      Pchar p = node->NodeValue();
      if (p && strlen(p) == 6) {
         ftEscChar    = p[0];
         ftCmdChar    = p[1];
         ftCmdStr[0]  = p[0]; ftCmdStr[1]  = p[1];
         ftVarChar    = p[2];
         ftPathChar   = p[3];
         ftPathStr[0] = p[0]; ftPathStr[1] = p[3];
         ftCtxtChar   = p[4];
         ftCtxtStr[0] = p[0]; ftCtxtStr[1] = p[4];
         ftEndChar    = p[5];
         ftEndStr[0]  = p[0]; ftEndStr[1]  = p[5];
         delete[] p;
         }
      }

   // Obtain the script path as specified in the script if needed
   if (StrEmpty(_path)) {
      Pchar p;
      Strcpy(path, Pchar("@script"), sizeString);
      PCNode node = xp.SelectSingleNode(scriptNode, path);
      if (node)
         p = node->NodeValue();
      else {
         p = new char[sizeLine];
         strcpy(p, Pchar("/*/text()"));
         }
      if (!p)
         THROW(Pchar("Error: script not specified"))
      Strcpy(path, p, sizeString);
      delete[] p;
      }
   else
      Strcpy(path, _path, sizeString);

   node = xp.SelectSingleNode(scriptNode, path);
   error = xp.Error();
   if (error != FVE_OK) {
      log(true, Pchar("Could not resolve XPath expression <%s>"), _path);
      return;
      }

   // Pull the body text from the script document and place it into a
   // strlist.
   Pchar p = node->NodeValue();
   PCAtom blk = new CBlock(p);
   delete[] p;
   body = new CStrList();
   body->BlockTo(blk);
   } // getScript

//---------------------------------------------------------------------
// Set the specified variable to the specified numeric value.
//
void CXText::setVar(PCVar _var, int _value) {
   char value[sizeString];
   sprintf(value, Pchar("%d"), _value);
   _var->Value(value);
   } // setVar

//---------------------------------------------------------------------
// Set the specified variable to the specified string value.
//
void CXText::setVar(PCVar _var, Pchar _value) {
   _var->Value(_value);
   } // setVar

//---------------------------------------------------------------------
// Set the specified variable to point to the specified atomic value.
//
void CXText::setVar(PCVar _var, PCAtom _value) {
   _var->Value(_value);
   } // setVar

//---------------------------------------------------------------------
// Note: caller must free memory when done using the string that this
// routine returns.
//
Pchar CXText::toString(PCVar _var) {
   Pchar rtn = 0;

   if (!_var || !_var->Value()) {
      rtn = new char[1];
      rtn[0] = 0;
      return rtn;
      }

   PCAtom atom = _var->Value();

   // Convert the block to a char array
   if      (_var->Itsa(iBlock)) {
      PCBlock blk = TYPECAST(atom, Block);
      if (!blk->Empty()) {
         Pchar p = blk->Lock();
         if (p) {
            rtn = new char[strlen(p)+1];
            strcpy(rtn, p);
            blk->Unlock();
            }
         }
      }

   // Convert a strlist to a char array
   else if (_var->Itsa(iStrList)) {
      PCStrList list = TYPECAST(atom, StrList);
      long len = list->ListSize();
      long nbr = list->Number();
      len      = len + nbr * 2 + 1;
      rtn      = new char[len];
      rtn[0]   = 0;
      if (list->GotoHead()) do {
         Strcat(rtn, list->StrItem(), len);
         if (!list->AtTail()) {
            Strcat(rtn, eolDelim, len);
            }
         } while (list->GotoNext());
      }

   // Convert a nodelist to a char array. (This one is harder because
   // we have to prefigure the length, which means we have to go
   // through everything an extra time)
   else if (_var->Itsa(iNodeList)) {
      PCNodeList nodes = TYPECAST(atom, NodeList);
      uint i;
      int len = 0;
      for (i = 0; i < nodes->Length(); i++) {
         PCNode node = nodes->Item(i);
         if      (node->Isa(iText) ||
                  node->Isa(iAttr)) {
            Pchar p = node->NodeValue();
            if (p)
               len += strlen(p);
            delete[] p;
            }
         else if (node->Isa(iElement)) {
            Pchar p = node->NodeName();
            if (p)
               len += strlen(p);
            delete[] p;
            }
         }

      rtn    = new char[len+1];
      rtn[0] = 0;

      for (i = 0; i < nodes->Length(); i++) {
         PCNode node = nodes->Item(i);
         if      (node->Isa(iText) ||
                  node->Isa(iAttr)) {
            Pchar p = node->NodeValue();
            if (p) {
               Strcat(rtn, p, len+1);
               }
            delete[] p;
            }
         else if (node->Isa(iElement)) {
            Pchar p = node->NodeName();
            if (p) {
               Strcat(rtn, p, len+1);
               }
            delete[] p;
            }
         }
      }

   // Convert a node (text, attribute, or element) to a char array
   else if (_var->Itsa(iNode)) {
      PCNode node = TYPECAST(atom, Node);
      if (node->Isa(iText) ||
          node->Isa(iAttr))
         rtn = node->NodeValue();

      // Save the node and its node tree to a char array
      else if (node->Isa(iDocument)) {
         PCDocument doc = TYPECAST(node, Document);
         PCBlock blk = new CBlock();
         doc->Save(blk);
         if (!blk->Empty()) {
            Pchar p = blk->Lock();
            if (p) {
               rtn = new char[strlen(p)+1];
               strcpy(rtn, p);
               blk->Unlock();
               }
            }
         }
      else if (node->Isa(iElement)) {
         PCBlock blk = new CBlock();
         node->Save(blk);
         if (!blk->Empty()) {
            Pchar p = blk->Lock();
            if (p) {
               rtn = new char[strlen(p)+1];
               strcpy(rtn, p);
               blk->Unlock();
               }
            }
         }
      }

   if (!rtn) {
      rtn = new char[1];
      rtn[0] = 0;
      }
   return rtn;
   } // toString

//=====================================================================
// Resolve class implementation
//=====================================================================

char CXText::Resolve::tagsPath[sizeString];
char CXText::Resolve::scriptPath[sizeString];

IVarHandler* CXText::Resolve::varHandler;
aCondition   CXText::Resolve::condition;

//---------------------------------------------------------------------
//
CXText::Resolve::Resolve(void):
   nrRegVars (0) {

   inpStack = new CStrList();
   } // Resolve

//---------------------------------------------------------------------
//
CXText::Resolve::~Resolve() {
   } // ~Resolve

//---------------------------------------------------------------------
// several debug levels:
//    0. no debug
//    1. proc calls
//    2. commands w/parameters before resolve
//    3. verbose
//       if - choice path
//       forEach - forEach vars
//
void CXText::Resolve::dbgLog(int _level, Pchar _message) {
   // Check debug level
   PCVar var = varHandler->Variable(Pchar("debugging"));
   int level = 0;
   if (var) {
      Pchar temp = toString(var);
      // Check for no debug log
      if (StrEmpty(temp)               ||
          strcmp  (temp, Pchar("0"))     == 0 ||
          stricmp (temp, Pchar("no"))    == 0 ||
          stricmp (temp, Pchar("off"))   == 0 ||
          stricmp (temp, Pchar("false")) == 0)
         return;

      // Check for default debug log level
      if (stricmp (temp, Pchar("yes"))   == 0 ||
          stricmp (temp, Pchar("or"))    == 0 ||
          stricmp (temp, Pchar("true"))  == 0)
         level = 2;                               // default test level
      // Get the debug log level
      else
         level = StrToShort(temp);

      // Check to actually do the debug log. Note that -Y flag must be
      // set to enable debug log.
      if (_level >= level)
         log(false, _message);
      delete[] temp;
      }
   } // dbgLog

void CXText::Resolve::dbgLog(int _level, Pchar _message, Pchar _text) {
   int len = strlen(_message) + strlen(_text) + 4;
   Pchar msg = new char[len];
   strcpy(msg, _message);
   strcat(msg, Pchar("/"));
   strcat(msg, _text);
   dbgLog(_level, msg);
   delete[] msg;
   } // dbgLog

void CXText::Resolve::dbgLog(int _level, Pchar _message, Pchar _text, Pchar _exp) {
   int len = strlen(_message) + strlen(_text) + strlen(_exp) + 4;
   Pchar msg = new char[len];
   strcpy(msg, _message);
   strcat(msg, Pchar("/"));
   strcat(msg, _text);
   strcat(msg, Pchar(": "));
   strcat(msg, _exp);
   dbgLog(_level, msg);
   delete[] msg;
   } // dbgLog

//---------------------------------------------------------------------
//
void CXText::Resolve::Initialize(Pchar        _tagsPath,
                                 Pchar        _scriptPath,
                                 IVarHandler* _varHandler) {
   Strcpy(tagsPath,   _tagsPath,   sizeString);
   Strcpy(scriptPath, _scriptPath, sizeString);

   varHandler = _varHandler;

   condition.Flush();
   condition.Push(Done);
   condition.Push(True);
   } // Initialize

//---------------------------------------------------------------------
// Set the specified variable to the specified numeric value.
//
void CXText::Resolve::setVar(PCVar _var, int _value) {
   char value[sizeString];
   sprintf(value, Pchar("%d"), _value);
   _var->Value(value);
   } // setVar

//---------------------------------------------------------------------
// Set the specified variable to the specified string value.
//
void CXText::Resolve::setVar(PCVar _var, Pchar _value) {
   _var->Value(_value);
   } // setVar

//---------------------------------------------------------------------
// Set the specified variable to point to the specified atomic value.
//
void CXText::Resolve::setVar(PCVar _var, PCAtom _value) {
   _var->Value(_value);
   } // setVar

//---------------------------------------------------------------------
// Note: caller must free memory when done using the string that this
// routine returns.
//
Pchar CXText::Resolve::toString(PCVar _var) {
   Pchar rtn = 0;

   if (!_var || !_var->Value()) {
      rtn = new char[1];
      rtn[0] = 0;
      return rtn;
      }

   PCAtom atom = _var->Value();

   // Convert the block to a char array
   if      (_var->Itsa(iBlock)) {
      PCBlock blk = TYPECAST(atom, Block);
      if (!blk->Empty()) {
         Pchar p = blk->Lock();
         if (p) {
            rtn = new char[strlen(p)+1];
            strcpy(rtn, p);
            blk->Unlock();
            }
         }
      }

   // Convert a strlist to a char array
   else if (_var->Itsa(iStrList)) {
      PCStrList list = TYPECAST(atom, StrList);
      long len = list->ListSize();
      long nbr = list->Number();
      len      = len + nbr * 2 + 1;
      rtn      = new char[len];
      rtn[0]   = 0;
      if (list->GotoHead()) do {
         Strcat(rtn, list->StrItem(), len);
         if (!list->AtTail()) {
            Strcat(rtn, eolDelim, len);
            }
         } while (list->GotoNext());
      }

   // Convert a nodelist to a char array. (This one is harder because
   // we have to prefigure the length, which means we have to go
   // through everything an extra time)
   else if (_var->Itsa(iNodeList)) {
      PCNodeList nodes = TYPECAST(atom, NodeList);
      uint i;
      int len = 0;
      for (i = 0; i < nodes->Length(); i++) {
         PCNode node = nodes->Item(i);
         if      (node->Isa(iText) ||
                  node->Isa(iAttr)) {
            Pchar p = node->NodeValue();
            if (p)
               len += strlen(p);
            delete[] p;
            }
         else if (node->Isa(iElement)) {
            Pchar p = node->NodeName();
            if (p)
               len += strlen(p);
            delete[] p;
            }
         }

      rtn    = new char[len+1];
      rtn[0] = 0;

      for (i = 0; i < nodes->Length(); i++) {
         PCNode node = nodes->Item(i);
         if      (node->Isa(iText) ||
                  node->Isa(iAttr)) {
            Pchar p = node->NodeValue();
            if (p) {
               Strcat(rtn, p, len+1);
               }
            delete[] p;
            }
         else if (node->Isa(iElement)) {
            Pchar p = node->NodeName();
            if (p) {
               Strcat(rtn, p, len+1);
               }
            delete[] p;
            }
         }
      }

   // Convert a node (text, attribute, or element) to a char array
   else if (_var->Itsa(iNode)) {
      PCNode node = TYPECAST(atom, Node);
      if (node->Isa(iText) ||
          node->Isa(iAttr))
         rtn = node->NodeValue();

      // Save the node and its node tree to a char array
      else if (node->Isa(iDocument)) {
         PCDocument doc = TYPECAST(node, Document);
         PCBlock blk = new CBlock();
         doc->Save(blk);
         if (!blk->Empty()) {
            Pchar p = blk->Lock();
            if (p) {
               rtn = new char[strlen(p)+1];
               strcpy(rtn, p);
               blk->Unlock();
               }
            }
         }
      else if (node->Isa(iElement)) {
         PCBlock blk = new CBlock();
         node->Save(blk);
         if (!blk->Empty()) {
            Pchar p = blk->Lock();
            if (p) {
               rtn = new char[strlen(p)+1];
               strcpy(rtn, p);
               blk->Unlock();
               }
            }
         }
      }

   if (!rtn) {
      rtn = new char[1];
      rtn[0] = 0;
      }
   return rtn;
   } // toString

//---------------------------------------------------------------------
//
void CXText::Resolve::toString(PCVar _var, Pchar _s, int _size) {
   memset(_s, 0, _size);

   Pchar p = toString(_var);
   if (p) {
      Strcpy(_s, p, _size);
      }
   delete[] p;
   } // toString

//---------------------------------------------------------------------
// Find the specified variable and convert its contents to a string. If
// there is no variable with the specified name, obtain the contents of
// the specified environment variable. If there is no environment
// variable with the specified name return the empty string. User must
// release memory returned.
//
Pchar CXText::Resolve::toString(Pchar _name) {
   PCVar var = varHandler->Variable(_name);
   if (!var) {
      Pchar env = new char[sizeString];
      env[0]    = 0;
      ::GetEnvironmentVariable(_name, env, sizeString);
      return env;
      }
   else
      return toString(var);
   } // toString

//---------------------------------------------------------------------
// Find the specified variable and convert its contents to a string. If
// there is no variable with the specified name, obtain the contents of
// the specified environment variable. If there is no environment
// variable with the specified name return the empty string.
//
void CXText::Resolve::toString(Pchar _name, Pchar _s, int _size) {
   PCVar var = varHandler->Variable(_name);
   if (!var) {
      memset(_s, 0, _size);
      ::GetEnvironmentVariable(_name, _s, _size);
      }
   else
      toString(var, _s, _size);
   } // toString

//---------------------------------------------------------------------
//
void CXText::Resolve::setDateTime(void) {
   Pchar p;
   CDate mydate;
   CTime mytime;

   varHandler->Variable(Pchar("year"), mydate.Year());
   varHandler->Variable(Pchar("month"), mydate.Month());
   p = mydate.MonthName();
   varHandler->Variable(Pchar("monthName"), p);
   delete[] p;
   varHandler->Variable(Pchar("day"), mydate.Day());
   p = mydate.DayName();
   varHandler->Variable(Pchar("dayOfWeek"), p);
   delete[] p;
   varHandler->Variable(Pchar("dayOfMonth"), mydate.DayOfMonth());
   varHandler->Variable(Pchar("dayOfYear"), mydate.DayOfYear());
   p = mytime.AsString();
   varHandler->Variable(Pchar("time"), p);
   delete[] p;
   } // setDateTime

//---------------------------------------------------------------------
// Convert a set of named values into named variables. Two forms are
// supported in the columns parameter:
//
//   1. ,name1,name2,...
//   2. ,name1{size1},name2{size2},...
//
void CXText::Resolve::toVars(Pchar _columns, Pchar _line) {
   // Parse the line into fields if columns is not empty
   char columns[sizeString];
   Strcpy(columns, _columns, sizeString);
   Trim(columns, backEnd);
   if (StrEmpty(columns))
      return;

   PCRecord cols = new CRecord(columns);

   char name[sizeString];
   char value[sizeString];

   // Determine if conversion from fixed-field or string-record
   if (ChrPos('{', _columns) > 0) {

      // Copy each fixed field into a named variable
      int len  = strlen(_line);
      int ofst = 0;
      for (int i = 1; i <= cols->Number(); i++) {
         cols->GetField(i, name, sizeString);
         Trim(name);

         // Locate the size expression
         int isz = ChrPos('{', name);
         if (isz < 0) {
            sprintf(message, Pchar("Size expression required 'name{size}'"
                                   "in field %d of columns variable <%s>"),
                             i,
                             _columns);
            throw(message);
            }

         // Extract the size and test if reasonable
         int size = StrToLong(name+isz+1);
         if (size <= 0 || ofst + size > len || size >= sizeString) {
            sprintf(message, Pchar("Specified field size in field %d "
                                   "of columns variable <%s> out of range"),
                             i,
                             _columns);
            throw(message);
            }

         // Isolate the name from the size expression
         name[isz] = 0;
         Trim(name);

         // Extract the value
         memset(value, 0, sizeString);
         memcpy(value, _line+ofst, size);

         // Create the named variable. (Note that fields with no name
         // are skipped)
         if (!StrEmpty(name))
            varHandler->Variable(name, value);

         ofst += size;
         }
      }
   else {
      // If the field delimiters are not the same, assume that the
      // line is not properly delimited, so use the columns delimiter
      // for the line.
      // Excel output will not have a leading delimiter of course, so
      // this is a way to handle it. Note that the columns string must
      // use the same delimiter that the line string uses for this to
      // work, and the columns string must start with the proper
      // delimiter.
      PCRecord line;
      if (isalnum(_line[0]) && _line[0] != columns[0])
         line = new CRecord(_line, columns);
      else
         line = new CRecord(_line);

      // Copy each string-record field into a named variable
      for (int i = 1; i <= line->Number(); i++) {
         cols->GetField(i, name, sizeString);
         Trim(name);
         line->GetField(i, value, sizeString);

         // Create the named variable
         if (!StrEmpty(name))
            varHandler->Variable(name, value);
         }
      }
   } // toVars

/*?
//---------------------------------------------------------------------
// debugging routine. Lists the names of the available varHandler.
//
void CXText::Resolve::dumpVars(PCVars _vars, bool _debug, Pchar _mssg) {
   if (!_debug && !CArgs::TheArgs->FArg(ftDebugFlag))
      return;

   char mssg[sizeString];
   if (StrEmpty(_mssg)) {
      Strcpy(mssg, Pchar("***Vars Dump:"), sizeString);
      }
   else if (strlen(_mssg) < sizeString) {
      Strcpy(mssg, _mssg, sizeString);
      }
   else
      Strcpy(mssg, Pchar("***Vars Dump:"), sizeString);

   log(false, mssg);

   PCStrList dump = new CStrList();
   for (int i = 0; i < _vars->GetWidth(); i++) {
      PCList list = _vars->GetList(i);
      if (list) {
         if (list->GotoHead()) do {
            PCItem item = list->Item();
            PCVar var   = TYPECAST(item, Var);
            PCAtom atom = var->Value();
            if (!atom)
               continue;

            Pchar line = toString(var);
            if (line && strlen(line) > 72)         // arbitrary, really
               line[71] = 0;

            char mssg[sizeLine];

            if      (var->Itsa(iBlock) ||
                     var->Itsa(iStrList)) {
               if      (!line) {
                  Strcpy(mssg, Pchar("variable empty"), sizeLine)
                  }
               else if (strlen(line) < sizeLine-2)
                  sprintf(mssg, "\"%s\Pchar(""), line);
               else {
                  memset(mssg, 0, sizeLine);
                  strncpy(mssg, line, sizeLine/2);
                  }
               }
            else if (var->Itsa(iNodeList)) {
               PCNodeList nodes = TYPECAST(atom, NodeList);
               sprintf(mssg, Pchar("nodeList(%d): %s"), nodes->Length(), line);
               }
            else if (var->Itsa(iNode)) {
               PCNode node = TYPECAST(atom, Node);
               if      (node->Isa(iText) ||
                        node->Isa(iAttr)) {
                  sprintf(mssg, "\"%s\Pchar(""), line);
                  }
               else if (node->Isa(iElement))
                  sprintf(mssg, Pchar("Element: %s"), line);
               else if (node->Isa(iDocument))
                  sprintf(mssg, Pchar("Document: %s"), line);
               else
                  sprintf(mssg, Pchar("%s type"), atom->IAm());
               }
            else
               sprintf(mssg, Pchar("%s type"), atom->IAm());

            delete[] line;

            char work[sizeLine];
            sprintf(work, Pchar("   var %s = %s"), var->name, mssg);
            work[128] = 0;
            dump->StrInsert(work, sorted);
            } while (list->GotoNext());
         }
      }

   if (dump->GotoHead()) do {
      dump->StrItem(mssg, sizeString);
      log(false, mssg);
      } while (dump->GotoNext());
   } // dumpVars
?*/

//---------------------------------------------------------------------
// Extract arg from Pchar("text(arg)") line. Ignore parentheses embedded in
// quotes.
//
void CXText::Resolve::extractArg(Pchar _line, short _size) {
   int beg = ChrPos(ftOpnChar, _line);

   if (beg < 0) {
      _line[0] = 0;
      return;
      }

   Strcpy(_line, _line+beg+1, _size);

   int i    = 1;
   int end  = 0;
   char inq = ' ';
   while (true) {
      if      (_line[end] == 0)
         THROW(Pchar("Badly formed argument"))
      else if (inq != ' ') {
         if (_line[end] == inq)
            inq = ' ';
         }
      else if (_line[end] == '\'')
         inq = '\'';
      else if (_line[end] == '"')
         inq = '"';
      else if (_line[end] == ftOpnChar)
         i++;
      else if (_line[end] == ftClsChar)
         i--;
      if      (i == 0)
         break;
      end++;
      }

   _line[end] = 0;
   Trim(_line);
   } // extractArg

//---------------------------------------------------------------------
//
void CXText::Resolve::expandInclude(PCStrList _list, Pchar _path) {
   _path;

   // Extract the include argument
   char line[sizeString];
   _list->StrItem(line, sizeString);
   extractArg(line, sizeString);

   // Expand the include argument
   PCVar var = resolveArg(line, true);

   sprintf(message, Pchar("Bad file name expansion for "
                          "Include command <%s>"),
                    line);

   if (!var)
      throw(message);

   if (!var->Itsa(iBlock))
      throw(message);

   PCAtom block = var->Value();
   PCStrList list = new CStrList();
   list->BlockTo(block);

   if (list->Number() != 1)
      throw(message);

   // Load the file specified in the include statement
   list->StrItem(1, line, sizeString);
   list->Flush();
   Trim(line);

   // Use include file name to load
   char fullName[sizeString];

   // Use script document path with include file name to load
   toString(Pchar("tagsPath"), tagsPath, sizeString);
   if (FindFileInPath(line, tagsPath, fullName, sizeString)) {
      if (!list->Load(fullName)) {
         sprintf(message, Pchar("Could not load include file %s"), line);
         throw(message);
         }
      }
   else {
      sprintf(message, Pchar("Could not load include file %s"), line);
      throw(message);
      }

   // Insert the contents of the included file into the main list
   int where = _list->Where();
   _list->ListInsert(list, after);
   _list->GotoItem(where);
   _list->Remove();
   _list->GotoItem(where);
   } // expandInclude

//---------------------------------------------------------------------
// Preprocess the includes, and then load the script document into
// the DOM.
//
PCDocument CXText::Resolve::LoadScript(PCStrList _scriptList, bool _preserveWhiteSpace) {
   // Resolve includes contained in the script document
   if (_scriptList->GotoHead()) do {
      char line[sizeString];
      _scriptList->StrItem(line, sizeString);
      Trim(line, frontEnd);
      CapStr(line);
      if (StrPos(ftIncStr, line) == 0) {
         expandInclude(_scriptList, scriptPath);
         _scriptList->GotoPrev();
         }
      } while (_scriptList->GotoNext());

//?if (CArgs::TheArgs->FArg(ftDebugFlag))
//?_scriptList->Save(Pchar("xtext.lst"));

   // Load the XML document into the DOM
   PCDocument script = new CDocument();
   script->PreserveWhiteSpace(_preserveWhiteSpace);
   if (!script->Load(_scriptList)) {
      PCDOMException error = script->GetError();
      sprintf(message, Pchar("Could not load the script document "
                             "into the DOM <%s>"),
                       error->Message());
      throw(message);
      }

   return script;
   } // LoadScript

//---------------------------------------------------------------------
// If the specified index is valid, returns the specified field from
// the delimited record in the first parameter. If the _len parameter is
// valid, concatenate up to that many fields/words/characters.
// Note: The manual says that negative indexes index from the right, or
// the bottom of the entity being indexed.
//
void CXText::Resolve::getField(Pchar _rec, char _mod, int _ndx, int _len, int _sz) {
   if (StrEmpty(_rec))
      return;
   
   if (_ndx == 0) {
      _rec[0] = 0;
      return;
      }
   
   _len = abs(_len);
   if (_len == 0)
      _len = 1;

   char mod = CapChr(_mod);
   if (_mod == ' ')
      mod = 'F';

   if      (mod == 'F') {                                   // by field
      PCRecord rec = new CRecord(_rec);
      int rnr = rec->Number();
      char delim = rec->GetDelimiter();
      rec->IgnoreQuotes();

      if (_ndx < 0)
         _ndx = rnr + _ndx + 1;
      if (_ndx > rnr)
         _ndx = rnr;
      if (_len > rnr - _ndx + 1)
         _len = rnr - _ndx + 1;

      _rec[0] = 0;
      Pchar wrk = new char[strlen(_rec+1)];
      for (int i = 0; i < _len; i++) {
         rec->GetField(_ndx+i, wrk, _sz);
         ChrCat(_rec, delim, _sz);
         Strcat(_rec, wrk, _sz);
         }
      if (_rec[0] == delim) {
         Strcpy(_rec+0, _rec+1, _sz);
         }
      }
   else if (mod == 'W') {                                    // by word
      PCRecord rec = new CRecord(_rec, Pchar(" "));
      int rnr = rec->Number();
      char delim = rec->GetDelimiter();
      rec->IgnoreQuotes();

      if (_ndx < 0)
         _ndx = rnr + _ndx + 1;
      if (_ndx > rnr)
         _ndx = rnr;
      if (_len > rnr - _ndx + 1)
         _len = rnr - _ndx + 1;

      _rec[0] = 0;
      Pchar wrk = new char[strlen(_rec+1)];
      for (int i = 0; i < _len; i++) {
         rec->GetField(_ndx+i, wrk, _sz);
         ChrCat(_rec, delim, _sz);
         Strcat(_rec, wrk, _sz);
         }
      if (_rec[0] == delim) {
         Strcpy(_rec+0, _rec+1, _sz);
         }
      }
   else if (mod == 'C') {                                    // by char
      int snr = strlen(_rec);
      if (_ndx < 0)
         _ndx = snr + _ndx + 1;
      if (_ndx > snr)
         _ndx = snr;
      Strcpy(_rec, _rec+_ndx-1, _sz);
      _len = abs(_len);
      if (_len >= snr - _ndx)
         _len = snr - _ndx;
      _rec[_len - 1] = 0;
      }
   } // getField

//---------------------------------------------------------------------
//
PCStrList CXText::Resolve::getListPart(PCStrList _list, char _mod, int _ndx, int _len) {
   PCStrList list = new CStrList();
   if (_list->Empty())
      return list;

   int lnr = _list->Number();
   if (_ndx < 0)
      _ndx = lnr + _ndx + 1;
   if (_ndx > lnr)
      _ndx = lnr;
   if (_len > lnr - _ndx + 1)
      _len = lnr - _ndx + 1;

   for (int i = 0; i < _len; i++)
      list->Insert(_list->Item(_ndx+i));

   return list;
   } // getListPart
   
//---------------------------------------------------------------------
//
PCNodeList CXText::Resolve::getNodeListPart(PCNodeList _nodes, char _mod, int _ndx, int _len) {
   PCNodeList nodes = new CNodeList();
   if (_nodes->Length() == 0)
      return nodes;

   int nnr = _nodes->Length();
   if (_ndx < 0)
      _ndx = nnr + _ndx + 1;
   if (_ndx > nnr)
      _ndx = nnr;
   if (_len > nnr - _ndx + 1)
      _len = nnr - _ndx + 1;

   for (int i = 0; i < _len; i++)
      nodes->InsertNode(_nodes->Item(_ndx+i));

   return nodes;
   } // getNodeListPart

//---------------------------------------------------------------------
// If the specified index is valid, returns the specified field from
// a delimited record, the specified line from a strlist, or the
// specified node from a nodelist. Does not affect vars containing
// other types. Note that the domtree definition of an index for a
// nodelist is zero-based. But in the interest of consistency, it is
// made one-based like the other types. Will have to make this
// consistent! This may come around to byte me if I don't.
//
PCVar CXText::Resolve::getPart(PCVar _var, char _mod, int _ndx, int _len) {
   if (!_var || _var->Empty())
      return new CVar();

   PCAtom atom = _var->Value();

   if      (_ndx == 0) {
      if      (_var->Itsa(iBlock)) {
         PCBlock blk = TYPECAST(atom, Block);
         Pchar p = blk->Lock();
         PCVar vout = new CVar(Pchar(""), p);
         blk->Unlock();
         return vout;
         }
      else if (_var->Itsa(iStrList)) {
         PCStrList list = TYPECAST(atom, StrList);
         PCAtom lout = new CStrList(list);
         return new CVar(Pchar(""), lout);
         }
      else if (_var->Itsa(iNodeList)) {
         PCNodeList nodes = TYPECAST(atom, NodeList);
         PCAtom nout = new CNodeList(nodes);
         return new CVar(Pchar(""), nout);
         }
      else
         return _var;
      }
   else {
      if      (_var->Itsa(iBlock)) {
         PCBlock blk = TYPECAST(atom, Block);
         Pchar p = blk->Lock();
         Pchar work = new char[strlen(p)+1];
         strcpy(work, p);
         blk->Unlock();
         getField(work, _mod, _ndx, _len, strlen(work)+1);
         PCAtom blk2 = new CBlock(work);
         delete[] work;
         return new CVar(Pchar(""), blk2);
         }
      else if (_var->Itsa(iStrList)) {
         PCStrList inpList = TYPECAST(atom, StrList);
         PCStrList outList = getListPart(inpList, _mod, _ndx, _len);
         PCAtom lOut       = new CStrList(outList);
         return new CVar(Pchar(""), lOut);
         }
      else if (_var->Itsa(iNodeList)) {
         PCNodeList inpNodes = TYPECAST(atom, NodeList);
         PCNodeList outNodes = getNodeListPart(inpNodes, _mod, _ndx, _len);
         PCAtom nOut         = new CNodeList(outNodes);
         return new CVar(Pchar(""), nOut);
         }
      else
         return _var;
      }
   } // getPart

//---------------------------------------------------------------------
// Use the XPath expression to search the specified node, or subscript.
// Calling XPath's Evaluate() method returns a CXPathItem, which is an
// object that can carry a result of any type that XPath works with.
// The resolvePath routine moves the contents of the returned
// CXPathItem into a CVar, converting to a string if specified in the
// _string parameter. Boolean CXPathItem values are always converted to
// strings.
//
// Called by resolveExp().
//
// If the string parameter is true, then attempt to return a CBlock
// object in the variable. If the string parameter is false, and the
// result of the XPath Evaluate call is a nodelist, then return a
// CNodeList. If no result or no result compatible with the string
// parameter, then the variable is returned empty.
//
PCVar CXText::Resolve::resolvePath(PCVar _context,
                                   Pchar _path,
                                   bool  _string) {
   // Check for an empty path
   if (StrEmpty(_path))
      THROW(Pchar("XPath search expression empty"))

if (!_context->Itsa(iNode))
   if (!_context->Itsa(iNode)) {
      char msg[sizeLine];
      sprintf(msg, Pchar("Path context node variable for %s must be node or nodeset"), _path);
      THROW(msg);
      }
   PCAtom atom = _context->Value();
   PCNode root = TYPECAST(atom, Node);

   // Evaluate the XPath expression against the contextNode
   PCXPathItem xItem = xp.Evaluate(root, _path);

   ulong error = xp.Error();
   if (error != FVE_OK) {
      sprintf(message, Pchar("XPath error %d encountered in expression <%s>"),
                       error,
                       _path);
      throw(message);
      }

   if (!xItem) {
      PCAtom block = new CBlock();
      PCVar vout   = new CVar();
      vout->Value(block);
      return vout;
      }

   // Process the results of the XPath search -------------------------

   // Create an unnamed variable for transport
   PCVar var = new CVar(Pchar(""));

   // Transfer the XPath search result into a variable. If a string,
   // then the variable contains a CBlock, if a nodelist, then the
   // variable contains the nodelist.
   uint type = xItem->Type();
   if      (type == XP_BOOLEAN) {
      char work[sizeWord];
      strcpy(work, xItem->Boolean() ? "true" : Pchar("false"));
      PCAtom block = new CBlock(work);
      var->Value(block);
      }

   else if (type == XP_NUMBER) {
      char work[sizeWord];
      sprintf(work, Pchar("%f"), *xItem->Number());
      // Trim insignificant zeros (n.0000 -> n)
      for (int i = strlen(work)-1; i > 0; i--)
         if      (work[i] == '.') {
            work[i] = 0;
            break;
            }
         else if (work[i] == '0')
            work[i] = 0;
         else
            break;

      PCAtom block = new CBlock(work);
      var->Value(block);
      }

   else if (type == XP_STRING) {
      Pchar p      = xItem->String();
      PCAtom block = new CBlock(p);
      delete[] p;
      var->Value(block);
      }

   else if (type == XP_NODESET) {
      PCNodeList nodes = xItem->NodeSet();
      PCNode node = nodes->Item(0);
      if (_string || (node && node->Isa(iText))) {
         if (node) {
            PCBlock block = new CBlock();
            if (node->Isa(iAttr) ||
                node->Isa(iText)) {
               Pchar p = node->NodeValue();
               block->NewBlock(p);
               delete[] p;
               }
            else                 // serialize the node and its children
               node->Save(block);
            PCAtom atom    = TYPECAST(block, Atom);
//?         PCStrList list = new CStrList();
//?         list->BlockTo(atom);
//?         atom = TYPECAST(list, Atom);
            var->Value(atom);
            }
         else { // return an empty block
            PCAtom block = new CBlock();
            var->Value(block);
            }
         }
      else {
         PCAtom atom = TYPECAST(nodes, Atom);
         var->Value(atom);
         }
      }

   return var;
   } // resolvePath

//---------------------------------------------------------------------
//
PCVar CXText::Resolve::getProperty(PCVar _var,
                                   bool  _type,
                                   bool  _size,
                                   bool  _flds,
                                   bool  _trim,
                                   bool  _cap) {
   if (!_var)
      return _var;

   PCAtom atom = _var->Value();

   if      (_type) {
      if      (_var->Itsa(iBlock)) {
         _var->Value(Pchar("string"));
         PCBlock blk = TYPECAST(atom, Block);
         Pchar p     = blk->Lock();
         if (IsNumber(p)) {
            _var->Value(Pchar("number"));
            }
         blk->Unlock();
         }
      else if (_var->Itsa(iStrList))
         _var->Value(Pchar("string_list"));
      else if (_var->Itsa(iNodeList))
         _var->Value(Pchar("node_list"));
      else if (_var->Itsa(iNode)) {
         PCNode node = TYPECAST(atom, Node);
         int type = node->NodeType();
         if      (type == ELEMENT_NODE)
            _var->Value(Pchar("element_node"));
         else if (type == ATTRIBUTE_NODE)
            _var->Value(Pchar("attribute_node"));
         else if (type == TEXT_NODE)
            _var->Value(Pchar("text_node"));
         else if (type == CDATA_SECTION_NODE)
            _var->Value(Pchar("cdata_section_node"));
         else if (type == ENTITY_REFERENCE_NODE)
            _var->Value(Pchar("entity_reference_node"));
         else if (type == ENTITY_NODE)
            _var->Value(Pchar("entity_node"));
         else if (type == PROCESSING_INSTRUCTION_NODE)
            _var->Value(Pchar("processing_instruction_node"));
         else if (type == COMMENT_NODE)
            _var->Value(Pchar("comment_node"));
         else if (type == DOCUMENT_NODE)
            _var->Value(Pchar("document_node"));
         else if (type == DOCUMENT_TYPE_NODE)
            _var->Value(Pchar("document_type_node"));
         else if (type == DOCUMENT_FRAGMENT_NODE)
            _var->Value(Pchar("document_fragment_node"));
         else if (type == NOTATION_NODE)
            _var->Value(Pchar("notation_node"));
         }
      }
   else if (_size) {
      if      (_var->Itsa(iBlock)) {
         PCBlock blk = TYPECAST(atom, Block);
         Pchar p     = blk->Lock();
         _var->Value(strlen(p));
         blk->Unlock();
         }
      else if (_var->Itsa(iStrList)) {
         PCStrList list = TYPECAST(atom, StrList);
         _var->Value(list->Number());
         }
      else if (_var->Itsa(iNodeList)) {
         PCNodeList nodes = TYPECAST(atom, NodeList);
         _var->Value(nodes->Length());
         }
      else if (_var->Itsa(iNode))
         _var->Value(1);
      }
   else if (_flds) {
      if      (_var->Itsa(iBlock)) {
         PCBlock blk = TYPECAST(atom, Block);
         Pchar p     = blk->Lock();
         PCRecord r  = new CRecord(p);
         _var->Value(r->Number());
         blk->Unlock();
         }
      else
         _var->Value(0L);
      }
   else if (_trim) {
      if (_var->Itsa(iBlock)) {
         PCBlock blk = TYPECAST(atom, Block);
         Pchar p     = blk->Lock();
         Trim(p, bothEnds);
         _var->Value(p);
         blk->Unlock();
         }
      }
   else if (_cap) {
      if (_var->Itsa(iBlock)) {
         PCBlock blk = TYPECAST(atom, Block);
         Pchar p     = blk->Lock();
         Trim(p, bothEnds);
         if (strlen(p) > 0) {
            CapChr(*p);
            }
         _var->Value(p);
         blk->Unlock();
         }
      }

   return _var;
   } // getProperty

//---------------------------------------------------------------------
// Checks the specified path expression to contain a subscript, and
// isolates the subscript index for later use. If the path expression
// is an XPath expression, then calls resolvePath. Otherwise, obtains
// the variable named by the expression. Once the result has been
// obtained, the subscript index (if any) is used to extract the string
// field (if the result is a string), or the node (if the result is a
// nodelist). If the result is not of either type, it is returned
// regardless of the value of the subscript index.
//
// Called by resolveArg() and resolveString().
//
PCVar CXText::Resolve::resolveExp(Pchar _path, bool _string) {
   if (StrEmpty(_path)) {
      PCAtom block = new CBlock();
      PCVar vout = new CVar();
      vout->Value(block);
      return vout;
      }

   char path[sizeString];
   Strcpy(path, _path, sizeString);
   char type = path[0];
   strcpy(path, path+1);

   // Check for a subscript.
   // Format is "{" ( "f" or "c" or none) index ["," length] "}"
   // where "f" means field index and "c" means character index. Only
   // use the split-spec with strings, ignored for lists. Serious note:
   // variable subscripts are not supported. This is very BAD!!!! It is
   // a parsing issue.
   char mod = ' ';
   int  ndx = 0;
   int  len = 0;
   int  i = ChrPos('{', path);
   int  j = ChrPos('}', path);
   if (i > 0) {
      if (j < i)
         THROW(Pchar("Subscript badly formed"))

      path[i] = 0;
      path[j] = 0;
      int icma = ChrPos(',', path+i+1);
      if (icma > 0) {
         path[i+icma+1] = 0;
         path[i+icma+1] = 0;
         Trim(path+i+icma+2, bothEnds);            // do this one first
         len = StrToShort(path+i+icma+2);
         }
      Trim(path+i+1, bothEnds);
      if (isalpha(*(path+i+1))) {
         mod = *(path+i+1);
         i++;
         }
      ndx = StrToShort(path+i+1);
      }

   // Process the specified XPath expression
   if      (type == ftPathChar) {
      // Use the node in the contextNode variable as the root of the
      // search
      PCVar var = varHandler->Variable(Pchar("contextNode"));
      if (!var)
         THROW(Pchar("Node variable missing"))

      if (var->Itsa(iNodeList)) {
         PCAtom atom      = var->Value();
         PCNodeList nodes = TYPECAST(atom, NodeList);
         if (nodes->Length()) {
            PCNode node = nodes->Item(0);
            atom        = TYPECAST(node, Atom);
            var->Value(atom);
            }
         }

      var = resolvePath(var, path, _string);
      PCVar vout = getPart(var, mod, ndx, len);
      return vout;
      }

   else if (type == ftCtxtChar) {
     // Use the node in the path string as the root of the search
     // form is $@varname!path$$
     char name[sizeLine];
     if (ChrPos('!', path) < 0)
        THROW(Pchar("Path expression badly formed"))

     // Get the context node variable name
     Before(Pchar("!"), path, name, sizeLine);
     Trim(name, bothEnds);
     if (StrEmpty(name))
        THROW(Pchar("Node variable name missing from path (viz., $@name!path$$)"))

     // Get the xpath expression
     char xpath[sizeLine];
     After(Pchar("!"), path, xpath, sizeLine);
     Trim(xpath, bothEnds);

     PCVar var = varHandler->Variable(name);
     if (!var) {
        sprintf(message, Pchar("Node variable name <%s> has no variable"), name);
        throw(message);
        }

     if (var->Itsa(iNodeList)) {
        PCAtom atom      = var->Value();
        PCNodeList nodes = TYPECAST(atom, NodeList);
        if (nodes->Length()) {
           PCNode node = nodes->Item(0);
           atom        = TYPECAST(node, Atom);
           var->Value(atom);
           }
        }

      var = resolvePath(var, xpath, _string);
      PCVar vout = getPart(var, mod, ndx, len);
      return vout;
      }

   // Process the specified variable reference
   else if (type == ftVarChar) {
      boolean varType = false;
      boolean varSize = false;
      boolean varFlds = false;
      boolean varTrim = false;
      boolean varCap  = false;
      if      (path[0] == ftTypeChar) {
         varType = true;
         strcpy(path, path+1);
         }
      else if (path[0] == ftSizeChar) {
         varSize = true;
         strcpy(path, path+1);
         }
      else if (path[0] == ftFldsChar) {
         varFlds = true;
         strcpy(path, path+1);
         }
      else if (path[0] == ftTrimChar) {
         varTrim = true;
         strcpy(path, path+1);
         }
      else if (path[0] == ftCapChar) {
         varCap = true;
         strcpy(path, path+1);
         }

      // Check for an empty name
      if (StrEmpty(path))
         THROW(Pchar("Name required to identify variable"))

      // Check for the Pchar("random") variable
      if (stricmp(path, Pchar("random")) == 0) {
         PCVar vout = new CVar();
         setVar(vout, rand());
         return getProperty(vout, varType, varSize, varFlds, varTrim, varCap);
         }

      // Check for uniqueID
      if (stricmp(path, Pchar("uniqueID")) == 0) {
         char value[sizeString];
         PCVar vout = new CVar();
         setVar(vout, GetID(value, sizeString));
         return getProperty(vout, varType, varSize, varFlds, varTrim, varCap);
         }

      // Find the value of the specified variable
      PCVar var;
      if (StrPos(Pchar("env."), path) != 0)
         var = varHandler->Variable(path);

      // Find the value of the requested environment variable
      else {
         strcpy(path, path+4);
         char value[sizeString];
         memset(value, 0, sizeString);
         PCAtom blk;
         if (::GetEnvironmentVariable(path, value, sizeString))
            blk = new CBlock(value);
         else
            blk = new CBlock();
         PCVar var = new CVar();
         var->Value(blk);
         PCVar vout = getPart(var, mod, ndx, len);
         return getProperty(vout, varType, varSize, varFlds, varTrim, varCap);
         }

      if      (!var) {
         PCAtom blk = new CBlock(Pchar(""));
         PCVar vout = new CVar();
         vout->Value(blk);
         return vout;
         }

      // Convert it to a string if requested
      else if (!var->Itsa(iBlock) && _string) {
         toString(var, path, sizeString);
         setVar(var, path);
         }

//?    var = new CVar(var);
       PCVar vout = getPart(var, mod, ndx, len);
       return getProperty(vout, varType, varSize, varFlds, varTrim, varCap);
      }

   else if (type == ftPopChar) {
      // Properties apply to the variable popped from the stack, not the
      //  stack itself.
      boolean varType = false;
      boolean varSize = false;
      boolean varFlds = false;
      boolean varTrim = false;
      boolean varCap  = false;
      if      (path[0] == ftTypeChar) {
         varType = true;
         strcpy(path, path+1);
         }
      else if (path[0] == ftSizeChar) {
         varSize = true;
         strcpy(path, path+1);
         }
      else if (path[0] == ftFldsChar) {
         varFlds = true;
         strcpy(path, path+1);
         }
      else if (path[0] == ftTrimChar) {
         varTrim = true;
         strcpy(path, path+1);
         }
      else if (path[0] == ftCapChar) {
         varCap = true;
         strcpy(path, path+1);
         }

      // Check for an empty stack variable name
      if (StrEmpty(path))
         THROW(Pchar("Name required to identify stack variable"))

      // Find the stack variable
      PCVar stk = varHandler->Variable(path);

      if (!stk) {
         PCAtom blk = new CBlock(Pchar(""));
         PCVar vout = new CVar();
         vout->Value(blk);
         return vout;
         }

      PCVar var = stk->Pop();

      // Convert to a string if requested
      if (!var->Itsa(iBlock) && _string) {
         toString(var, path, sizeString);
         setVar(var, path);
         }

       PCVar vout = getPart(var, mod, ndx, len);
       return getProperty(vout, varType, varSize, varFlds, varTrim, varCap);
      }

   else
      THROW(Pchar("Unknown expression type"))
   } // resolveExp

//---------------------------------------------------------------------
//
bool CXText::Resolve::findExp(Pchar _line, int& _beg, int& _end) {
   int ofst = 0;
   for (;;) {
      // Find the right-hand end of the first expression in the line
      _end = StrPos(Pchar(ftEndStr), _line+ofst);
      if (_end < 0)
         return false;

      _end += ofst;

      // Find the left-hand end of the first expression in the line
      _beg = _end - 2;
      while (_beg && (_line[_beg]   != ftEscChar  ||
                     (_line[_beg+1] != ftVarChar  &&
                      _line[_beg+1] != ftPathChar &&
                      _line[_beg+1] != ftCtxtChar)))
         _beg--;

      if (_line[_beg]   != ftEscChar  ||
         (_line[_beg+1] != ftVarChar  &&
          _line[_beg+1] != ftPathChar &&
          _line[_beg+1] != ftCtxtChar))
         ofst = _end+1;
      else
         return true;
      }
   } // findExp

//---------------------------------------------------------------------
//
PCVar CXText::Resolve::resolveArg(Pchar _line, bool _string) {
   // Validate the expression
   if (StrEmpty(_line)) {
      PCVar var    = new CVar();
      PCAtom block = new CBlock();
      var->Value(block);
      return var;
      }

   char line[sizeString];
   Strcpy(line, _line, sizeString);

   for (;;) {
      // Find the first expression in the line. Note that the line may
      // not actually contain an expression.
      int beg, end;
      if (!findExp(line, beg, end)) {
         PCVar vout = new CVar();
         vout->Value(line);
         return vout;
         }

      // Remove the part of the line to the right of the expression
      char rightPart[sizeString];
      Strcpy(rightPart, line+end+2, sizeString);
      line[end] = 0;

      // Remove the part of the line to the left of the expression
      char leftPart[sizeString];
      line[beg] = 0;
      Strcpy(leftPart, line, sizeString);
      strcpy(line, line+beg+1);

      // Expand the expression
      PCVar var = resolveExp(line, false);

      if (var->Itsa(iNode) || var->Itsa(iNodeList)) {
         if (_string) {
            Pchar p = toString(var);
            PCAtom blk = new CBlock(p);
            delete[] p;
            var->Value(blk);
            }
         return var;
         }

      // Process the resolve result -----------------------------------

      bool more = (!StrEmpty(leftPart) ||
                   !StrEmpty(rightPart));

//?   if (!_string && !more)
//?      return var;

      // Reconstitute the resolved string
      toString(var, line, sizeString);
      Strcat(leftPart, line, sizeString);
      Strcat(leftPart, rightPart, sizeString);

      if      (more || StrPos(Pchar(ftEndStr), leftPart) > 0) {
         Strcpy(line, leftPart, sizeString);
         }
      else if (var->Itsa(iBlock)) {
         PCVar vout = new CVar();
         setVar(vout, leftPart);
         return vout;
         }
      else if (_string) {
         PCVar vout = new CVar();
         setVar(vout, leftPart);
         return vout;
         }
      else
         return var;
      }
   } // resolveArg

//---------------------------------------------------------------------
// Extract arguments from Pchar("arg1(arg2)") line. Ignore parentheses
// embedded in quotes and parentheses preceeded by the escape
// character (`).
//
void CXText::Resolve::extractArgs(Pchar _line, Pchar _arg1, short _sz1,
                                               Pchar _arg2, short _sz2) {
   _arg1[0] = 0;
   _arg2[0] = 0;

   Pchar arg = _arg1;
   short sz  = _sz1;

   int i    = 1;
   int end  = 0;
   char inq = ' ';
   while (true) {
      if      (_line[end] == 0)
         return;
      else if (inq != ' ') {
         ChrCat(arg, _line[end], sz);
         if (_line[end] == inq)
            inq = ' ';
         }
      else if (_line[end] == '`') {
         ChrCat(arg, _line[end+1], sz);
         end++;
         }
      else if (_line[end] == '\'') {
         ChrCat(arg, _line[end], sz);
         inq = '\'';
         }
      else if (_line[end] == '"') {
         ChrCat(arg, _line[end], sz);
         inq = '"';
         }
      else if (_line[end] == ftOpnChar) {
         if (arg == _arg1) {
            arg = _arg2;
            sz  = _sz2;
            }
         else {
            ChrCat(arg, '(', sz);
            i++;
            }
         }
      else if (_line[end] == ftClsChar) {
         i--;
         if (i != 0)
            ChrCat(arg, ')', sz);
         }
      else
         ChrCat(arg, _line[end], sz);

      if (i == 0)
         break;
      end++;
      }
   } // extractArgs

//---------------------------------------------------------------------
//
bool CXText::Resolve::match(Pchar _text, Pchar _exp) {
   PCVar vtext = resolveArg(_text, true);
   Pchar text = toString(vtext);

   PCVar vexp  = resolveArg(_exp,  true);
   char exp[sizeString];
   toString(vexp, exp, sizeString);

   ulong nr;

   // Match the regular expression against the specified text
   CRegX regx(exp);
   if (regx.Match(text))
      nr = regx.Number();
   else
      nr = 0;

   // Create a variable to contain the number of substring matches
   varHandler->Variable(Pchar("regXCount"), nr);

/*?
   // Create a string record variable to contain the substring matches
   char regX[sizeString];
   
   regX[0] = 0xff;
   regX[1] = 0;
   for (uint i = 0; i < nr; i++) {
      Pchar val = regx.SubString(i, text);
      Strcat(regX, val, sizeString);
      ChrCat(regX, 0xff, sizeString);
      delete[] val;
      }
   varHandler->Variable(Pchar("regX"), regX);
?*/   
   // Create variables for each matched substring
   char nm[sizeString];
   for (uint i = 0; i < nr; i++) {
      sprintf(nm, Pchar("regX%d"), i);
      Pchar val = regx.SubString(i, text);
      varHandler->Variable(nm, val);
      delete[] val;
      }

   delete[] text;

   nrRegVars = nr;
   return (nr != 0);
   } // match

//---------------------------------------------------------------------
// Expand the specified text line to return a string list.
//
PCStrList CXText::Resolve::resolveString(Pchar _line, bool _resolve) {
   // Isolate the expression (if any) before expanding it
   if (StrEmpty(_line))
      return 0;

   char line[sizeString];
   Strcpy(line, _line, sizeString);
//?string *sline = new string(_line);

   // Find the first expression in the line. Note that the line may
   // not actually contain an expression.
   int beg, end;
   if (!findExp(line, beg, end))
      return 0;

   // Remove the part of the line to the right of the expression
   char rightPart[sizeString];
   Strcpy(rightPart, line+end+2, sizeString);
   line[end] = 0;

   // Remove the part of the line to the left of the expression
   char leftPart[sizeString];
   line[beg] = 0;
   Strcpy(leftPart, line, sizeString);
   strcpy(line, line+beg+1);

   // Expand the expression
   PCVar var = resolveExp(line, true);

   // Process the result of the search --------------------------------

   if (!var->Itsa(iBlock))
      THROW(Pchar("String expected"))

   PCStrList work = new CStrList();
   work->BlockTo(var->Value());

   // Attach the saved parts to the list
   if      (work->GotoHead()) {
      // Attach the saved left part to the first line
      work->StrItem(line, sizeString);
      Strcat(leftPart, line, sizeString);
      work->StrReplace(leftPart);

      if (_resolve) {
         PCStrList out = new CStrList();
         Resolve* r = new Resolve();
         r->resolve(work, out);
         delete r;
         out->GotoTail();
         out->StrItem(line, sizeString);
         Strcat(line, rightPart, sizeString);
         out->StrReplace(line);
         return out;
         }
      else {
         work->GotoTail();
         work->StrItem(line, sizeString);
         Strcat(line, rightPart, sizeString);
         work->StrReplace(line);
         return work;
         }

/*?
      PCStrList out;
      if (!_output)
         out = new CStrList();
      else
         out = _output;

      // Resolve the contents of the list
      Resolve(work, out);

      // Attach the saved right part to the last line
      out->GotoTail();
      out->StrItem(line, sizeString);
      Strcat(line, rightPart, sizeString);
      out->StrReplace(line);
      return out;
?*/
      }

   else {
      Strcat(leftPart, rightPart, sizeString);
//?   if (!StrEmpty(leftPart))
         work->StrInsert(leftPart);
      return work;
      }
   } // resolveString

//---------------------------------------------------------------------
//
bool CXText::Resolve::isFalse(Pchar _line) {
   return (StrEmpty(_line)               ||
           strcmp  (_line, Pchar("0"))     == 0 ||
           stricmp (_line, Pchar("no"))    == 0 ||
           stricmp (_line, Pchar("off"))   == 0 ||
           stricmp (_line, Pchar("false")) == 0);
   } // isFalse

//---------------------------------------------------------------------
//
bool CXText::Resolve::isTrue(Pchar _exp) {
   PCVar var   = resolveArg(_exp, false);
   if (!var)
      return false;

   PCAtom atom = var->Value();
   if (!atom)
      return false;

   if (var->Itsa(iBlock)) {
      PCBlock blk = TYPECAST(atom, Block);
      if (blk->Empty())
         return false;

      Pchar p = blk->Lock();
      Pchar line = new char[strlen(p)+1];
      strcpy(line, p);
      blk->Unlock();
      Trim(line);
      Squeeze(line, '0');
      bool test = !isFalse(line);
      delete[] line;
      return test;
      }

   else if (var->Itsa(iStrList)) {
      PCStrList lst = TYPECAST(atom, StrList);
      if      (lst->Empty())
         return false;
      else if (lst->Number() > 1)
         return true;
      else {
         char line[sizeString];
         lst->StrItem(1, line, sizeString);
         Trim(line);
         Squeeze(line, '0');
         return !isFalse(line);
         }
      }

   else if (var->Itsa(iNode))
      return true;

   else if (var->Itsa(iNodeList)) {
      PCNodeList nodes = TYPECAST(atom, NodeList);
      return (nodes->Length() > 0);
      }

   else
      return false;
   } // isTrue

//---------------------------------------------------------------------
// if tos is hardfalse, push hardfalse.
// if tos is false, push hardfalse.
// if tos is true, and this if is true, push true.
// if tos is true, and this if is false, push false.
//
void CXText::Resolve::cIf(Pchar _text, Pchar _exp) {
   dbgLog(commandLog, Pchar("if %s (%s)"), _text, _exp);

   if      (condition.IsHardFalse())
      condition.Push(HardFalse);
   else if (condition.IsFalse())
      condition.Push(HardFalse);
   else if (StrEmpty(_text))
      if (isTrue(_exp))
         condition.Push(True);
      else
         condition.Push(False);
   else if (match(_text, _exp))
      condition.Push(True);
   else
      condition.Push(False);
   } // cIf

//---------------------------------------------------------------------
// if tos is hardfalse, push hardfalse.
// if tos is false, push hardfalse.
// if tos is true, and this if is false, push true.
// if tos is true, and this if is true, push false.
//
void CXText::Resolve::cIfN(Pchar _text, Pchar _exp) {
   dbgLog(commandLog, Pchar("ifn %s (%s)"), _text, _exp);

   if      (condition.IsHardFalse())
      condition.Push(HardFalse);
   else if (condition.IsFalse())
      condition.Push(HardFalse);
   else if (StrEmpty(_text))
      if (!isTrue(_exp))
         condition.Push(True);
      else
         condition.Push(False);
   else if (!match(_text, _exp))
      condition.Push(True);
   else
      condition.Push(False);
   } // cIfN

//---------------------------------------------------------------------
// if tos is hardfalse, ignore this elif.
// if tos is false, and this elif is true, set tos to true.
// if tos is true, set tos to hardfalse.
//
void CXText::Resolve::cElif(Pchar _text, Pchar _exp) {
   dbgLog(commandLog, Pchar("elif %s (%s)"), _text, _exp);

   if      (condition.IsHardFalse())
      return;
   else if (condition.IsFalse()) {
      if (StrEmpty(_text)) {
         if (isTrue(_exp))
            condition.Set(True);
//?         else
//?            condition.Set(False);
         }
      else if (match(_text, _exp))
         condition.Set(True);
//?      else
//?         condition.Set(False);
      }
   else
      condition.Set(HardFalse);
   } // cElif

//---------------------------------------------------------------------
// if tos is hardfalse, ignore this elifN.
// if tos is false, and this elif is false, set tos to true.
// if tos is true, set tos to hardfalse.
//
void CXText::Resolve::cElifN(Pchar _text, Pchar _exp) {
   dbgLog(commandLog, Pchar("elifn %s (%s)"), _text, _exp);

   if      (condition.IsHardFalse())
      return;
   else if (condition.IsFalse())
      if (StrEmpty(_text))
         if (!isTrue(_exp))
            condition.Set(True);
//?         else
//?            condition.Set(False);
      else if (!match(_text, _exp))
         condition.Set(True);
//?      else
//?         condition.Set(False);
   else
      condition.Set(HardFalse);
   } // cElifN

//---------------------------------------------------------------------
// if tos is hardfalse, ignore this else.
// if tos is false, set tos to true.
// if tos is true, set tos to hardfalse.
//
void CXText::Resolve::cElse(void) {
   dbgLog(commandLog, Pchar("else"));

   if      (condition.IsHardFalse())
      return;
   else if (condition.IsFalse())
      condition.Set(True);
   else
      condition.Set(HardFalse);
   } // cElse

//---------------------------------------------------------------------
//
void CXText::Resolve::cEnd(void) {
   dbgLog(commandLog, Pchar("end"));

   condition.Pop();
   } // cEnd

//---------------------------------------------------------------------
// Extract the portion of the specified body list enclosed by the
// forEach..end or while..end
//
PCStrList CXText::Resolve::extractForLoop(PCStrList _body) {
   PCStrList body = new CStrList();
   int iEnd       = 1;
   char line[sizeString];
   if (_body->GotoNext()) do {
      _body->StrItem(line, sizeString);
      body->StrInsert(line);
      Trim(line);
      CapStr(line);
      int ofst = strlen(Pchar(ftCmdStr));
      if (StrPos(Pchar(ftCmdStr), line) == 0) { // is it a command line
         if      (strncmp(Pchar("IF"),      line+ofst, 2) == 0)
            iEnd++;
         else if (strncmp(Pchar("FOREACH"), line+ofst, 7) == 0)
            iEnd++;
         else if (strncmp(Pchar("WHILE"),   line+ofst, 5) == 0)
            iEnd++;
         else if (strncmp(Pchar("FORELSE"), line+ofst, 7) == 0) {
            iEnd--;
            if (iEnd == 0) {
               _body->GotoPrev();
               body->StrReplace(Pchar("$#end"));
               break;
               }
            }
         else if (strncmp(Pchar("END"),     line+ofst, 3) == 0) {
            iEnd--;
            // The line below did not work correctly when on sep lines.
            // A Borland Bug!?
            if (iEnd == 0) break;
            }
         }
      } while (_body->GotoNext());

   if (iEnd)
      THROW(Pchar("Unmatched forEach/while command"))

   return body;
   } // extractForLoop

//---------------------------------------------------------------------
//
bool CXText::Resolve::getNextRow(ODBCInterface &_db,
                                 PCBlock        _blk,
                                 Pchar          _columns) {
   PCList row = _db.GetRow();
   if (!row)
      return false;

   PCRecord columns = new CRecord(_columns);

   // obtain the values from the result list
   for (int i = 1; i <= row->Number(); i++) {
      PCItem item = row->Item(i);
      PCODBCBindCol colval = TYPECAST(item, ODBCBindCol);

      int len = colval->GetSize();
      if (colval->IsEmpty())
         len = 0;

      Pchar value = new char[len+2];

      value[0] = char(0xff);
      value[1] = 0;

      // Append the value to a delimited string in the specified block
      if (!colval->IsEmpty()) {
         Strcpy(value+1, colval->GetValue(), len-1);
         }
      _blk->AddString(value);

      // Create a variable whose name is specified in columns
      int size = columns->LenField(i)+1;
      Pchar name = new char[size];
      columns->GetField(i, name, size);
      Trim(name);
      if (StrEmpty(name)) {
         delete[] value;
         sprintf(message, Pchar("name expected at %d in %s."),
                          i,
                          columns->GetDelimStr(false));
         throw(message);
         }

      // Initialize the named variable to the new value
      varHandler->Variable(name, value+1);
      delete[] value;
      }

   return true;
   } // getNextRow

//---------------------------------------------------------------------
//
bool CXText::Resolve::doLoop(long      _iTop,
                             PCStrList _body,
                             PCStrList _output) {
   // Isolate the body of the forEach/while loop
   _body->GotoItem(_iTop);
   PCStrList body = extractForLoop(_body);

   condition.Push(True);

   Resolve* r = new Resolve();
   r->resolve(body, _output);
   delete r;
   if (lastLoop || nextLoop) {
      while (!condition.IsMark())
         condition.Pop();
      if (lastLoop)
         return false;
      }

   return true;
   } // doLoop

//---------------------------------------------------------------------
// Loop for each character in the string.
//
bool CXText::Resolve::forEachChar(Pchar        _name,
                                  Pchar        _exp,
                                  PCStrList    _body,
                                  PCStrList    _output,
                                  forEachVars& _vars) {
   char str[sizeString];
   toString(_vars.var, str, sizeString);
   if (!StrEmpty(str)) {
      // Perform the forEach loop
      ulong nr = strlen(str);
      _vars.last->Value(nr);
      for (uint i = 0; i < nr; i++) {
         _vars.position->Value(i+1);

         // Set the foreach variable to the current character
         char ch[2];
         ch[0] = str[i];
         ch[1] = 0;

         // If the grep string is not empty, match it against the word
         if (!StrEmpty(_vars.grepStr) && match(ch, _vars.grepStr) == 0)
             continue;

         PCAtom blk = new CBlock(ch);
         _vars.foreach->Value(blk);

         PCAtom atom;

         // Restore the contextNode variable
         atom = _vars.holdNd->Value();
         _vars.ctxtnode->Value(atom);

         // Clear the grep variable
         if (_vars.grep)
            setVar(_vars.grep, Pchar(""));

         if (!doLoop(_vars.iTop, _body, _output))
            break;
         if (!condition.IsMark())
            THROW(Pchar("Stack corrupted in forEach field command"))
         }
      }
   else {       // Skip the body code within the forEach..end structure
      extractForLoop(_body);
      return false;
      }

   return true;
   } // forEachChar

//---------------------------------------------------------------------
// Loop for each number between 1 and value.
//
bool CXText::Resolve::forEachCount(Pchar        _name,
                                   Pchar        _exp,
                                   PCStrList    _body,
                                   PCStrList    _output,
                                   forEachVars& _vars) {
   char value[sizeString];
   toString(_vars.var, value, sizeString);
   long hi = StrToLong(value);
   if (hi > 0)
      for (long i = 1; i <= hi; i++) {
         _vars.last->Value(hi);
         _vars.position->Value(i);
         _vars.foreach->Value(i);

         // Restore the contextNode variable
         PCAtom atom = _vars.holdNd->Value();
         _vars.ctxtnode->Value(atom);

         if (!doLoop(_vars.iTop, _body, _output))
            break;
         if (!condition.IsMark())
            THROW(Pchar("Stack corrupted in forEach count command"))
         }
   else {       // Skip the body code within the forEach..end structure
      extractForLoop(_body);
      return false;
      }

   return true;
   } // forEachCount

//---------------------------------------------------------------------
// Loop for each field in the record in value
//
bool CXText::Resolve::forEachField(Pchar        _name,
                                   Pchar        _exp,
                                   PCStrList    _body,
                                   PCStrList    _output,
                                   forEachVars& _vars) {
   char str[sizeString];
   toString(_vars.var, str, sizeString);
   if (!StrEmpty(str)) {
      // Perform the forEach loop
      PCRecord rec = new CRecord(str);
      ulong nr = rec->Number();
      for (uint i = 1; i <= nr; i++) {
         _vars.last->Value(nr);
         _vars.position->Value(i);

         // Set the foreach variable to the current field
         rec->GetField(i, str, sizeString);

         // If the grep string is not empty, match it against the
         // field
         if (!StrEmpty(_vars.grepStr) && match(str, _vars.grepStr) == 0)
             continue;

         PCAtom blk = new CBlock(str);
         _vars.foreach->Value(blk);

         PCAtom atom;

         // Restore the contextNode variable
         atom = _vars.holdNd->Value();
         _vars.ctxtnode->Value(atom);

         // Clear the grep variable
         if (_vars.grep)
            setVar(_vars.grep, Pchar(""));

         if (!doLoop(_vars.iTop, _body, _output))
            break;
         if (!condition.IsMark())
            THROW(Pchar("Stack corrupted in forEach field command"))
         }
      }
   else {       // Skip the body code within the forEach..end structure
      extractForLoop(_body);
      return false;
      }

   return true;
   } // forEachField

//---------------------------------------------------------------------
// Loop for each file in a directory. The form of the forEach argument
// is
//    dirName|fileMask|searchtype
// where
//    searchtype={ *root | tree & *data | dir | any }
//
bool CXText::Resolve::forEachFile(Pchar        _name,
                                  Pchar        _exp,
                                  PCStrList    _body,
                                  PCStrList    _output,
                                  forEachVars& _vars) {
   PCAtom atom;

   // Save the fileinfo variable
   PCVar fi = varHandler->Variable(Pchar("fileInfo"));
   Assert(fi);
   atom = fi->Value();
   PCVar holdFI = new CVar();
   holdFI->Value(atom);

   // Save the filedrive variable
   PCVar fd = varHandler->Variable(Pchar("fileDrive"));
   Assert(fd);
   atom = fd->Value();
   PCVar holdFD = new CVar();
   holdFD->Value(atom);

   // Save the filepath variable
   PCVar fp = varHandler->Variable(Pchar("filePath"));
   Assert(fp);
   atom = fp->Value();
   PCVar holdFP = new CVar();
   holdFP->Value(atom);

   // Save the contents of the filename variable
   PCVar fn = varHandler->Variable(Pchar("fileName"));
   Assert(fn);
   atom = fn->Value();
   PCVar holdFN = new CVar();
   holdFN->Value(atom);

   // Parse the directory search string
   char dirName[sizeString];
   toString(_vars.var, dirName, sizeString);
   Trim(dirName);

   // Separate the directory name from the file mask and search type
   PCRecord r = new CRecord(dirName);
   r->GetField(1, dirName, sizeString);
   if (StrEmpty(dirName))
      ChrCpy(dirName, '.');
   if (dirName[strlen(dirName)-1] != '\\')// platform dependence!!??
      ChrCat(dirName, '\\', sizeString);

   // Get the file mask
   char fileMask[sizeString];
   r->GetField(2, fileMask, sizeString);

   // Get the searchType = { *root | tree & *data | dir | any }
   char searchType[sizeString];
   r->GetField(3, searchType, sizeString);

   if (!FindFileInPath(dirName, scriptPath, dirName, sizeString)) {
//?      sprintf(message, Pchar("Could not find forEach-file directory <%s>"), dirName);
//?      throw(message);
      }

   // Set up the directory search object
   PCFileList dir = new CFileList();
   CapStr(searchType);
   dir->GetFileNames(dirName,
                     StrPos(Pchar("TREE"), searchType) >= 0,
                     StrPos(Pchar("DIR"),  searchType) >= 0 ? dirFile :
                     StrPos(Pchar("ANY"),  searchType) >= 0 ? anyFile :
                                                       datFile);
   dir->SetMask(fileMask);

   char name[sizeString];
   int pos = 0;

   // Perform the forEach loop
   if (dir->GetFirstName(name, sizeString)) do {
      _vars.last->Value(-1);         // last is not valid for this loop

      // Set the position variable
      pos++;
      _vars.position->Value(pos);

      // Set the foreach variable to the line
      _vars.foreach->Value(name);

      // Get the file information record into a Tags variable
      PCRecord fileInfo = new CRecord();
      dir->GetFileInfo(fileInfo);
      Pchar info = fileInfo->GetDelimStr(false);
      varHandler->Variable(Pchar("fileInfo"), info);

      char driv[sizeString]; driv[0] = 0;
      char path[sizeString]; path[0] = 0;
      char nwrk[sizeString]; nwrk[0] = 0;
      char next[sizeString]; next[0] = 0;

      // Set the file variables
      Strcpy(nwrk, name, sizeString);
      ParseName(path, nwrk, next, sizeString, 0);
      Before(Pchar(":"), path, driv, sizeString);
      ChrCat(driv, ':', sizeString);
      fd->Value(driv);
      After (Pchar(":"), path, driv, sizeString);
      fp->Value(driv);
      if (strlen(next) > 0) {
         ChrCat(nwrk, '.', sizeString);
         Strcat(nwrk, next, sizeString);
         }
      fn->Value(nwrk);

      // Restore the contextNode variable
      atom = _vars.holdNd->Value();
      _vars.ctxtnode->Value(atom);

      if (!doLoop(_vars.iTop, _body, _output))
         break;
      if (!condition.IsMark())
         THROW(Pchar("Stack corrupted in forEach file command"))
      } while (dir->GetNextName(name, sizeString));
   else {            // Skip the code within the forEach..end structure
      extractForLoop(_body);
      return false;
      }

   // Reset the file path, name, and extension variables back
   atom = holdFI->Value();
   fi->Value(atom);
   atom = holdFD->Value();
   fd->Value(atom);
   atom = holdFP->Value();
   fp->Value(atom);
   atom = holdFN->Value();
   fn->Value(atom);

   return true;
   } // forEachFile

//---------------------------------------------------------------------
// Loop for each input from the console until the user enters an empty
// line. This is not very good human engineering, I know, but I don't
// want to rewrite the CConsole.ReadString() code. If the ReadString()
// method could return an ok/cancel value, that would be very good. Ok,
// how about a follow-up test API?
//
bool CXText::Resolve::forEachGet(Pchar        _name,
                                 Pchar        _exp,
                                 PCStrList    _body,
                                 PCStrList    _output,
                                 forEachVars& _vars) {
   if (!CConsole::TheConsole)
      new CConsole();

   // Obtain the contents of the getColumns variable
   char columns[sizeString];
   toString(Pchar("getColumns"), columns, sizeString);

   for (int i = 1; ;i++) {
      char prompt[sizeString];
      Strcpy(prompt, _exp, sizeString);
      PCStrList temp = resolveString(prompt, true);

      if (temp) {
         PCVar var = new CVar();
         PCAtom atom = TYPECAST(temp, Atom);
         var->Value(atom);
         toString(var, prompt, sizeString);
         }

      ChrCat(prompt, ' ', sizeString);

      CConsole::TheConsole->WriteSubStr(prompt);
      CConsole::TheConsole->FlushInput();
      CConsole::TheConsole->ReadString(prompt, sizeString);
      while (prompt[strlen(prompt)-1] == '\r' ||
             prompt[strlen(prompt)-1] == '\n')
         prompt[strlen(prompt)-1] = 0;
      if (StrEmpty(prompt)) {
         _body->GotoItem(_vars.iTop);
         extractForLoop(_body);
         return false;
         }

      // Convert the text line into a set of named variables.
      toVars(columns, prompt);

      _vars.position->Value(i);

      if (prompt[0] != '<') {
         // Set the foreach variable to the current input
         _vars.foreach->Value(prompt);
         }
      else {
         // Convert the text into a DOM document
         PCBlock blk    = new CBlock(prompt);
         PCDocument doc = new CDocument();
         doc->PreserveWhiteSpace(true);
         if (!doc->Load(blk)) {
            sprintf(message, Pchar("Could not parse forEach get document %s"),
                             prompt);
            throw(message);
            }

         // Set the foreach variable to the document
         PCAtom atom = TYPECAST(doc, Atom);
         _vars.foreach->Value(atom);
         }

      // Restore the contextNode variable
      PCAtom atom = _vars.holdNd->Value();
      _vars.ctxtnode->Value(atom);

      if (!doLoop(_vars.iTop, _body, _output))
         break;
      if (!condition.IsMark())
         THROW(Pchar("Stack corrupted in forEach get command"))
      }

   return true;
   } // forEachGet

//---------------------------------------------------------------------
// Loop for each line in the text file.
//
bool CXText::Resolve::forEachLine(Pchar        _name,
                                  Pchar        _exp,
                                  PCStrList    _body,
                                  PCStrList    _output,
                                  forEachVars& _vars) {
   // Obtain the contents of the columns variable
   char columns[sizeString];
   toString(Pchar("lineColumns"), columns, sizeString);

   // Ignore blank lines if forEach sub-command is Pchar("lineb"), or if
   // the -U flag is set
   bool ignoreBlanks = stricmp(_name, Pchar("LINEB")) == 0 ||
                       CArgs::TheArgs->FArg(ftBlankFlag);

   char fileName[sizeString];
   toString(_vars.var, fileName, sizeString);
   Trim(fileName);

   PCFile input;
   if (!StrEmpty(fileName)) {
      toString(Pchar("tagsPath"), tagsPath, sizeString);
      if (!FindFileInPath(fileName, tagsPath, fileName, sizeString)) {
         sprintf(message, Pchar("Could not find file <%s> "
                                "for forEach line commad"),
                          fileName);
         throw(message);
         }

      input = new CFile(fileName);
      sprintf(message, Pchar("Could not open file <%s> for forEach line"),
                    fileName);

      if (input->Error())
         throw(message);
      if (!input->OpenInput())
         throw(message);
      }

   // TBD: If input has not been initialized, then input is taken
   // from stdin
   PCTXTFile reader = new CTXTFile(input);

   int pos = 0;
   // Perform the forEach loop
   for (;;) {
      _vars.last->Value(-1);         // last is not valid for this loop

      // Load the text line from the input filelist
      inpStack->Flush();
      PCBlock blk = reader->GetLine();
      if (!blk)
         break;

//?      if (CArgs::TheArgs->FArg(ftDebugFlag))
//?         blk->Save(Pchar("nextline.dbg"));

      Pchar p = blk->Lock();
      char name[sizeString];
      Strcpy(name, p, sizeString);
      blk->Unlock();

      // (a blank line is a line containing either nothing or only
      // characters less than or equal to spaces)
      Trim(name);
      if (strlen(name) == 0)
         if (ignoreBlanks)
            continue;

      // If the grep variable is not empty, match it against the line
      if (!StrEmpty(_vars.grepStr) && match(name, _vars.grepStr) == 0)
          continue;

      // Set the foreach variable to the line
      PCAtom atom = TYPECAST(blk, Atom);
      _vars.foreach->Value(atom);

      // If the columns string is not empty, convert the text into a
      // set of named variables.
      if (!StrEmpty(columns)) {
         p = blk->Lock();
         toVars(columns, p);
         blk->Unlock();
         }

      pos++;
      _vars.position->Value(pos);

      // Restore the contextNode variable
      atom = _vars.holdNd->Value();
      _vars.ctxtnode->Value(atom);

      // Clear the grep variable
      if (_vars.grep)
         setVar(_vars.grep, Pchar(""));

      if (!doLoop(_vars.iTop, _body, _output))
         break;
      if (!condition.IsMark())
         THROW(Pchar("Stack corrupted in forEach line command"))
      }

   if (pos == 0) {   // Skip the code within the forEach..end structure
      extractForLoop(_body);
      input->Close();
      return false;
      }
   input->Close();

   return true;
   } // forEachLine

//---------------------------------------------------------------------
// Loop for each node in the nodelist.
//
bool CXText::Resolve::forEachNode(Pchar        _name,
                                  Pchar        _exp,
                                  PCStrList    _body,
                                  PCStrList    _output,
                                  forEachVars& _vars) {
   if (!_vars.value->Isa(iNodeList))
      THROW(Pchar("forEach-node argument is not a nodelist"))

   PCNodeList myNodes = TYPECAST(_vars.value, NodeList);
   if (myNodes->Length()) {
      // Perform the forEach loop
      ulong len = myNodes->Length();
      for (ulong iNode = 0; iNode < len; iNode++) {
         _vars.last->Value(len);
         _vars.position->Value(iNode+1);

         // Set the foreach variable to the current node
         PCNode nd = myNodes->Item(iNode);
         PCAtom atom = TYPECAST(nd, Atom);
         _vars.foreach->Value(atom);

         // Restore the contextNode variable
         atom = _vars.holdNd->Value();
         _vars.ctxtnode->Value(atom);

         if (!doLoop(_vars.iTop, _body, _output))
            break;
         if (!condition.IsMark())
            THROW(Pchar("Stack corrupted in forEach node command"))
         }
      }
   else {       // Skip the body code within the forEach..end structure
      extractForLoop(_body);
      return false;
      }

   return true;
   } // forEachNode

//---------------------------------------------------------------------
// Loop for each topic in the ODB result list. Form is
//    #forEach odb(|query|type)
//
bool CXText::Resolve::forEachODB(Pchar        _name,
                                 Pchar        _exp,
                                 PCStrList    _body,
                                 PCStrList    _output,
                                 forEachVars& _vars) {
#ifdef FV_ODB
   char query[sizeString];
   toString(_vars.var, query, sizeString);
   Trim(query);

   PCRecord rec = new CRecord(query);

   rec->GetField(1, query, sizeString);
   Trim(query);

   char keyType[sizeString];
   rec->GetField(2, keyType, sizeString);
   Trim(keyType);
   CapStr(keyType);

   // Process the key type, if present
  ODB_LINKTYPE type;
   if (!StrEmpty(keyType)) {
      switch (keyType[0]) {          // Use only the first character
         case 'N':
            type = ODB_NAME;
            break;
         case 'P':
            type = ODB_PARENT;
            break;
         case 'C':
            type = ODB_CHILD;
            break;
         case 'R':
            type = ODB_REF;
            break;
         case 'Q':
            type = ODB_QUERY;
         default:
            break;
         }
      }

   // If no ODB object, then open the ODB specified by the name in
   // ?ODBPATH
   char mssg[sizeString];
   bool odbLocal = false;
   if (!odb) {
      toString(Pchar("ODBPATH"), mssg, sizeString);
      if (StrEmpty(mssg))
         THROW(Pchar("ODB path in ?ODBPATH variable is ")
               Pchar("missing in forEach ODB command"))
      odbLocal = true;
      odb = new CODB(mssg);
      }

   if (!odb->Query(query, type))
      log(true, Pchar("XText: ")
                Pchar("Empty result set for query <%s> ")
                 Pchar("of type <%s> in forEach ODB command"),
          query,
          keyType);

   ulong len = odb->Length();
   if (len > 0)
      for (int i = 1; i <= len; i++) {
         _vars.last->Value(len);
         _vars.position->Value(i);
         PCBlock blk = odb->Topic(i-1);

         PCAtom atom  = TYPECAST(blk, Atom);

/*? if you want a dom object, use #setasxml
         // If text is XML or HTML, convert to a DOM
         TXT_Type type = CTXTType().Type(blk);
         if (type == TXT_HTMLText) {
            HTMLCleaner().FormHTML(blk);
            type = TXT_XMLText;
            }
//?         blk->Save(Pchar("odb.lst"));
         if (type == TXT_XMLText) {
            PCDocument doc = new CDocument();
            doc->PreserveWhiteSpace(true);
            if (doc->Load(blk))
               atom = TYPECAST(doc, Atom);
            else
               THROW(Pchar("Value not valid XML"))
            }
?*/

         _vars.foreach->Value(atom);

         // Restore the contextNode variable
         atom = _vars.holdNd->Value();
         _vars.ctxtnode->Value(atom);

         if (!doLoop(_vars.iTop, _body, _output))
            break;
         if (!condition.IsMark())
            THROW(Pchar("Stack corrupted in forEach ODB command"))
         }
   else {       // Skip the body code within the forEach..end structure
      extractForLoop(_body);

      // Close the odb if it is local
      if (odbLocal) {
         odb->Close();
         odb.Release();
         }

      return false;
      }

   // Close the odb if it is local
   if (odbLocal) {
      odb->Close();
      odb.Release();
      }
#endif // FV_ODB

   return true;
   } // forEachODB

//---------------------------------------------------------------------
// Loop for each row in the SQL select result set.
//
bool CXText::Resolve::forEachSQL(Pchar        _name,
                                 Pchar        _exp,
                                 PCStrList    _body,
                                 PCStrList    _output,
                                 forEachVars& _vars) {
   // Obtain the contents of the SQLColumns variable
   char savcols[sizeString];
   char columns[sizeString];
   toString(Pchar("SQLColumns"), savcols, sizeString);
   strcpy(columns, savcols);

   // Get the query and set up the bindery
   char query[sizeString];
   toString(_vars.var, query, sizeString);
   Trim(query);

   // Get the DSN
   char dsn[sizeString];
   toString(Pchar("dsn"), dsn, sizeString);

   // Get the user name
   char userName[sizeString];
   toString(Pchar("userName"), userName, sizeString);

   // Get the password
   char password[sizeString];
   toString(Pchar("password"), password, sizeString);

   // Connect to the database
   ODBCInterface db;
   if (!db.Open(dsn, userName, password)) {
      sprintf(message, Pchar("forEach-SQL could not connect to DB with %s"), dsn);
      ::OutputDebugString(message);
      throw(message);
      }

   // Perform the query
   int count = db.Query(query);        // returns the number of columns
   if (count <= 0) {
      sprintf(message, Pchar("forEach-SQL query failed with <%s>(%d)"), query, count);
      ::OutputDebugString(message);
      throw(message);
      }

   // Get the column names from the DB if available
   if (db.BindColumns(count, columns, sizeString))
      varHandler->Variable(Pchar("SQLColumns"), columns);

   // Perform the forEach loop
   _vars.last->Value(count);
   int pos = 0;
   PCBlock blk = new CBlock();
   for (;;) {

      // Get a row from the result set
      blk->Copy(Pchar(""), 0);
      if (!getNextRow(db, blk, columns))
         break;

      // Set the forEach variable to the line
      PCAtom atom = TYPECAST(blk, Atom);
      _vars.foreach->Value(atom);

      pos++;
      _vars.position->Value(pos);

      // Restore the contextNode variable
      atom = _vars.holdNd->Value();
      _vars.ctxtnode->Value(atom);

      if (!doLoop(_vars.iTop, _body, _output))
         break;

      // Clear the value of the column variables
      PCRecord rec = new CRecord(columns);
      for (int i = 1; i <= rec->Number(); i++) {
         char varName[sizeLine];
         rec->GetField(i, varName, sizeLine);
         if (StrEmpty(varName))
            continue;
         PCVar var = varHandler->Variable(varName);
         if (var)
            var->Clear();
         }

      if (!condition.IsMark())
         THROW(Pchar("Stack corrupted in forEach SQL command"))
      }

   // Clear the column variables (vars.h says not to remove them?!)
   PCRecord rec = new CRecord(columns);
   for (int i = 1; i <= rec->Number(); i++) {
      char varName[sizeLine];
      rec->GetField(i, varName, sizeLine);
      if (StrEmpty(varName))
         continue;
      PCVar var = varHandler->Variable(varName);
      if (var)
         var->Clear();
      }

   varHandler->Variable(Pchar("SQLColumns"), savcols);

   if (pos == 0) {   // Skip the code within the forEach..end structure
      extractForLoop(_body);
      db.Close();
      return false;
      }

   db.Close();

   return true;
   } // forEachSQL

//---------------------------------------------------------------------
// Loop for each string in the strlist(tbd: /block).
//
bool CXText::Resolve::forEachStr(Pchar        _name,
                                 Pchar        _exp,
                                 PCStrList    _body,
                                 PCStrList    _output,
                                 forEachVars& _vars) {
   // Obtain the contents of the columns variable
   char columns[sizeString];
   toString(Pchar("columns"), columns, sizeString);

   // Ignore blank lines if forEach sub-command is Pchar("strb"), or if
   // the -U flag is set
   bool ignoreBlanks = stricmp(_name, Pchar("STRB")) == 0 ||
                       CArgs::TheArgs->FArg(ftBlankFlag);

   char varName[sizeString];
   toString(_vars.var, varName, sizeString);
   Trim(varName);
   _vars.var = varHandler->Variable(varName);

   if (!_vars.var) {
      THROW(Pchar("forEach-str argument is not a string list"))
      }

   // Convert a block-type variable to a strlist-type variable
   if (_vars.var->Value()->Isa(iBlock)) {
      PCAtom atm    = _vars.var->Value();
      PCStrList lst = new CStrList();
      lst->BlockTo(atm);
      atm = TYPECAST(lst, Atom);
      _vars.var->Value(atm);
      }

   _vars.value = _vars.var->Value();

   if (!_vars.value->Isa(iStrList)) { // && !value->Isa(iBlock))
      THROW(Pchar("forEach-str argument is not a string list"))
      }

   PCStrList myList = TYPECAST(_vars.value, StrList);
   if (myList->Number()) {
      // Perform the forEach loop
      ulong len = myList->Number();
      for (ulong iStr = 1; iStr <= len; iStr++) {
         _vars.last->Value(len);
         _vars.position->Value(iStr);

         // Get the line and check if blank
         char str[sizeString];
         myList->StrItem(iStr, str, sizeString);
         Trim(str);
         int len = strlen(str);
         if (len == 0)
            if (ignoreBlanks)
               continue;

         // If the grep variable is not empty, match it against the
         // string
         if (!StrEmpty(_vars.grepStr) && match(str, _vars.grepStr) == 0)
             continue;

         // Set the foreach variable to the line
         PCBlock blk = new CBlock(str);
         if (!blk)
            break;

         PCAtom atom = TYPECAST(blk, Atom);
         _vars.foreach->Value(atom);

         // If the columns string is not empty, convert the text into a
         // set of named variables.
         if (!StrEmpty(columns)) {
            Pchar p = blk->Lock();
            toVars(columns, p);
            blk->Unlock();
            }

         // Clear the grep variable
         if (_vars.grep)
            setVar(_vars.grep, Pchar(""));

         if (!doLoop(_vars.iTop, _body, _output))
            break;
         if (!condition.IsMark())
            THROW(Pchar("Stack corrupted in forEach str command"))
         }
      }
   else {       // Skip the body code within the forEach..end structure
      extractForLoop(_body);
      return false;
      }

   return true;
   } // forEachStr

//---------------------------------------------------------------------
// Loop for each word in the string.
//
bool CXText::Resolve::forEachWord(Pchar        _name,
                                  Pchar        _exp,
                                  PCStrList    _body,
                                  PCStrList    _output,
                                  forEachVars& _vars) {
   char str[sizeString];
   toString(_vars.var, str, sizeString);
   Trim(str, bothEnds);
   if (!StrEmpty(str)) {
      // Perform the forEach loop
      PCRecord words = new CRecord(str, Pchar(" "));
      words->TrimAll();
      char word[sizeString];

      // Remove empty words and delimiters
      int i = 1;
      while (i <= words->Number())
         if (words->FieldEmpty(i))
            words->DeleteField(i);
         else {
            words->GetField(i, word, sizeString);
            int j = strlen(word)-1;
            bool done = false;
            while (!done) {
               if      (j == 0)
                  done = true;
               else if (IsDelim(word[j]))
                  word[j] = 0;
               else
                  done = true;
               j--;
               }
            words->PutField(i, word);
            i++;
            }

      ulong nr = words->Number();
      for (uint i = 1; i <= nr; i++) {
         _vars.last->Value(nr);
         _vars.position->Value(i);

         // Set the foreach variable to the current word
         words->GetField(i, word, sizeString);

         // If the grep string is not empty, match it against the word
         if (!StrEmpty(_vars.grepStr) && match(word, _vars.grepStr) == 0)
             continue;

         PCAtom blk = new CBlock(word);
         _vars.foreach->Value(blk);

         PCAtom atom;

         // Restore the contextNode variable
         atom = _vars.holdNd->Value();
         _vars.ctxtnode->Value(atom);

         // Clear the grep variable
         if (_vars.grep)
            setVar(_vars.grep, Pchar(""));

         if (!doLoop(_vars.iTop, _body, _output))
            break;
         if (!condition.IsMark())
            THROW(Pchar("Stack corrupted in forEach field command"))
         }
      }
   else {       // Skip the body code within the forEach..end structure
      extractForLoop(_body);
      return false;
      }

   return true;
   } // forEachWord

//---------------------------------------------------------------------
// Loop for each <element>..</element> object in the input file.
//
bool CXText::Resolve::forEachXML(Pchar        _name,
                                 Pchar        _exp,
                                 PCStrList    _body,
                                 PCStrList    _output,
                                 forEachVars& _vars) {
   char fileName[sizeString];
   toString(_vars.var, fileName, sizeString);
   Trim(fileName);

   toString(Pchar("tagsPath"), tagsPath, sizeString);
   if (!FindFileInPath(fileName, tagsPath, fileName, sizeString)) {
      sprintf(message, Pchar("Could not find forEach XML file <%s>"), fileName);
      throw(message);
      }

   TXT_Type type = CTXTType().Type(fileName, Pchar(""));
   if (type != TXT_XMLText) {
      sprintf(message, Pchar("forEach XML content in file %s is not XML"),
                       fileName);
      throw(message);
      }

   PCFile input = new CFile(fileName);
   sprintf(message, Pchar("Could not open forEach XML file <%s>"), fileName);

   if (input->Error())
      throw(message);
   if (!input->OpenInput())
      throw(message);

   // Perform the forEach loop
   int pos = 0;
   CXMLFile reader(input);
   for (;;) {
      _vars.last->Value(-1);         // last is not valid for this loop
      // Load the next element in the input file list
      inpStack->Flush();
      PCBlock blk = reader.GetXML();
      if (!blk)
         break;
      if (blk->GetSize() == 0)
         continue;

//?      if (CArgs::TheArgs->FArg(ftDebugFlag))
//?         blk->Save(Pchar("foreach.dbg"));

      // Convert the XML text into a DOM document
      PCDocument doc = new CDocument();
      doc->PreserveWhiteSpace(true);
      if (!doc->Load(blk)) {
         sprintf(message, Pchar("Could not parse forEach XML element "
                                "%d in file %s"),
                          pos,
                          fileName);
         throw(message);
         }

      // Set the foreach variable to the document
      PCAtom atom = TYPECAST(doc, Atom);
      _vars.foreach->Value(atom);

      pos++;
      _vars.position->Value(pos);

      // Restore the contextNode variable
      atom = _vars.holdNd->Value();
      _vars.ctxtnode->Value(atom);

      if (!doLoop(_vars.iTop, _body, _output))
         break;
      if (!condition.IsMark())
         THROW(Pchar("Stack corrupted in forEach-XML command"))
      }

   if (pos == 0) {   // Skip the code within the forEach..end structure
      extractForLoop(_body);
      input->Close();
      return false;
      }
   input->Close();

   return true;
   } // forEachXML

//---------------------------------------------------------------------
// Loop for each line in the object. Convert the object to a strList if
// it isn't already.
//
bool CXText::Resolve::forEachDefault(Pchar        _name,
                                     Pchar        _exp,
                                     PCStrList    _body,
                                     PCStrList    _output,
                                     forEachVars& _vars) {
   PCStrList myList;

   // Obtain the contents of the columns variable
   char columns[sizeString];
   toString(Pchar("columns"), columns, sizeString);

   // Convert the first node of a node list to a string list
   if      (_vars.value->Isa(iNodeList)) {
      PCNodeList myNodes = TYPECAST(_vars.value, NodeList);
      if (myNodes->Length()) {
         PCNode nd = myNodes->Item(0);
         if (nd->Isa(iAttr) ||
             nd->Isa(iText)) {
            Pchar p = nd->NodeValue();
            PCAtom atom = new CBlock(p);
            myList = new CStrList();
            myList->BlockTo(atom);
            delete[] p;
            }
         else {                  // serialize the node and its children
            myList = new CStrList();
            nd->Save(myList);
            return true;
            }
         }
      else
         myList = new CStrList();
      }

   // Convert the node to a string list
   else if (_vars.value->Isa(iNode)) {
      PCNode nd = TYPECAST(_vars.value, Node);
      if (nd->Isa(iAttr) ||
          nd->Isa(iText)) {
         Pchar p = nd->NodeValue();
         PCAtom atom = new CBlock(p);
         delete[] p;
         myList = new CStrList();
         myList->BlockTo(atom);
         }
      else {                     // serialize the node and its children
         myList = new CStrList();
         nd->Save(myList);
         return true;
         }
      }

   // Convert a block to a string list
   else if (_vars.value->Isa(iBlock)) {
      PCAtom blk = TYPECAST(_vars.value, Block);
      myList = new CStrList();
      myList->BlockTo(blk);
      }

   // Already a string list
   else if (_vars.value->Isa(iStrList))
      myList = TYPECAST(_vars.value, StrList);

   // Don't know about this one
   else
      return false;

   // Perform the forEach loop
   ulong nr = myList->Number();
   if (myList->GotoHead()) do {
      _vars.last->Value(nr);
      _vars.position->Value(myList->Where());

      // Set the foreach variable to the current line
      Pchar p = myList->StrItem(false);

      // If the grep variable is not empty, match it against the
      // string
      if (!StrEmpty(_vars.grepStr) && match(p, _vars.grepStr) == 0)
          continue;

      PCAtom blk = new CBlock(p);
      _vars.foreach->Value(blk);

      // Convert the text line into a set of named variables.
      toVars(columns, p);

      // Restore the contextNode variable
      PCAtom atom = _vars.holdNd->Value();
      _vars.ctxtnode->Value(atom);

      // Clear the grep variable
      if (_vars.grep)
         setVar(_vars.grep, Pchar(""));

      if (!doLoop(_vars.iTop, _body, _output))
         break;
      if (!condition.IsMark())
         THROW(Pchar("Stack corrupted in forEach command"))
      } while (myList->GotoNext());
   else {       // Skip the body code within the forEach..end structure
      extractForLoop(_body);
      return false;
      }

   return true;
   } // forEachDefault

//---------------------------------------------------------------------
//
void CXText::Resolve::cForEach(Pchar     _name,
                               Pchar     _exp,
                               PCStrList _body,
                               PCStrList _output) {
   dbgLog(commandLog, Pchar("forEach %s (%s)"), _name, _exp);

   if (condition.IsFalse()) {
      condition.Push(HardFalse);
      return;
      }

   char temp[sizeString];
   PCAtom atom;

   forEachVars vars;

   // Save the contents of the contextNode variable
   vars.ctxtnode = varHandler->Variable(Pchar("contextNode"));
   Assert(vars.ctxtnode);
   atom = vars.ctxtnode->Value();
   vars.holdNd = new CVar();
   vars.holdNd->Value(atom);

   // Save the contents of the foreach variable
   vars.foreach = varHandler->Variable(Pchar("forEach"));
   Assert(vars.foreach);
   atom = vars.foreach->Value();
   vars.holdFE = new CVar();
   vars.holdFE->Value(atom);

   // Save the contents of the last variable
   vars.last = varHandler->Variable(Pchar("last"));
   Assert(vars.last);
   atom = vars.last->Value();
   vars.holdLT = new CVar();
   vars.holdLT->Value(atom);

   // Save the contents of the position variable
   vars.position = varHandler->Variable(Pchar("position"));
   Assert(vars.position);
   atom = vars.position->Value();
   vars.holdPS = new CVar();
   vars.holdPS->Value(atom);

   // Save the contents of the grep variable
   vars.grep = varHandler->Variable(Pchar("grep"));
   Assert(vars.grep);
   atom = vars.grep->Value();
   vars.holdGP = new CVar();
   vars.holdGP->Value(atom);
   toString(vars.grep, temp, sizeString);
   if (strlen(temp)) {
      vars.grepStr = new char[strlen(temp)+1];
      strcpy(vars.grepStr, temp);
      }

   // Remember where we are in the body list
   vars.iTop = _body->Where();

   // Mark the current level on the condition stack
   condition.Push(Mark);

   // Resolve the name
   Pchar name = new char[strlen(_name)+1];
   strcpy(name, _name);

   // Resolve name
   if (StrEmpty(name))
      strcpy(message, name);
   else {
      PCStrList list = resolveString(name, true);
      while (list) {
         if (list->Number() == 1) {
            list->GotoHead();
            list->StrItem(name, sizeString);
            }
         list = resolveString(name, true);
         }
      }

   // Resolve the argument
   vars.var = resolveArg(_exp, false);
   if (!vars.var) {
      sprintf(message, Pchar("Could not resolve argument <%s> in forEach command"), _exp);
      throw(message);
      }

   vars.value = vars.var->Value();

   //------------------------------------------------------------------
   // The following section contains the individual loops for each type
   // of object upon which the forEach command can loop.
   //
   //  1. Char      - loops through characters in a string.
   //  2. Count     - loops a specified number of times.
   //  3. Field+    - loops through fields in a delimited-string record.
   //  4. File      - loops through file names listed in a directory.
   //  5. Get*      - loops through user input until the user terminates.
   //  6. Line*+    - loops through lines in a text file.
   //  7. Lineb*+   - loops through lines in a text file ignoring blank
   //                 lines.
   //  8. Node      - loops through nodes in a node list.
   //  9. ODB*      - loops through an ODB result set.
   // 10. SQL*      - loops through an ODBC result set.
   // 11. Str*+     - loops through lines in a string list.
   // 12. Strb*+    - loops through lines in a string list ignoring blank
   //                 lines.
   // 13. Word      - loops through words in a string.    
   // 10. XML       - loops through xml documents in a document list
   //                 file.
   // 11. default*+ - loops through a list of lines in a node list,
   //                 a node, a strlist, or a block.
   //
   // * For the marked forEach modifiers, Tags recognizes two variables
   //   that can contain instructions about how to parse the contents
   //   of the ?forEach$ variable into a specified set of variables.
   //
   // + If the grep variable is not empty, forEach uses it to match
   //   against each string in the list and propogates only matching
   //   strings. The grep variable is taken as a regular expression.
   //
   //   If a result from the ODB or SQL query is an XML document, it is
   //   automatically loaded into a DOM.
   //
   // columns     - Get, Line and Lineb, StrList, Block
   // sqlcolumns  - SQL
   //
   //------------------------------------------------------------------

   bool didLoop = false;

   //------------------------------------------------------------------
   // cForEach char: loop for each character in the string in value
   //
   for (;;) {
   if (stricmp(name, Pchar("CHAR")) == 0) {
      didLoop = forEachChar(name,
                            _exp,
                            _body,
                            _output,
                            vars);
      break;
      } // end of forEach char()

   //------------------------------------------------------------------
   // cForEach count: loop for each number between 1 and value
   //
   if (stricmp(name, Pchar("COUNT")) == 0) {
      didLoop = forEachCount(name,
                             _exp,
                             _body,
                             _output,
                             vars);
      break;
      } // end of forEach count()

   //------------------------------------------------------------------
   // cForEach field: loop for each field in the record in value
   //
   if (stricmp(name, Pchar("FIELD"))  == 0) {
      didLoop = forEachField(name,
                             _exp,
                             _body,
                             _output,
                             vars);
      break;
      } // end of forEach field()

   //------------------------------------------------------------------
   // cForEach file: loop for each file in a directory. The form of the
   // forEach argument is
   //
   //    dirName|fileMask|searchtype
   // where
   //    searchtype={ *root | tree & *data | dir | any }
   //
   if (stricmp(name, Pchar("FILE")) == 0) {
      didLoop = forEachFile(name,
                            _exp,
                            _body,
                            _output,
                            vars);
      break;
      } // end of forEach file()

   //------------------------------------------------------------------
   // cForEach get: loop for each input from the console until the user
   // enters an empty line. This is not very good human engineering, I
   // know, but I don't want to rewrite the CConsole.ReadString() code.
   // If the ReadString() method could return an ok/cancel value, that
   // would be very good. Ok, how about a follow-up test API?
   //
   if (stricmp(name, Pchar("GET"))   == 0) {
      didLoop = forEachGet(name,
                           _exp,
                           _body,
                           _output,
                           vars);
      break;
      } // end of forEach get()

   //------------------------------------------------------------------
   // cForEach line: loop for each line in the text file.
   //
   if (stricmp(name, Pchar("LINE"))  == 0 ||
       stricmp(name, Pchar("LINEB")) == 0) {
      didLoop = forEachLine(name,
                            _exp,
                            _body,
                            _output,
                            vars);
      break;
      } // end of forEach line()

   //------------------------------------------------------------------
   // cForEach node: loop for each node in the nodelist
   //
   if (stricmp(name, Pchar("NODE")) == 0) {
      didLoop = forEachNode(name,
                            _exp,
                            _body,
                            _output,
                            vars);
      break;
      } // end of forEach node()

   //------------------------------------------------------------------
   // cForEach ODB: loop for each topic in the ODB result list
   //    form is #forEach odb(|query|type)
   //
   if (stricmp(name, Pchar("ODB")) == 0) {
#ifndef FV_ODB
      Assert(false);
#else
      didLoop = forEachODB(name,
                           _exp,
                           _body,
                           _output,
                           vars);
#endif // FV_ODB
      break;
      } // end of forEach odb()

   //------------------------------------------------------------------
   // cForEach SQL: loop for each row in the SQL select result set
   //
   if (stricmp(name, Pchar("SQL")) == 0) {
      didLoop = forEachSQL(name,
                           _exp,
                           _body,
                           _output,
                           vars);
      break;
      } // end of forEach SQL()

   //------------------------------------------------------------------
   // cForEach str: loop for each string in the strlist(tbd: /block)
   //
   if (stricmp(name, Pchar("STR"))  == 0 ||
       stricmp(name, Pchar("STRB")) == 0) {
      didLoop = forEachStr(name,
                           _exp,
                           _body,
                           _output,
                           vars);
      break;
      } // end of forEach str()

   //------------------------------------------------------------------
   // cForEach word: loop for each word in the string in value
   //
   if (stricmp(name, Pchar("WORD")) == 0) {
      didLoop = forEachWord(name,
                            _exp,
                            _body,
                            _output,
                            vars);
      break;
      } // end of forEach word()

   //------------------------------------------------------------------
   // cForEach XML: loop for each <element>..</element> object in the
   // input file.
   //
   if (stricmp(name, Pchar("XML"))  == 0) {
      didLoop = forEachXML(name,
                           _exp,
                           _body,
                           _output,
                           vars);
      break;
      } // end of forEach XML()

   //------------------------------------------------------------------
   // cForEach: loop for each line in the object. Convert the object to
   // a strList if it isn't already.
   //
   if (StrEmpty(name)) {
      didLoop = forEachDefault(name,
                               _exp,
                               _body,
                               _output,
                               vars);
      break;
      } // end of forEach ()

   // Bypass the loop code
   extractForLoop(_body);
   break;
   } // end of for (;;) {

   lastLoop = false;
   nextLoop = false;

   // Check and clear the forEach Mark from the condition stack
   if (!condition.IsMark())
      THROW(Pchar("Stack corrupted in forEach command"))
   condition.Pop();

   // Restore the contextNode variable
   atom = vars.holdNd->Value();
   vars.ctxtnode->Value(atom);

   // Restore the forEach variable
   atom = vars.holdFE->Value();
   vars.foreach->Value(atom);

   // Restore the last variable
   atom = vars.holdLT->Value();
   vars.last->Value(atom);

   // Restore the position variable
   atom = vars.holdPS->Value();
   vars.position->Value(atom);

   // Restore the grep variable
   if (vars.grep)
      setVar(vars.grep, vars.grepStr);

/*?
   if (didLoop)
      condition.Push(True);
   else
      condition.Push(False);
?*/
   } // cForEach

//---------------------------------------------------------------------
// if tos is hardfalse, ignore this forelse.
// if tos is false, set tos to true.
// if tos is true, set tos to hardfalse.
//
void CXText::Resolve::cForElse(void) {
   dbgLog(commandLog, Pchar("forElse"));

   if      (condition.IsHardFalse())
      return;
   else if (condition.IsFalse())
      condition.Set(True);
   else
      condition.Set(HardFalse);
   } // cForElse

//---------------------------------------------------------------------
//
void CXText::Resolve::cWhile(Pchar     _text,
                             Pchar     _exp,
                             PCStrList _body,
                             PCStrList _output) {
   dbgLog(commandLog, Pchar("while %s (%s)"), _text, _exp);

   if (condition.IsFalse())
      return;

   // Remember where we are in the body list
   long iTop = _body->Where();

   // Mark the current level on the condition stack
   condition.Push(Mark);

   if (StrEmpty(_text))
      if (isTrue(_exp))
         while (isTrue(_exp)) {
            if (!doLoop(iTop, _body, _output))
               break;
            if (!condition.IsMark())
               THROW(Pchar("Stack corrupted in while command"));
            }
      else
         extractForLoop(_body);
   else
      if (match(_text, _exp))
         while (match(_text, _exp)) {
            if (!doLoop(iTop, _body, _output))
               break;
            if (!condition.IsMark())
               THROW(Pchar("Stack corrupted in while command"));
            }
      else
         extractForLoop(_body);

   lastLoop = false;
   nextLoop = false;

   // Check and clear the loop Mark from the condition stack
   if (!condition.IsMark())
      THROW(Pchar("Stack corrupted in while command"));
   condition.Pop();
   } // cWhile

//---------------------------------------------------------------------
//
void CXText::Resolve::cWhileN(Pchar     _text,
                              Pchar     _exp,
                              PCStrList _body,
                              PCStrList _output) {
   dbgLog(commandLog, Pchar("whileN %s (%s)"), _text, _exp);

   if (condition.IsFalse())
      return;

   // Remember where we are in the body list
   long iTop = _body->Where();

   // Mark the current level on the condition stack
   condition.Push(Mark);

   if (StrEmpty(_text))
      if (!isTrue(_exp))
         while (!isTrue(_exp)) {
            if (!doLoop(iTop, _body, _output))
               break;
            if (!condition.IsMark())
               THROW(Pchar("Stack corrupted in whileN command"));
            }
      else
         extractForLoop(_body);
   else
      if (!match(_text, _exp))
         while (!match(_text, _exp)) {
            if (!doLoop(iTop, _body, _output))
               break;
            if (!condition.IsMark())
               THROW(Pchar("Stack corrupted in whileN command"));
            }
      else
         extractForLoop(_body);

   lastLoop = false;
   nextLoop = false;

   // Check and clear the loop Mark from the condition stack
   if (!condition.IsMark())
      THROW(Pchar("Stack corrupted in whileN command"));
   condition.Pop();
   } // cWhileN

//---------------------------------------------------------------------
//
void CXText::Resolve::cLast(Pchar _text, Pchar _exp) {
   dbgLog(commandLog, Pchar("last %s (%s)"), _text, _exp);

   if (condition.IsFalse())
      return;
   else if (StrEmpty(_text) &&
            StrEmpty(_exp))
      lastLoop = true;
   else if (StrEmpty(_text)) {
      if (isTrue(_exp))
         lastLoop = true;
      }
   else if (StrEmpty(_exp))
      lastLoop = true;
   else if (match(_text, _exp))
      lastLoop = true;
   } // cLast

//---------------------------------------------------------------------
//
void CXText::Resolve::cLastN(Pchar _text, Pchar _exp) {
   dbgLog(commandLog, Pchar("lastN %s (%s)"), _text, _exp);

   if (condition.IsTrue())
      return;
   else if (StrEmpty(_text) &&
            StrEmpty(_exp))
      lastLoop = false;
   else if (StrEmpty(_text))
      lastLoop = true;
   else if (!match(_text, _exp))
      lastLoop = true;
   } // cLastN

//---------------------------------------------------------------------
//
void CXText::Resolve::cNext(Pchar _text, Pchar _exp) {
   dbgLog(commandLog, Pchar("next %s (%s)"), _text, _exp);

   if (condition.IsFalse())
      return;
   else if (StrEmpty(_text) &&
            StrEmpty(_exp))
      nextLoop = true;
   else if (StrEmpty(_text)) {
      if (isTrue(_exp))
         nextLoop = true;
      }
   else if (StrEmpty(_exp))
      nextLoop = true;
   else if (match(_text, _exp))
      nextLoop = true;
   } // cNext

//---------------------------------------------------------------------
//
void CXText::Resolve::cNextN(Pchar _text, Pchar _exp) {
   dbgLog(commandLog, Pchar("nextN %s (%s)"), _text, _exp);

   if (condition.IsFalse())
      return;
   else if (StrEmpty(_text) &&
            StrEmpty(_exp))
      nextLoop = false;
   else if (StrEmpty(_text))
      nextLoop = true;
   else if (!match(_text, _exp))
      nextLoop = true;
   } // cNextN

//---------------------------------------------------------------------
// Convert the named variable to a CStrList. If the expression is a
// nodelist, then insert the string values of the nodes in the list into
// the named variable. If the expression is a node, then insert the
// string value of the node into the named variable. If the expression is
// a CStrList, then insert the list into the named variable. If the
// expression is a string, then insert the string into the named variable.
// If the expression is empty, then don't do anything. Note that XText
// supports several insertion methods: atTop, atEnd, sorted, and unique.
// Use $#addT, $#add, $#addS, and $#addU, respectively.
//
void CXText::Resolve::cAdd(Pchar _name, Pchar _exp, AtWhere _where) {
   dbgLog(commandLog, Pchar("add %s (%s)"), _name, _exp);

   if (condition.IsFalse())
      return;

   char name[sizeString];
   Strcpy(name, _name, sizeString);

   PCStrList list = resolveString(name, true);
   while (list) {
      if (list->Number() == 1) {
         list->GotoHead();
         list->StrItem(name, sizeString);
         }
      list = resolveString(name, true);
      }
   if (StrEmpty(name))
      THROW(Pchar("Invalid variable name in Add command"));

   // Get the destination variable
   PCVar var = varHandler->Variable(name);
   PCAtom atm;
   if (!var) {
      atm = new CBlock();
      var = varHandler->Variable(name, atm);
      }
   else {
      atm = var->Value();

      if (!atm) {
         THROW(Pchar("Add container must be a block"));
         }

      if (atm->Isa(iStrList)) {
         PCStrList list = TYPECAST(atm, StrList);
         PCAtom blk = new CBlock();
         list->ToBlock(blk);
         atm = blk;
         }
      }

   // If no input value, then nothing to add
   if (StrEmpty(_exp))
      return;

   PCVar temp = resolveArg(_exp, false);
   PCAtom tatm = temp->Value();

   // Note: Only supports a string list container

   // Append the expression value to the container variable
   if      (tatm->Isa(iNode)) {
      PCNode node = TYPECAST(tatm, Node);
      if (node->NodeType() == ATTRIBUTE_NODE ||
          node->NodeType() == TEXT_NODE) {

         PCStrList lista = new CStrList();
         lista->BlockTo(atm);

         PCStrList listb = new CStrList();
         Pchar p = node->NodeValue();
         PCAtom blk      = new CBlock(p);
         delete[] p;
         listb->BlockTo(blk);

         PCStrList listc = new CStrList();
         Resolve* r = new Resolve();
         r->resolve(listb, listc);
         delete r;

         lista->ListInsert(listc, _where);
         atm = new CBlock();
         lista->ToBlock(atm);
         var->Value(atm);
         }
      else {
         sprintf(message, Pchar("Cannot add type = %d in  %s(%s)"), tatm->IAm(), name, _exp);
         throw(message);
         }
      }
   else if (tatm->Isa(iNodeList)) {
      PCStrList lista = new CStrList();
      lista->BlockTo(atm);

      PCNodeList list = TYPECAST(tatm, NodeList);
      for (uint i = 0; i < list->Length(); i++) {
         PCNode node = list->Item(i);
         if (node->NodeType() == ATTRIBUTE_NODE ||
             node->NodeType() == TEXT_NODE) {

            PCStrList listb = new CStrList();
            Pchar p = node->NodeValue();
            PCAtom blk      = new CBlock(p);
//?         blk->Save(Pchar("blk.txt"));
            delete[] p;
            listb->BlockTo(blk);
//?         listb->Save(Pchar("listb.txt"));

            PCStrList listc = new CStrList();
            Resolve* r = new Resolve();
            r->resolve(listb, listc);
            delete r;
//?         listc->Save(Pchar("listc.txt"));
//?         lista->Save(Pchar("listapre.txt"));
            lista->ListInsert(listc, _where);
//?         lista->Save(Pchar("listapost.txt"));
            var->Value(atm);
            }
         else {
            sprintf(message, Pchar("Cannot add type = %d in  %s(%s)"), tatm->IAm(), name, _exp);
            throw(message);
            }
         }
      atm = new CBlock();
      lista->ToBlock(atm);
      var->Value(atm);
      }
   else if (tatm->Isa(iStrList)) {

      PCStrList lista = new CStrList();
      lista->BlockTo(atm);

      PCStrList listb = TYPECAST(tatm, StrList);
      PCStrList listc = new CStrList();
      Resolve* r = new Resolve();
      r->resolve(listb, listc);
      delete r;

      lista->ListInsert(listc, _where);
      atm = new CBlock();
      lista->ToBlock(atm);
      var->Value(atm);
      }
   else if (tatm->Isa(iBlock)) {

      PCStrList lista = new CStrList();
      lista->BlockTo(atm);

      PCStrList listb = new CStrList();
      listb->BlockTo(tatm);

      PCStrList listc = new CStrList();
      Resolve* r = new Resolve();
      r->resolve(listb, listc);
      delete r;

      lista->ListInsert(listc, _where);
      atm = new CBlock();
      lista->ToBlock(atm);
      var->Value(atm);
      }
   else {
      sprintf(message, Pchar("Cannot add type = %d in  %s(%s)"), tatm->IAm(), name, _exp);
      throw(message);
      }
   } // cAdd

//---------------------------------------------------------------------
//
void CXText::Resolve::cConsole(Pchar _arg) {
   dbgLog(commandLog, Pchar("console (%s)"), _arg);

   if (condition.IsFalse())
      return;

   PCVar var = resolveArg(_arg, true);

   char line[sizeString];
   toString(var, line, sizeString);

   Trim(line);
   Squeeze(line, '0');

   if (isFalse(line)) {
      if (CConsole::TheConsole)
         CConsole::TheConsole.Release();
      }
   else
      if (!CConsole::TheConsole)
         new CConsole();
   } // cConsole

//---------------------------------------------------------------------
//
void CXText::Resolve::cDebug(Pchar _arg) {
   dbgLog(commandLog, Pchar("debug (%s)"), _arg);

   if (condition.IsFalse())
      return;

   PCVar var = resolveArg(_arg, true);

   char line[sizeString];
   toString(var, line, sizeString);

   Trim(line);
   Squeeze(line, '0');

   if (isFalse(line))
      CArgs::TheArgs->ClrFArg(ftDebugFlag);
   else
      CArgs::TheArgs->SetFArg(ftDebugFlag);
   } // cDebug

//---------------------------------------------------------------------
//
void CXText::Resolve::cDefer(Pchar _name, Pchar _exp) {
   dbgLog(commandLog, Pchar("defer %s (%s)"), _name, _exp);

   if (condition.IsFalse())
      return;

   char name[sizeString];
   Strcpy(name, _name, sizeString);

   PCStrList list = resolveString(name, true);
   while (list) {
      if (list->Number() == 1) {
         list->GotoHead();
         list->StrItem(name, sizeString);
         }
      list = resolveString(name, true);
      }
   if (StrEmpty(name))
      THROW(Pchar("Invalid variable name in AsIs command"));

   // Initialize a variable for the specified name
   varHandler->Variable(name, _exp);
   } // cDefer

//---------------------------------------------------------------------
// If the specified atom is a node other than a document, return its
// owner document, otherwise return zero. Protected.
//
PCDocument CXText::Resolve::getDoc(PCAtom _atom) {
   if (!_atom)
      return 0;
   PCAtom atom = _atom;
   if (atom->Isa(iNodeList)) {
      PCNodeList nodes = TYPECAST(atom, NodeList);
      if (nodes->Length() > 0)
         atom = nodes->Item(0);
      }
   if (!atom)
      return 0;
   if (atom->Isa(iNode)) {
      PCNode node = TYPECAST(atom, Node);
      if (!node->Isa(iDocument)) {
         PCDocument doc = node->OwnerDocument();
         return doc;
         }
      }

   return 0;
   } // getDoc

/*?
//---------------------------------------------------------------------
//
void CXText::Resolve::cDrop(Pchar _name) {
   if (condition.IsFalse())
      return;

   char name[sizeString];
   Strcpy(name, _name, sizeString);

   PCStrList list = resolveString(name, true);
   while (list) {
      if (list->Number() == 1) {
         list->GotoHead();
         list->StrItem(name, sizeString);
         }
      list = resolveString(name, true);
      }
   if (StrEmpty(name))
      THROW(Pchar("Invalid variable name in Set command"));

   PCVar var = varHandler->Variable(name);
   if (!var)
      return;

   PCDocument doc = getDoc(var->Value());
   if (doc)
      doc->DrgRef();

   varHandler->Remove(); TBD - not sure about the Drop command really.
   } // cDrop
?*/

//---------------------------------------------------------------------
//
void CXText::Resolve::cExec(Pchar _name, Pchar _arg) {
   dbgLog(commandLog, Pchar("exec %s (%s)"), _name, _arg);

   if (condition.IsFalse())
      return;

   char name[sizeString];
   Strcpy(name, _name, sizeString);

   PCStrList list = resolveString(name, true);
   while (list) {
      if (list->Number() == 1) {
         list->GotoHead();
         list->StrItem(name, sizeString);
         }
      list = resolveString(name, true);
      }

   bool wait = true;
   if (isFalse(name))
      wait = false;

   PCVar var = resolveArg(_arg, true);

   char line[sizeString];
   toString(var, line, sizeString);

   Trim(line);
   CProcess launch;
   launch.Run(line, Pchar(""), wait);

   if (wait)
      // Handle return code
      varHandler->Variable(Pchar("error"), launch.Error());
   } // cExec

//---------------------------------------------------------------------
//
void CXText::Resolve::cGet(Pchar _name, Pchar _prompt) {
   dbgLog(commandLog, Pchar("get %s (%s)"), _name, _prompt);

   if (condition.IsFalse())
      return;

   char name[sizeString];
   Strcpy(name, _name, sizeString);

   PCStrList list = resolveString(name, true);
   while (list) {
      if (list->Number() == 1) {
         list->GotoHead();
         list->StrItem(name, sizeString);
         }
      list = resolveString(name, true);
      }
   if (StrEmpty(name))
      THROW(Pchar("Invalid variable name in Get command"));

   PCVar var;

   char prompt[sizeString];
   Strcpy(prompt, _prompt, sizeString);
   PCStrList temp = resolveString(prompt, true);

   if (temp) {
      var = new CVar();
      PCAtom atom = TYPECAST(temp, Atom);
      var->Value(atom);

      toString(var, prompt, sizeString);
      }
   ChrCat(prompt, ' ', sizeString);

   if (!CConsole::TheConsole)
      new CConsole();

   CConsole::TheConsole->WriteSubStr(prompt);
   CConsole::TheConsole->FlushInput();
   CConsole::TheConsole->ReadString(prompt, sizeString);

   // Initialize a variable for the specified name
   Pchar p = prompt;
   while (*p) {
      if (*p < ' ')
         strcpy(p, p+1);
      else
         p++;
      }

   varHandler->Variable(name, prompt);
   } // cGet

//---------------------------------------------------------------------
// Called by cInput and cInputP. Protected.
//
PCStrList CXText::Resolve::getHTTP(Pchar _url, PCVar _post) {
   Pchar post = 0;
   long  plen = 0;
   if (!(!_post)) {
      post = toString(_post);
      if (post)
         plen = strlen(post);
      }

   PCVar vhead = varHandler->Variable(Pchar("HTTPHeaders"));
   Pchar head = toString(vhead);

   PCBlock input = new CBlock();
   if (!HTTPReader().ReadHTTP(_url, head, post, plen, input)) {
      if (plen)
         sprintf(message, Pchar("HTTP Put request failed for %s"), _url);
      else
         sprintf(message, Pchar("HTTP Get request failed for %s"), _url);
      delete[] head;
      delete[] post;
      throw(message);
      }

   PCStrList list = new CStrList();
   PCAtom atom = TYPECAST(input, Atom);
   list->BlockTo(atom);

   delete[] head;
   delete[] post;
   return list;
   } // getHTTP

//---------------------------------------------------------------------
// Called by cInput. Protected.
//
PCStrList CXText::Resolve::getFile(Pchar _name, bool _ignoreBlanks) {
   PCStrList input = new CStrList();

   char path[sizeString];
   path[0] = 0;

   toString(Pchar("tagsPath"), tagsPath, sizeString);
   if (FindFileInPath(_name, tagsPath, path, sizeString)) {
      strcpy(_name, path);
      if (!input->Load(_name)) {
         sprintf(message, Pchar("Could not load the load file %s"), _name);
         throw(message);
         }
      }
/* No file found with the specified name - don't report it, tags script
   should check for an empty result.
   else {
      sprintf(message, Pchar("Could not find the load file %s"), _name);
      throw(message);
      }
*/

   GetFilePath(path, _name, sizeString);

   // Resolve includes contained in the script
   if (input->GotoHead()) do {
      char line[sizeString];
      input->StrItem(line, sizeString);
      Trim(line, frontEnd);
      if (_ignoreBlanks) {
         input->Remove(before);
         continue;
         }
      CapStr(line);
      if (StrPos(ftIncStr, line) == 0) {
         expandInclude(input, path);
         input->GotoPrev();
         }
      } while (input->GotoNext());

   return input;
   } // getFile

//---------------------------------------------------------------------
// Load the specified document into a DOM, unless it is plaintext.
//
void CXText::Resolve::cInput(Pchar _name, Pchar _exp, bool _ignoreBlanks) {
   dbgLog(commandLog, Pchar("input %s (%s)"), _name, _exp);

   if (condition.IsFalse())
      return;

   char name[sizeString];
   Strcpy(name, _name, sizeString);

   PCStrList list = resolveString(name, true);
   while (list) {
      if (list->Number() == 1) {
         list->GotoHead();
         list->StrItem(name, sizeString);
         }
      list = resolveString(name, true);
      }

   if (StrEmpty(name))
      THROW(Pchar("Missing variable name in Input command"));

   PCVar tvar = resolveArg(_exp, true);
   if (!tvar)
      THROW(Pchar("Bad file name expansion for Input command"));

   char fileName[sizeString];
   toString(tvar, fileName, sizeString);

   PCStrList input;

   // Check for file name or HTTP url
   if (strlen(fileName) < 7)
      input = getFile(fileName, _ignoreBlanks);
   else {
      char sv7 = fileName[7];
      fileName[7] = 0;
      if (stricmp(fileName, Pchar("HTTP://")) == 0) {
         fileName[7] = sv7;
         PCVar tmp;
         input = getHTTP(fileName, tmp);
         }
      else {
         fileName[7] = sv7;
         input = getFile(fileName, _ignoreBlanks);
         }
      }

   // Determine the type of the document: plaintext, HTML, or XML
   TXT_Type type = CTXTType().Type(input);

   // If it is HTML, clean it up
   if (type == TXT_HTMLText)
      HTMLCleaner().FormHTML(input);

/*?
   if (CArgs::TheArgs->FArg(ftDebugFlag)) {
      char fname[sizeString];
      Strcpy(fname, name, sizeString);
      Strcat(fname, Pchar(".dbg"), sizeString);
      input->Save(fname);
      }
?*/

   PCAtom atom;
   if (type != TXT_PlainText) {
//?   input->Save(Pchar("input.lst"));

      // Load the XML/HTML document into the DOM
      PCDocument doc = new CDocument();
      doc->PreserveWhiteSpace(true);
      if (!doc->Load(input)) {
         PCDOMException error = doc->GetError();
         sprintf(message, Pchar("Input: Failed to load document <%s> into the DOM.\n"
                                "Error at line %d"),
                          error->Message(),
                          error->Line());
         throw(message);
         }

      // Save the document node in the user's specified variable
      atom = TYPECAST(doc, Atom);
      varHandler->Variable(name, atom);
      }
   else {
      // Save the file as a string list
      atom = TYPECAST(input, Atom);
      varHandler->Variable(name, atom);
      }
   } // cInput

//---------------------------------------------------------------------
// Load the specified document into a DOM, unless it is plaintext.
// Implements load using an HTTP Put.
//
void CXText::Resolve::cInputP(Pchar _name, Pchar _exp) {
   dbgLog(commandLog, Pchar("inputP %s (%s)"), _name, _exp);

   if (condition.IsFalse())
      return;

   char name[sizeString];
   Strcpy(name, _name, sizeString);

   PCStrList list = resolveString(name, true);
   while (list) {
      if (list->Number() == 1) {
         list->GotoHead();
         list->StrItem(name, sizeString);
         }
      list = resolveString(name, true);
      }

   if (StrEmpty(name))
      THROW(Pchar("Invalid variable name in InputP command"));

   PCVar pvar = varHandler->Variable(name);
   if (!pvar)
      THROW(Pchar("Undefined variable in InputP command"));

   PCVar tvar = resolveArg(_exp, true);
   if (!tvar)
      THROW(Pchar("Bad file name or url expansion for InputP command"));

   char fileName[sizeString];
   toString(tvar, fileName, sizeString);

   PCStrList input;

   // Check for file name or HTTP url
   if (strlen(fileName) < 7)
      input = getFile(fileName, /* _ignoreBlanks */ false);
   else {
      char sv7 = fileName[7];
      fileName[7] = 0;
      if (stricmp(fileName, Pchar("HTTP://")) == 0) {
         fileName[7] = sv7;
         input = getHTTP(fileName, pvar);
         }
      else {
         fileName[7] = sv7;
         input = getFile(fileName, /* _ignoreBlanks */ false);
         }
      }

   // Determine the type of the document: plaintext, HTML, or XML
   TXT_Type type = CTXTType().Type(input);

   // If it is HTML, clean it up
   if (type == TXT_HTMLText)
      HTMLCleaner().FormHTML(input);

/*?
   if (CArgs::TheArgs->FArg(ftDebugFlag)) {
      char fname[sizeString];
      Strcpy(fname, name, sizeString);
      Strcat(fname, Pchar(".dbg"), sizeString);
      input->Save(fname);
      }
?*/

   PCAtom atom;
   if (type != TXT_PlainText) {
      // Load the XML/HTML document into the DOM
      PCDocument doc = new CDocument();
      doc->PreserveWhiteSpace(true);
      if (!doc->Load(input)) {
         PCDOMException error = doc->GetError();
         sprintf(message, Pchar("InputP: Failed to load document <%s> into the DOM.\n"
                                "Error at line %d"),
                          error->Message(),
                          error->Line());
         throw(message);
         }

      // Save the document node in the user's specified variable
      atom = TYPECAST(doc, Atom);
      varHandler->Variable(name, atom);
      }
   else {
      // Save the file as a string list
      atom = TYPECAST(input, Atom);
      varHandler->Variable(name, atom);
      }
   } // cInputP

//---------------------------------------------------------------------
// Log the contents of the argument.
//
void CXText::Resolve::cLog(Pchar _name, Pchar _exp) {
   dbgLog(commandLog, Pchar("log %s (%s)"), _name, _exp);

   if (condition.IsFalse())
      return;

   char name[sizeString];
   Strcpy(name, _name, sizeString);

   if (!StrEmpty(name)) {
      PCStrList list = resolveString(name, true);
      while (list) {
         if (list->Number() == 1) {
            list->GotoHead();
            list->StrItem(name, sizeString);
            }
         list = resolveString(name, true);
         }
      }

   PCVar var = resolveArg(_exp, true);

   if (!var)
      sprintf(message, Pchar("Logging %s(%s)"), name, _exp);
   else {
      char temp[sizeString];
      toString(var, temp, sizeString);
      if (StrEmpty(name)) {
         Strcpy(message, temp, sizeString);
         }
      else
         sprintf(message, Pchar("%s='%s'"), name, temp);
      }

   log(true, message);
   } // cLog

//---------------------------------------------------------------------
//
void CXText::Resolve::cMatch(Pchar _text, Pchar _exp) {
   dbgLog(commandLog, Pchar("match %s (%s)"), _text, _exp);

   if (condition.IsFalse())
      return;

   match(_text, _exp);
   } // cMatch

//---------------------------------------------------------------------
//
void CXText::Resolve::cMssg(Pchar _arg) {
   dbgLog(commandLog, Pchar("mssg (%s)"), _arg);

   if (condition.IsFalse())
      return;

   PCVar var = resolveArg(_arg, true);

   char line[sizeString];
   toString(var, line, sizeString);

// Trim(line);

   if (!CConsole::TheConsole)
      return;

   if (!StrEmpty(line))
      CConsole::TheConsole->WriteString(line);

   CConsole::TheConsole->WriteString(Pchar(""));
   log(true, line);
   } // cMssg

/*?
//---------------------------------------------------------------------
// Uses the two fields in the specified path to build a reference from
// the specified topic.
//
// The format of path is
//
//    |[prefix]|XPath expression
//
// The prefix is optional. If it is present, then it is added to the
// reference before the value obtained using the XPath expression. The
// form of the result of this method is
//
//    [prefix.]value
//
// If the XPath expression specifies an element, then Pchar("/text()") is
// appended to access the text that the element contains. The check is
// pretty naive about XPath expressions: if it sees an at-sign (@) any-
// where in the expression, it just assumes that the expression is
// referencing an attribute (i.e., the expression is complete).
//
bool CXText::getRef(PCDocument _topic,
                    Pchar      _path,
                    Pchar      _value,
                    int        _size) {
#ifdef FV_ODB
   Assert(_topic);
   Assert(!StrEmpty(_path));
   Assert(!StrEmpty(_value));

   PCRecord rec = new CRecord(_path);
   char path[sizeString];
   rec->GetField(2, path, sizeString);

   _value[0] = 0;

   // If not a path to an attribute...
   if (ChrPos('@', path) < 0) {
      Strcat(path, Pchar("/text()"), sizeString);
      }

   PCNode node = TYPECAST(_topic, Node);
   node        = xp.SelectSingleNode(node, path);
   if (!node)
      return false;

   Pchar value = node->NodeValue();
   if (StrEmpty(value)) {
      delete[] value;
      return false;
      }

   rec->GetField(1, _value, _size);
   if (!StrEmpty(_value))
      ChrCat(_value, '.', _size);

   Strcat(_value, value, _size);
   delete[] value;

#endif // FV_ODB
   return true;
   } // getRef

//---------------------------------------------------------------------
// Uses the specified path information to obtain a list of reference
// values from the specified topic. See getRef above for particulars.
//
PCStrList CXText::getRefs(PCDocument _topic, PCRecord _keys) {
#ifdef FV_ODB
   PCStrList list = new CStrList();
   char seckey[sizeString];
   for (int ipri = 1; ipri <= _keys->Number(); ipri++) {
      _keys->GetField(ipri, seckey, sizeString);
      if (!StrEmpty(seckey)) {
         getRef(_topic, seckey, seckey, sizeString);
         if (!StrEmpty(seckey))
            list->StrInsert(seckey);
         }
      }
#endif // FV_ODB
   return list;
   } // getRefs
?*/

//---------------------------------------------------------------------
// Open the specified file using the default program for the file type.
//
void CXText::Resolve::cOpen(Pchar _name, Pchar _exp) {
   dbgLog(commandLog, Pchar("open %s (%s)"), _name, _exp);

   if (condition.IsFalse())
      return;

   char name[sizeString];
   Strcpy(name, _name, sizeString);

   PCStrList list = resolveString(name, true);
   while (list) {
      if (list->Number() == 1) {
         list->GotoHead();
         list->StrItem(name, sizeString);
         }
      list = resolveString(name, true);
      }

   bool wait = true;
   if (isFalse(name))
      wait = false;

   PCVar var = resolveArg(_exp, true);

   if (!var)
      THROW(Pchar("Bad file name expansion for Open command"));

   char fileName[sizeString];
   fileName[0] = '"';
   toString(var, fileName+1, sizeString-1);
   ChrCat(fileName, '"', sizeString);

   CProcess launch;
   Pchar exec = launch.FindExecutable(fileName);
   if (!StrEmpty(exec)) {
      // Make sure the file exists
      ulong err;
      if (!FileExists(fileName, err))
         CFile(fileName).Create();

      // Open the document with the program
      launch.Run(exec, fileName, wait);

      if (wait)
         // Handle return code
         varHandler->Variable(Pchar("error"), launch.Error());
      }

   delete[] exec;
   } // cOpen

//---------------------------------------------------------------------
// Output the contents of the output list to the specified file. Then
// flush the list.
//
void CXText::Resolve::cOut(Pchar     _name,
                           Pchar     _exp,
                           PCStrList _output,
                           bool      _append) {
   dbgLog(commandLog, Pchar("out %s (%s)"), _name, _exp);

   if (condition.IsFalse())
      return;

   char name[sizeString];
   Strcpy(name, _name, sizeString);

   if (!StrEmpty(name)) {
      PCStrList list = resolveString(name, true);
      while (list) {
         if (list->Number() == 1) {
            list->GotoHead();
            list->StrItem(name, sizeString);
            }
         list = resolveString(name, true);
         }
      }

   PCVar var = resolveArg(_exp, true);

   if (!var)
      THROW(Pchar("Bad file name expansion for Out command"));

   char fileName[sizeString];
   toString(var, fileName, sizeString);

   // Write the specified variable to the specified file
   if      (!StrEmpty(name) && stricmp(name, Pchar("output")) != 0) {
      var = varHandler->Variable(name);
      if         (var) {
         PCAtom atom = var->Value();
         if      (atom->Isa(iStrList)) {
            PCStrList list = TYPECAST(atom, StrList);
            if (!dumpOutput(list, fileName, _append)) {
               sprintf(message, Pchar("Couldn't write to %s"), fileName);
               throw(message);
               }
            }
         else if (atom->Isa(iBlock)) {
            PCBlock blk = TYPECAST(atom, Block);
            Pchar p = blk->Lock();
            blk->Unlock();
            PCStrList list = new CStrList();
            list->BlockTo(atom);
            if (!dumpOutput(list, fileName, _append)) {
               sprintf(message, Pchar("Couldn't write to %s"), fileName);
               throw(message);
               }
            }
         else if (atom->Isa(iNode)) {
            Pchar p = toString(var);
            PCAtom blk = new CBlock(p);
            delete[] p;
            PCStrList list = new CStrList();
            list->BlockTo(blk);
            if (!dumpOutput(list, fileName, _append)) {
               sprintf(message, Pchar("Couldn't write to %s"), fileName);
               throw(message);
               }
            }
         else if (atom->Isa(iNodeList)) {
            Pchar p = toString(var);
            PCAtom blk = new CBlock(p);
            delete[] p;
            PCStrList list = new CStrList();
            list->BlockTo(blk);
            if (!dumpOutput(list, fileName, _append)) {
               sprintf(message, Pchar("Couldn't write to %s"), fileName);
               throw(message);
               }
            }
         else if (atom->Isa(iDocument)) {
            PCDocument doc = TYPECAST(atom, Document);
            if (!doc->Save(fileName)) {
               sprintf(message, Pchar("Couldn't write to %s"), fileName);
               throw(message);
               }
            }
         else {                    // This is really some kind of error
            Pchar p = toString(var);
            PCAtom blk = new CBlock(p);
            delete[] p;
            PCStrList list = new CStrList();
            list->BlockTo(blk);
            if (!dumpOutput(list, fileName, _append)) {
               sprintf(message, Pchar("Couldn't write to %s"), fileName);
               throw(message);
               }
            }
         }
      }

   // Write the tags output list to the specified file
   else if (!dumpOutput(_output, fileName, _append)) {
      sprintf(message, Pchar("Couldn't write to %s"), fileName);
      throw(message);
      }
   else
      _output->Flush();

   didOut = true;
   } // cOut

//---------------------------------------------------------------------
//
void CXText::Resolve::cPause(Pchar _arg) {
   dbgLog(commandLog, Pchar("pause (%s)"), _arg);

   if (condition.IsFalse())
      return;

   PCVar var = resolveArg(_arg, true);

   char line[sizeString];
   toString(var, line, sizeString);

//?Trim(line);

   if (!CConsole::TheConsole)
      return;

   if (!StrEmpty(line))
      CConsole::TheConsole->WriteString(line);

   CConsole::TheConsole->WriteString(Pchar(""));
   CConsole::TheConsole->WriteSubStr(Pchar("> "));
   CConsole::TheConsole->FlushInput();
   CConsole::TheConsole->ReadChar();
   } // cPause

//---------------------------------------------------------------------
// Play the specified wave-file
//
void CXText::Resolve::cPlay(Pchar _arg) {
   dbgLog(commandLog, Pchar("play (%s)"), _arg);

   if (condition.IsFalse())
      return;

   PCVar var = resolveArg(_arg, true);

   if (!var)
      THROW(Pchar("Bad file name expansion for Out command"));

   char temp[sizeString];
   toString(var, temp, sizeString);

#ifdef FV_SOUND
   CSound().PlaySound(receiver, temp, soundPath);
#endif
   } // cPlay

//---------------------------------------------------------------------
// $#pop var(stack)
//
void CXText::Resolve::cPop(Pchar _name, Pchar _exp) {
   dbgLog(commandLog, Pchar("pop %s (%s)"), _name, _exp);

   if (condition.IsFalse())
      return;

   char name[sizeString];
   Strcpy(name, _name, sizeString);

   PCStrList list = resolveString(name, true);
   while (list) {
      if (list->Number() == 1) {
         list->GotoHead();
         list->StrItem(name, sizeString);
         }
      list = resolveString(name, true);
      }
   if (StrEmpty(name))
      THROW(Pchar("Invalid variable name in Pop command"));

   PCVar var = varHandler->Variable(name);
   if (!var) {
      PCAtom atm = new CBlock();
      var = varHandler->Variable(name, atm);
      }

   Strcpy(name, _exp, sizeString);
   list = resolveString(name, true);
   while (list) {
      if (list->Number() == 1) {
         list->GotoHead();
         list->StrItem(name, sizeString);
         }
      list = resolveString(name, true);
      }
   if (StrEmpty(name))
      THROW(Pchar("Invalid stack name in Pop command"));

   PCVar stk = varHandler->Variable(name);
   if (!stk)
      THROW(Pchar("Missing stack in Pop command"));

   PCVar temp = stk->Pop();
   PCAtom atm = temp->Value();
   var->Value(atm);
  } // cPop

//---------------------------------------------------------------------
// $#push stack(var)
//
void CXText::Resolve::cPush(Pchar _name, Pchar _exp) {
   dbgLog(commandLog, Pchar("push %s (%s)"), _name, _exp);

   if (condition.IsFalse())
      return;

   char name[sizeString];
   Strcpy(name, _name, sizeString);

   PCStrList list = resolveString(name, true);
   while (list) {
      if (list->Number() == 1) {
         list->GotoHead();
         list->StrItem(name, sizeString);
         }
      list = resolveString(name, true);
      }
   if (StrEmpty(name))
      THROW(Pchar("Invalid variable name in Push command"));

   PCVar stk = varHandler->Variable(name);
   if (!stk) {
      PCAtom atm = new CBlock();
      stk = varHandler->Variable(name, atm);
      }

   PCVar var = resolveArg(_exp, false);
   if (!var)
      var = new CVar();

   stk->Push(var);
   } // cPush

//---------------------------------------------------------------------
//
void CXText::Resolve::cSet(Pchar _name, Pchar _exp) {
   dbgLog(commandLog, Pchar("set %s (%s)"), _name, _exp);

   if (condition.IsFalse())
      return;

   char name[sizeString];
   Strcpy(name, _name, sizeString);

   PCStrList list = resolveString(name, true);
   while (list) {
      if (list->Number() == 1) {
         list->GotoHead();
         list->StrItem(name, sizeString);
         }
      list = resolveString(name, true);
      }
   if (StrEmpty(name))
      THROW(Pchar("Invalid variable name in Set command"));

   // If set random, then randomize
   if (stricmp(name, Pchar("random")) == 0) {
#ifdef __BORLANDC__
      randomize();
#endif // __BORLANDC__
      return;
      }

   // Get the destination variable
   PCAtom atm;
   PCVar var  = varHandler->Variable(name);
   if (!var) {
      atm = new CBlock();
      var = varHandler->Variable(name, atm);
      }
   else
      atm = var->Value();

   PCVar temp;
   if (!StrEmpty(_exp))
      temp = resolveArg(_exp, false);

   // If no expression, then set the variable to empty
   else {
      temp = new CVar();
      temp->Value(Pchar(""));
      }

   PCAtom tatm = temp->Value();

   // Store the expression value as the variable
   if      (tatm->Isa(iNode)) {
      var = varHandler->Variable(name, tatm);
      return;
      }

   else if (tatm->Isa(iNodeList)) {
      var = varHandler->Variable(name, tatm);
      return;
      }

   else if (tatm->Isa(iBlock)) {
      PCStrList lista = new CStrList();
      lista->BlockTo(tatm);
      PCStrList listb = new CStrList();
      Resolve* r = new Resolve();
      r->resolve(lista, listb);
      delete r;
      atm = new CBlock();
      listb->ToBlock(atm);
      var->Value(atm);
      }

   else if (tatm->Isa(iStrList)) {
      PCStrList lista = TYPECAST(tatm, StrList);
      PCStrList listb = new CStrList();
      Resolve* r = new Resolve();
      r->resolve(lista, listb);
      delete r;
      atm = new CBlock();
      listb->ToBlock(atm);
      var->Value(atm);
      }

   else {
      sprintf(message, Pchar("Cannot add type = %d in  %s(%s)"), tatm->IAm(), name, _exp);
      throw(message);
      }

   atm = var->Value();
   if (!atm)
      THROW(Pchar("Value corrupt in Set command"));

/*?
   // If the temp variable contains XML in text form, convert it to
   // a DOM
   if      (atm->Isa(iBlock)) {
      PCBlock blk = TYPECAST(atm, Block);
      Pchar p = blk->Lock();
      if (!StrEmpty(p) && *p == '<') {
         PCDocument doc = new CDocument();
         doc->PreserveWhiteSpace(true);
         if (doc->Load(blk)) {
            PCAtom atom = TYPECAST(doc, Atom);
            varHandler->Variable(name, atom);
            }
         }
      blk->Unlock();
      }
   else if (atm->Isa(iStrList)) {
      PCStrList lst = TYPECAST(atm, StrList);
      Pchar p = lst->StrItem(1);
      if (!StrEmpty(p) && *p == '<') {
         PCDocument doc = new CDocument();
         doc->PreserveWhiteSpace(true);
         if (doc->Load(lst)) {
            PCAtom atom = TYPECAST(doc, Atom);
            varHandler->Variable(name, atom);
            }
         }
      }

   atm = var->Value();

   PCDocument olddoc = getDoc(var->Value());
   PCDocument newdoc = getDoc(atm);

   if (newdoc)
      newdoc->RegRef();
   if (olddoc)
      olddoc->DrgRef();
?*/
   } // cSet

//---------------------------------------------------------------------
//
void CXText::Resolve::cSetAsText(Pchar _name, Pchar _exp) {
   dbgLog(commandLog, Pchar("settext %s (%s)"), _name, _exp);

   if (condition.IsFalse())
      return;

   char name[sizeString];
   Strcpy(name, _name, sizeString);

   PCStrList list = resolveString(name, true);
   while (list) {
      if (list->Number() == 1) {
         list->GotoHead();
         list->StrItem(name, sizeString);
         }
      list = resolveString(name, true);
      }
   if (StrEmpty(name))
      THROW(Pchar("Invalid variable name in Set command"));

   // If set random, then randomize
   if (stricmp(name, Pchar("random")) == 0) {
#ifdef __BORLANDC__
      randomize();
#endif // __BORLANDC__
      return;
      }

   // Get the destination variable
   PCAtom atm;
   PCVar var  = varHandler->Variable(name);
   if (!var) {
      atm = new CStrList();
      var = varHandler->Variable(name, atm);
      }
   else {
      atm = var->Value();
      if      (!atm) {
         THROW(Pchar("Destination variable undefined"));
         }
      else if (atm->Isa(iBlock)) {
         PCStrList list = new CStrList();
         list->BlockTo(atm);
         atm = list;
         }
      }

   PCVar temp;
   if (!StrEmpty(_exp))
      temp = resolveArg(_exp, false);

   // If no expression, then set to empty
   else {
      temp = new CVar();
      temp->Value(Pchar(""));
      }

   PCAtom tatm = temp->Value();

   // Append the expression value to the container variable
   if      (tatm->Isa(iNode)) {
      PCNode node = TYPECAST(tatm, Node);
      if (node->NodeType() == ATTRIBUTE_NODE ||
          node->NodeType() == TEXT_NODE) {
         PCStrList lista = new CStrList();
         Pchar p         = node->NodeValue();
         PCAtom blk      = new CBlock(p);
         delete[] p;
         lista->BlockTo(blk);
         PCStrList listb = new CStrList();
         Resolve* r = new Resolve();
         r->resolve(lista, listb);
         delete r;
         atm = new CBlock();
         listb->ToBlock(atm);
         var->Value(atm);
         }
      else {
         sprintf(message, Pchar("Cannot add type = %d in  %s(%s)"), tatm->IAm(), name, _exp);
         throw(message);
         }
      }
   else if (tatm->Isa(iNodeList)) {
      PCStrList lista = new CStrList();
      PCNodeList list = TYPECAST(tatm, NodeList);
      for (int i = 0; i < list->Length(); i++) {
         PCNode node = list->Item(i);
         if (node->NodeType() == ATTRIBUTE_NODE ||
             node->NodeType() == TEXT_NODE) {
            PCStrList listb = new CStrList();
            Pchar p         = node->NodeValue();
            PCAtom blk      = new CBlock(p);
            delete[] p;
            listb->BlockTo(blk);
            PCStrList listc = new CStrList();
            Resolve* r = new Resolve();
            r->resolve(listb, listc);
            delete r;
            lista->ListInsert(listc, atEnd);
            }
         else {
            sprintf(message, Pchar("Cannot add type = %d in  %s(%s)"), tatm->IAm(), name, _exp);
            throw(message);
            }
         }
      atm = new CBlock();
      lista->ToBlock(atm);
      var->Value(atm);
      }
   else if (tatm->Isa(iStrList)) {
      PCStrList lista = TYPECAST(tatm, StrList);
      PCStrList listb = new CStrList();
      Resolve* r = new Resolve();
      r->resolve(lista, listb);
      delete r;
      atm = new CBlock();
      listb->ToBlock(atm);
      }
   else if (tatm->Isa(iBlock)) {
      PCStrList lista = new CStrList();
      lista->BlockTo(tatm);
      PCStrList listb = new CStrList();
      Resolve* r = new Resolve();
      r->resolve(lista, listb);
      delete r;
      atm = new CBlock();
      listb->ToBlock(atm);
      var->Value(atm);
      }
   else {
      sprintf(message, Pchar("Cannot add type = %d in  %s(%s)"), tatm->IAm(), name, _exp);
      throw(message);
      }
   } // cSetAsText

//---------------------------------------------------------------------
// This command should support both XML and HTML. Doesn't yet support
// the latter.
//
void CXText::Resolve::cSetAsXML(Pchar _name, Pchar _exp) {
   dbgLog(commandLog, Pchar("setxml %s (%s)"), _name, _exp);

   if (condition.IsFalse())
      return;

   char name[sizeString];
   Strcpy(name, _name, sizeString);

   PCStrList list = resolveString(name, true);
   while (list) {
      if (list->Number() == 1) {
         list->GotoHead();
         list->StrItem(name, sizeString);
         }
      list = resolveString(name, true);
      }
   if (StrEmpty(name))
      THROW(Pchar("Invalid variable name in Set command"));

   // If set random, then randomize
   if (stricmp(name, Pchar("random")) == 0) {
#ifdef __BORLANDC__
      randomize();
#endif // __BORLANDC__
      return;
      }

   // Get the destination variable
   PCAtom atm;
   PCVar var  = varHandler->Variable(name);
   if (!var) {
      atm = new CStrList();
      var = varHandler->Variable(name, atm);
      }
   else
      atm = var->Value();

   PCVar temp;
   if (!StrEmpty(_exp))
      temp = resolveArg(_exp, false);

   // Initialize the variable to an empty XML document
   else {
      temp = new CVar();
      PCAtom atm = new CDocument();
      temp->Value(atm);
      }

   PCAtom tatm = temp->Value();

   // Append the expression value to the container variable
   if      (tatm->Isa(iNode)) {
      PCNode node = TYPECAST(tatm, Node);
      if (node->NodeType() == ATTRIBUTE_NODE ||
          node->NodeType() == TEXT_NODE) {
         PCStrList lista = new CStrList();
         Pchar p         = node->NodeValue();
         PCAtom blk      = new CBlock(p);
         delete[] p;
         lista->BlockTo(blk);
         PCStrList listb = new CStrList();
         Resolve* r      = new Resolve();
         r->resolve(lista, listb);
         delete r;
         atm = new CBlock();
         listb->ToBlock(atm);
         var->Value(atm);
         }
      else {
         sprintf(message, Pchar("Cannot add type = %d in  %s(%s)"), tatm->IAm(), name, _exp);
         throw(message);
         }
      }
   else if (tatm->Isa(iNodeList)) {
      PCStrList lista = new CStrList();
      PCNodeList list = TYPECAST(tatm, NodeList);
      for (int i = 0; i < list->Length(); i++) {
         PCNode node = list->Item(i);
         if (node->NodeType() == ATTRIBUTE_NODE ||
             node->NodeType() == TEXT_NODE) {
            PCStrList listb = new CStrList();
            Pchar p         = node->NodeValue();
            PCAtom blk      = new CBlock(p);
            delete[] p;
            listb->BlockTo(blk);
            PCStrList listc = new CStrList();
            Resolve* r      = new Resolve();
            r->resolve(listb, listc);
            delete r;
            lista->ListInsert(listc, atEnd);
            }
         else {
            sprintf(message, Pchar("Cannot add type = %d in  %s(%s)"), tatm->IAm(), name, _exp);
            throw(message);
            }
         }
      atm = new CBlock();
      lista->ToBlock(atm);
      var->Value(atm);
      }
   else if (tatm->Isa(iStrList)) {
      PCStrList lista = TYPECAST(tatm, StrList);
      PCStrList listb = new CStrList();
      Resolve* r      = new Resolve();
      r->resolve(lista, listb);
      delete r;
      atm = new CBlock();
      listb->ToBlock(atm);
      var->Value(atm);
      }
   else if (tatm->Isa(iBlock)) {
      PCStrList lista = new CStrList();
      lista->BlockTo(tatm);
      PCStrList listb = new CStrList();
      Resolve* r      = new Resolve();
      r->resolve(lista, listb);
      delete r;
      atm = new CBlock();
      listb->ToBlock(atm);
      var->Value(atm);
      }
   else {
      sprintf(message, Pchar("Cannot add type = %d in  %s(%s)"), tatm->IAm(), name, _exp);
      throw(message);
      }

   atm = var->Value();
   if (!atm)
      THROW(Pchar("Value corrupt in XML command"));

   // If the variable contains XML in text form, load it into a DOM
   if      (atm->Isa(iBlock)) {
      PCDocument doc = new CDocument();
      doc->PreserveWhiteSpace(true);
      PCBlock blk = TYPECAST(atm, Block);
      if (doc->Load(blk))
         atm = TYPECAST(doc, Atom);
      else {
         sprintf(message, Pchar("Value in <%s> not valid XML"), _exp);
         throw(message);
         }
      }
   else if (atm->Isa(iStrList)) {
      PCDocument doc = new CDocument();
      doc->PreserveWhiteSpace(true);
      PCStrList lst = TYPECAST(atm, StrList);
//?   lst->Save(Pchar("setAsXML.lst"));
      if (doc->Load(lst))
         atm = TYPECAST(doc, Atom);
      else {
         sprintf(message, Pchar("Value in <%s> not valid XML"), _exp);
         throw(message);
         }
      }
   else {
      sprintf(message, Pchar("Value in <%s> not text"), _exp);
      throw(message);
      }

   atm = var->Value();

   PCDocument olddoc = getDoc(var->Value());
   PCDocument newdoc = getDoc(atm);

   if (newdoc)
      newdoc->RegRef();
   if (olddoc)
      olddoc->DrgRef();
   } // cSetAsXML

//---------------------------------------------------------------------
//
void CXText::Resolve::cSleep(Pchar _arg) {
   dbgLog(commandLog, Pchar("sleep (%s)"), _arg);

   if (condition.IsFalse())
      return;

   PCVar var = resolveArg(_arg, true);

   char line[sizeString];
   toString(var, line, sizeString);

   Trim(line);

   if (!StrEmpty(line)) {
      long sleepTime = atoi(line) * 1000;
      ::Sleep(sleepTime);
      }
   } // cSleep

//---------------------------------------------------------------------
//
// name = variable name, exp = SQL command (Update, Create, etc)
//
void CXText::Resolve::cSQLCommand(Pchar _name, Pchar _exp) {
   dbgLog(commandLog, Pchar("SQLCommand %s (%s)"), _name, _exp);

   if (condition.IsFalse())
      return;

   char name[sizeString];
   Strcpy(name, _name, sizeString);

   PCStrList list = resolveString(name, true);
   while (list) {
      if (list->Number() == 1) {
         list->GotoHead();
         list->StrItem(name, sizeString);
         }
      list = resolveString(name, true);
      }
   if (StrEmpty(name))
      THROW(Pchar("Invalid variable name in Set command"));

   PCVar temp = resolveArg(_exp, false);
   if (!temp)
      THROW(Pchar("Value is not text"));

   // Get the SQL statement
   char todo[sizeString];
   toString(temp, todo, sizeString);
   Trim(todo);

   // Get the DSN
   char dsn[sizeString];
   toString(Pchar("dsn"), dsn, sizeString);

   // Get the user name
   char userName[sizeString];
   toString(Pchar("userName"), userName, sizeString);

   // Get the password
   char password[sizeString];
   toString(Pchar("password"), password, sizeString);

   // Connect to the database
   ODBCInterface db;
   if (!db.Open(dsn, userName, password)) {
      sprintf(message, Pchar("SQLWrite could not connect to DB with %s"), dsn);
      throw(message);
      }

   // Perform the SQL command
   int count = db.Execute(todo);
   if (count <= 0) {
      sprintf(message, Pchar("SQLWrite command failed with <%s>"), todo);
      throw(message);
      }

   // Set variable to the number of rows effected
   if (!StrEmpty(name))
      PCVar var = varHandler->Variable(name, count);
   db.Close();
   } // cSQLCommand

//---------------------------------------------------------------------
// name = variable name, exp = SQL Query (Select)
//
void CXText::Resolve::cSQLQuery(Pchar _name, Pchar _exp) {
   dbgLog(commandLog, Pchar("SQLQuery %s (%s)"), _name, _exp);

   if (condition.IsFalse())
      return;

   char name[sizeString];
   Strcpy(name, _name, sizeString);

   PCStrList list = resolveString(name, true);
   while (list) {
      if (list->Number() == 1) {
         list->GotoHead();
         list->StrItem(name, sizeString);
         }
      list = resolveString(name, true);
      }
   if (StrEmpty(name))
      THROW(Pchar("Invalid variable name in Set command"));

   PCVar temp = resolveArg(_exp, false);
   if (!temp)
      THROW(Pchar("Value is not text"));

   // Obtain the contents of the sqlcolumns variable
   char columns[sizeString];
   toString(Pchar("sqlcolumns"), columns, sizeString);

   // Get the query and set up the bindery
   char query[sizeString];
   toString(temp, query, sizeString);
   Trim(query);

   // Get the DSN
   char dsn[sizeString];
   toString(Pchar("dsn"), dsn, sizeString);

   // Get the user name
   char userName[sizeString];
   toString(Pchar("userName"), userName, sizeString);

   // Get the password
   char password[sizeString];
   toString(Pchar("password"), password, sizeString);

   // Connect to the database
   ODBCInterface db;
   if (!db.Open(dsn, userName, password)) {
      sprintf(message, Pchar("SQLQuery could not connect to DB with %s"), dsn);
      throw(message);
      }

   // Perform the query
   int count = db.Query(query);
   if (count <= 0) {
      sprintf(message, Pchar("SQLQuery query failed with <%s>"), query);
      throw(message);
      }

   // Need the sqlcolumns variable as a string if available
   toString(Pchar("SQLColumns"), columns, sizeString);

   // Get the column names from the DB if available
   if (db.BindColumns(count, columns, sizeString))
      varHandler->Variable(Pchar("SQLColumns"), columns);

   // Collect the responses into a StrList
   PCStrList lst = new CStrList();
   for (;;) {
      // Get a row from the result set
      PCBlock blk = new CBlock();
      blk->Copy(Pchar(""), 0);
      if (!getNextRow(db, blk, columns))
         break;

      Pchar p = blk->Lock();
      lst->StrInsert(p);
      }

   if (!StrEmpty(name)) {
      PCAtom atm = TYPECAST(lst, Atom);
      PCVar var = varHandler->Variable(name, atm);
      }
   db.Close();
   } // SQLQuery

//---------------------------------------------------------------------
//
void CXText::Resolve::cStop(Pchar _arg) {
   dbgLog(commandLog, Pchar("stop (%s)"), _arg);

   if (condition.IsFalse())
      return;

   PCVar var = resolveArg(_arg, true);
   toString(var, message, sizeString);

//?Trim(line);

   if (CConsole::TheConsole) {
   if (!StrEmpty(message))
      CConsole::TheConsole->WriteString(message);

      CConsole::TheConsole->WriteString(Pchar(""));
      CConsole::TheConsole->WriteSubStr(Pchar("> "));
      CConsole::TheConsole->FlushInput();
      CConsole::TheConsole->ReadChar();
      }

   throw(message);
   } // cStop

//---------------------------------------------------------------------
//
void CXText::Resolve::cTrace(Pchar _arg) {
   dbgLog(commandLog, Pchar("trace (%s)"), _arg);

   if (condition.IsFalse())
      return;

   PCVar var = resolveArg(_arg, true);

   char line[sizeString];
   toString(var, line, sizeString);

   Trim(line);
   Squeeze(line, '0');
   } // cTrace

//---------------------------------------------------------------------
// Format of input _arg is (fromCharSet,toCharSet).
// Format of output is
//
//   ,translate,arg,fun                                    translate on
// or
//   ,translate,                                          translate off
//
// This version only translates the contents of the specified variable.
//
//   $#translate varname(fromCharSet,toCharSet)
//
void CXText::Resolve::cTranslate(Pchar _name, Pchar _exp) {
   dbgLog(commandLog, Pchar("translate %s (%s)"), _name, _exp);

   if (condition.IsFalse())
      return;

   if (StrEmpty(_name))
      return;
   if (StrEmpty(_exp))
      return;

   char name[sizeString];
   Strcpy(name, _name, sizeString);

   if (!StrEmpty(name)) {
      PCStrList list = resolveString(name, true);
      while (list) {
         if (list->Number() == 1) {
            list->GotoHead();
            list->StrItem(name, sizeString);
            }
         list = resolveString(name, true);
         }
      }

   // No variable to translate
   if (StrEmpty(name))
      return;

   // Get the variable to translate
   PCAtom atm;
   PCVar var  = varHandler->Variable(name);
   if (!var)
      return;

   atm = var->Value();
   if      (!atm) {
      THROW(Pchar("Translate Error: Variable undefined"));
      }

   if (!atm->Isa(iBlock)) {
      THROW(Pchar("Translate Error: Variable must be type string"));
      }

   PCBlock block = TYPECAST(atm, Block);

   PCRecord rec = new CRecord(_exp, Pchar(","));

   int largs = rec->LenField(1)+1;
   int lfuns = rec->LenField(2)+1;
   if (largs > lfuns) {
      THROW(Pchar("Translate Error: Function string cannot be shorter than the argument string."));
      }

   Pchar args = new char[largs];
   rec->GetField(1, args, largs);

   Pchar funs = new char[lfuns];
   rec->GetField(2, funs, lfuns);

   // Map each arg character to its corresponding fun character
   block->RepChars(args, funs);

   delete[] args;
   delete[] funs;
   } // cTranslate

//---------------------------------------------------------------------
// Format of input _arg is (fromCharSet,toCharSet).
// Format of output is
//
//   ,translate,arg,fun                                    translate on
// or
//   ,translate,                                          translate off
//
// Also a less expansive form:
//
//   $#translate varname(fromCharSet,toCharSet)
//
// translates the contents of the specified variable.
//
void CXText::Resolve::cTranslate(Pchar _arg, PCStrList _output) {
   dbgLog(commandLog, Pchar("translate %s"), _arg);

   if (condition.IsFalse())
      return;

   char arg[sizeString];
   char line[sizeString];
   After(Pchar("("),  _arg, arg,  sizeString);
   Before(Pchar(")"), arg,  line, sizeString);
   Trim(line);

   if (StrEmpty(line)) {
      _output->StrInsert(Pchar(",translate,"), atEnd);
      }
   else {
      PCRecord rec = new CRecord(line, Pchar(","));
      rec->InsertField(1, Pchar("translate"));

      char arg[sizeString];
      rec->GetField(2, arg, sizeString);
      Trim(arg);
      rec->PutField(2, arg);

      char fun[sizeString];
      rec->GetField(3, fun, sizeString);
      Trim(fun);
      rec->PutField(3, fun);

      if (strlen(arg) > strlen(fun))
         THROW(Pchar("Bad Translate command - first list longer than the second list"));

      log(false, rec->GetDelimStrD(false));
      _output->StrInsert(rec->GetDelimStrD(false), atEnd);
      }
   } // cTranslate

   /*?
//---------------------------------------------------------------------
//
void CXText::Resolve::cVars(Pchar _arg) {
   if (condition.IsFalse())
      return;

   PCVar var = resolveArg(_arg, true);

   char line[sizeString];
   toString(var, line, sizeString);

   Trim(line);
   dumpVars(varHandler, true, line);
   } // cVars
?*/

//---------------------------------------------------------------------
// Return true if this line is to be ignored. Note that the command 
// prefix must be the first non-blank occupant in the line. I.e.,
// $#out should be ignored because the prefix is not the first thing on
// the line, Pchar("//") is.
//
bool CXText::Resolve::command(PCStrList _body,
                              PCStrList _output) {
   setDateTime();

   // Check for a command line
   char line[sizeString];
   _body->StrItem(line, sizeString);
   Trim(line);
   int i = StrPos(Pchar(ftCmdStr), line);

   // 2019.06.05 pjm: Check if the prefix is present and that it is the
   // first non-blank occupant. Interesting note: I wrote a short script
   // to modify cpp-files and it worked for all but this file (xtext.cpp).
   // The reason is because Pchar("$#") appears several places in the
   // file, and the interpreter tried to execute the command - which
   // was  not a command. So I added the requirement that it cannot be
   // preceeded on the line by other stuff. This was implied in the
   // manual, but not enforced. Actually, the problems are bigger than I
   // thought. I gave up getting this file to work as fodder for a tags
   // script.
   if (i != 0)
      return false;

   // Test for comment
   if (line[i+2] == ' ')
      return true;

   strcpy(line, line+i+2);
   if (StrEmpty(line))
      return true;

   // Extract the command name
   char cmmd[sizeString];
   i = ChrPos(' ', line);
   int j = ChrPos('(', line);
   if      (i < 0 && j < 0) {
      Strcpy(cmmd, line, sizeString);
      line[0] = 0;
      }
   else if (i < 0) {
      line[j] = 0;
      Strcpy(cmmd, line, sizeString);
      line[j] = '(';
      strcpy(line, line+j);
      }
   else if (j < 0) {
      line[i] = 0;
      Strcpy(cmmd, line, sizeString);
      strcpy(line, line+i+1);
      }
   else if (j < i) {
      line[j] = 0;
      Strcpy(cmmd, line, sizeString);
      line[j] = '(';
      strcpy(line, line+j);
      }
   else {
      line[i] = 0;
      Strcpy(cmmd, line, sizeString);
      strcpy(line, line+i+1);
      }

/*?
   // Extract the command name
   char cmmd[sizeString];
   i = ChrPos(' ', line);
   if (i > 0) {
      line[i] = 0;
      Strcpy(cmmd, line, sizeString);
      strcpy(line, line+i+1);
      }
   else {
      Strcpy(cmmd, line, sizeString);
      line[0] = 0;
      }
?*/

   CapStr(cmmd);

   // Extract the name and argument (if present)
   char name[sizeString];
   char arg[sizeString];
   if (strcmp(cmmd, Pchar("TRANSLATE")) == 0 && line[0] == '(')
      _body->StrItem(arg, sizeString);
   else {
      extractArgs(line, name, sizeString, arg, sizeString);
      Trim(name);
      }

   // Process the command
   if      (strcmp(cmmd, Pchar("IF"))        == 0)
      cIf(name, arg);
   else if (strcmp(cmmd, Pchar("IFN"))       == 0)
      cIfN(name, arg);
   else if (strcmp(cmmd, Pchar("ELIF"))      == 0)
      cElif(name, arg);
   else if (strcmp(cmmd, Pchar("ELIFN"))     == 0)
      cElifN(name, arg);
   else if (strcmp(cmmd, Pchar("ELSE"))      == 0)
      cElse();
   else if (strcmp(cmmd, Pchar("END"))       == 0)
      cEnd();

   else if (strcmp(cmmd, Pchar("FOREACH"))   == 0)
      cForEach(name, arg, _body, _output);
   else if (strcmp(cmmd, Pchar("FORELSE"))   == 0)
      cForElse();
   else if (strcmp(cmmd, Pchar("WHILE"))     == 0)
      cWhile(name, arg, _body, _output);
   else if (strcmp(cmmd, Pchar("WHILEN"))    == 0)
      cWhileN(name, arg, _body, _output);
   else if (strcmp(cmmd, Pchar("LAST"))      == 0)
      cLast(name, arg);
   else if (strcmp(cmmd, Pchar("LASTN"))     == 0)
      cLastN(name, arg);
   else if (strcmp(cmmd, Pchar("NEXT"))      == 0)
      cNext(name, arg);
   else if (strcmp(cmmd, Pchar("NEXTN"))     == 0)
      cNextN(name, arg);

   else if (!condition.IsFalse()) {

      if      (strcmp(cmmd, Pchar("ADD"))       == 0)
         cAdd(name, arg, atEnd);
      else if (strcmp(cmmd, Pchar("ADDS"))      == 0)
         cAdd(name, arg, sorted);
      else if (strcmp(cmmd, Pchar("ADDT"))      == 0)
         cAdd(name, arg, atTop);
      else if (strcmp(cmmd, Pchar("ADDU"))      == 0)
         cAdd(name, arg, unique);
      else if (strcmp(cmmd, Pchar("CONSOLE"))   == 0)
         cConsole(arg);
      else if (strcmp(cmmd, Pchar("DEBUG"))     == 0)
         cDebug(arg);
      else if (strcmp(cmmd, Pchar("DEFER"))     == 0)
         cDefer(name, arg);
   //?else if (strcmp(cmmd, Pchar("DROP"))      == 0)
   //?   cDrop(arg);
      else if (strcmp(cmmd, Pchar("EXEC"))      == 0)
         cExec(name, arg);
   /*? TBD:
      else if (strcmp(cmmd, Pchar("FILEREAD"))  == 0)
         cFILERead(name, arg);
      else if (strcmp(cmmd, Pchar("FILEWRITE")) == 0)
         cFILEWrite(name, arg);
   ?*/
      else if (strcmp(cmmd, Pchar("GET"))       == 0)
         cGet(name, arg);
      else if (strcmp(cmmd, Pchar("IN"))        == 0)
         cInput(name, arg, /* _ignoreBlanks */ false);
      else if (strcmp(cmmd, Pchar("INB"))       == 0)
         cInput(name, arg, /* _ignoreBlanks */ true);
      else if (strcmp(cmmd, Pchar("INP"))       == 0)
         cInputP(name, arg);
      else if (strcmp(cmmd, Pchar("LOG"))       == 0)
         cLog(name, arg);
      else if (strcmp(cmmd, Pchar("MATCH"))     == 0)
         cMatch(name, arg);
      else if (strcmp(cmmd, Pchar("MESSAGE"))   == 0)
         cMssg(arg);
      else if (strcmp(cmmd, Pchar("MSG"))       == 0)
         cMssg(arg);
   //?else if (strcmp(cmmd, Pchar("ODBREAD"))   == 0)
   //?   cODBRead(name, arg);
   //?else if (strcmp(cmmd, Pchar("ODBWRITE"))  == 0)
   //?   cODBWrite(name, arg);
      else if (strcmp(cmmd, Pchar("OPEN"))      == 0)
         cOpen(name, arg);
      else if (strcmp(cmmd, Pchar("OUT"))       == 0)
         cOut(name, arg, _output, /* append */ false);
      else if (strcmp(cmmd, Pchar("OUTA"))      == 0)
         cOut(name, arg, _output, /* append */ true);
      else if (strcmp(cmmd, Pchar("PAUSE"))     == 0)
         cPause(arg);
      else if (strcmp(cmmd, Pchar("PLAY"))      == 0)
         cPlay(arg);
      else if (strcmp(cmmd, Pchar("POP"))       == 0)
         cPop(name, arg);
      else if (strcmp(cmmd, Pchar("PUSH"))      == 0)
         cPush(name, arg);
      else if (strcmp(cmmd, Pchar("SET"))       == 0)
         cSet(name, arg);
      else if (strcmp(cmmd, Pchar("TEXT"))      == 0)
         cSetAsText(name, arg);
      else if (strcmp(cmmd, Pchar("TXT"))       == 0)
         cSetAsText(name, arg);
      else if (strcmp(cmmd, Pchar("XML"))       == 0)
         cSetAsXML(name, arg);
      else if (strcmp(cmmd, Pchar("SQLQUERY"))  == 0)
         cSQLQuery(name, arg);
      else if (strcmp(cmmd, Pchar("SQLCOMMAND"))== 0)
         cSQLCommand(name, arg);
      else if (strcmp(cmmd, Pchar("SLEEP"))     == 0)
         cSleep(arg);
      else if (strcmp(cmmd, Pchar("STOP"))      == 0)
         cStop(arg);
      else if (strcmp(cmmd, Pchar("TRACE"))     == 0)
         cTrace(arg);
      else if (strcmp(cmmd, Pchar("TRANSLATE")) == 0) {
         if (!StrEmpty(name))
            cTranslate(name, arg);
         else
            cTranslate(name, _output);
         }
//?   else if (strcmp(cmmd, Pchar("VARS"))      == 0)
//?      cVars(arg);
      else if (CConsole::TheConsole) {
         _body->StrItem(line, sizeString);
         i = StrPos(Pchar(ftCmdStr), line);
         strcpy(line+i, line+i+2);

//?      PCStrList list = resolveString(line, true);
//?      if (list && list->GotoHead())
//?         list->StrItem(line, sizeString);

         CConsole::TheConsole->WriteString(line);
         }
      }

   return true;
   } // command

//---------------------------------------------------------------------
//
void CXText::Resolve::resolveBody(PCStrList _body, PCStrList _output) {
   if (_body->GotoHead()) do {
      if (lastLoop || nextLoop)
         break;

      Pchar line = _body->StrItem(false);
      if      (condition.IsTrue())
         log(false, line);
      else if (condition.IsFalse() && (StrPos(Pchar("$#el"), line) >= 0 ||
                                       StrPos(Pchar("$#en"), line) >= 0))
         log(false, line);

      // Process the line if it is a command
      if (command(_body, _output))
         if (condition.IsDone()) {
            THROW(Pchar("Condition stack failed"))
            }
         else if (condition.IsMark())
            break;
         else
            continue;

      // Not a command

      if (condition.IsFalse())
         continue;

      // Expand and process the line as specified
      PCStrList list = resolveString(line, false);
      if (list) {
//?      _body->Save(Pchar("bodybefore.txt"));
         long where = _body->Where();
         _body->ListInsert(list, after);
         _body->GotoItem(where);
         _body->Remove();
//?      _body->Save(Pchar("bodyafter.txt"));
         if (where > _body->Number())
            break;
         _body->GotoItem(where-1);
         }
      else
         _output->StrInsert(line, atEnd);

      } while (_body->GotoNext());
  } // resolveBody

//---------------------------------------------------------------------
// The Langoliers are cleaning up
//
class Langolier {
   private:
      Pchar line;

   public:
      Langolier(void):
         line(0) {
         }

      Langolier(Pchar _line):
         line(0) {

         if (strlen(_line) == 0) {
            line = new char[2];
            line[0] = 0x01;
            line[1] = 0;
            }
         else {
            int size = strlen(_line);
            line = new char[size+2];
            strcpy(line, _line);
            if (line[size-1] >= ' ') {
               ChrCat(line, 0x01, size+2);
               }

            Pchar next = line;
            while (*next) {
               if (*next < ' ') {
                  *next++ = 0x01;
                  if (*next) {
                     if (*next < ' ')
                        strcpy(next, next+1);
                     else
                        next++;
                     }
                  }
               else
                  next++;
               }
            }
         }

      virtual ~Langolier() {
         delete[] line;
         }

      Pchar nextLine(void) {
         if (line == 0)
            return 0;

         // Scan for the end of the line segment
         Pchar next;
         for (next = line; *next; next++)
            if (*next == 0x01) {
               *next = 0;
               Pchar temp = 0;
               STRINIT(temp, line);
               strcpy(line, next+1);

               return temp;
               }

         return 0;
         } // nextLine
         
   }; // Langolier

//---------------------------------------------------------------------
// Pull out the body text from the DOM, and then parse it.
//
// Traverse the body text a line at a time. If the line is a
// command line, then process the command. If not, replace strings
// on the line, and copy the line to the output. Note that replacing
// strings may create more lines. This function can be recursive,
// depending on the script.
//
void CXText::Resolve::resolve(PCStrList _body, PCStrList _output) {
   ulong sleep = CArgs::TheArgs->IArg();
   ulong islp  = sleep;

   // Sleep some on request
   if (sleep && --islp <= 0) {
      ::Sleep(1);
      islp = sleep;
      }

/*?
   // Convert Pchar("composited") lines into simple text lines. This is needed
   // because the lines in _body were probably obtained using an XPath
   // expression ending with ../text(), which gets all the text lines
   // within the element.
   PCStrList body = new CStrList();
   if (_body->GotoHead()) do {
      Pchar p = _body->StrItem(false);

      Langolier* lang = new Langolier(p);
      Pchar line = lang->nextLine();
      while (line) {
         body->StrInsert(line, atEnd);
         line = lang->nextLine();
         }
      delete lang;
      } while (_body->GotoNext());
?*/

   // Now, we can go through the lines in the body, one by one.
   resolveBody(_body, _output);
   } // resolve

//---------------------------------------------------------------------
//
void CXText::Resolve::translate(Pchar _line, Pchar _map) {
   for (Pchar p = _line; *p; p++)
      *p = _map[*p];
   } // translate

//---------------------------------------------------------------------
// Tab marks have the form "$\t{n}. If the ?tab$ variable exists, then
// the {n} indexes into that table the variable is assumed to contain.
// If the variable does not exist, the {n} is taken as the actual column
// number. If the {n} index is omitted, then a standard tab (0x09)
// character is the value.
//
void CXText::Resolve::expandTab(int _tabOfst, Pchar _line, int _size) {
   if (_tabOfst >= strlen(_line))
      THROW(Pchar("Tab offset beyond the line to expand"))

   // Copy the part of the line left of the tab mark to a buffer
   char left[sizeString];
   _line[_tabOfst] = 0;
   Strcpy(left, _line, sizeString);

   // Discard Pchar("$\t") to access the tab table index
   strcpy(_line, _line+_tabOfst+3);

   // If no index, then replace with a tab character
   if (_line[0] !='{') {
      ChrCat(left, '\t', sizeString);
      Strcat(left, _line, sizeString);
      Strcpy(_line, left, _size);
      return;
      }

   // Discard the tab index opener character
   strcpy(_line, _line+1);

   // Locate the tab index closure character
   int i = ChrPos('}', _line);
   if (i < 0 || i > strlen(_line))
      THROW(Pchar("Tab mark badly formatted (form is $\\t{n})"))

   // Capture the tab table index
   _line[i] = 0;
   int tndx = StrToShort(_line);
   if (tndx < 1)
      THROW(Pchar("Tab index is out of range"))

   // Discard the remains of the tab mark
   strcpy(_line, _line+i+1);

   // Get the tab offset from the tab table (if available)
   char ttmp[sizeString];
   toString(Pchar("tabs"), ttmp, sizeString);
   if (!StrEmpty(ttmp)) {
      PCRecord tabs = new CRecord(ttmp);
      if (tndx > tabs->Number())
         THROW(Pchar("Tab index is out of range"))

      tndx = tabs->GetIntField(tndx);
      }

   if (tndx > sizeString)
      THROW(Pchar("Tab offset is to big for implementation"))

   // Pad the string to the tab offset
   PadStr(left, tndx-1, sizeString);

   // Append the rest of the line to the left part
   Strcat(left, _line, sizeString);

   Strcpy(_line, left, _size);
   } // expandTab

//---------------------------------------------------------------------
//
bool CXText::Resolve::setMap(Pchar _translate, Pchar _map) {
   PCRecord rec = new CRecord(_translate);
   char arg[sizeString];
   char fun[sizeString];
   rec->GetField(2, arg, sizeString);
   rec->GetField(3, fun, sizeString);

   if (StrEmpty(arg))
      return false;

   unsigned int i;
   for (i = 0; i < 256; i++)
      _map[i] = char(i);
   for (i = 0; i < strlen(arg); i++)
      _map[arg[i]] = fun[i];

   return true;
   } // setMap

//---------------------------------------------------------------------
// Copy the output list to the return list, cut and joined by the
// embedded line control commands, and translated by the embedded
// ,translate lines.
//
bool CXText::Resolve::getOutput(PCStrList _output, PCStrList _return) {
   bool toMap = false;
   char map[256];

   char mssg[sizeString];

//?if (CArgs::TheArgs->FArg(ftDebugFlag))
//?   _output->Save(Pchar("output.dbg"));

   pCFile out = 0;

   char outLine[sizeString];
   outLine[0] = 0;

   if (_output->GotoHead()) do {
      char inpLine[sizeString];
      _output->StrItem(inpLine, sizeString);

      if (StrPos(Pchar(",translate,"), inpLine) == 0) {
         toMap = setMap(inpLine, map);
         continue;
         }

      else {
         Strcat(outLine, inpLine, sizeString);

         bool toJoin = false;
         int lc = StrPos(Pchar(ftLineCmd), outLine);
         while (lc >= 0) {
            if      (outLine[lc+2] == 'c' ||
                     outLine[lc+2] == 'C') {
               outLine[lc] = 0;
               Strcpy(mssg, outLine, sizeString);
               Strcat(mssg, Pchar("<![CDATA["), sizeString);
               out->WriteSubStr(mssg);
               strcpy(outLine, outLine+lc+3);
               }
            else if (outLine[lc+2] == 'd' ||
                     outLine[lc+2] == 'D') {
               outLine[lc] = 0;
               Strcpy(mssg, outLine, sizeString);
               Strcat(mssg, Pchar("]]>"), sizeString);
               out->WriteSubStr(mssg);
               strcpy(outLine, outLine+lc+3);
               }

            // Process line join mark
            else if (outLine[lc+2] == 'j' ||
                     outLine[lc+2] == 'J') {
               // Setup to append the next line
               toJoin      = true;
               outLine[lc] = 0;
               }

            // Process next line mark
            else if (outLine[lc+2] == 'n' ||
                     outLine[lc+2] == 'N') {
               // Split the line into two lines
               outLine[lc] = 0;
               if (toMap)
                  translate(outLine, map);
               out->WriteString(outLine);
               strcpy(outLine, outLine+lc+3);
               }

            // Process tab mark
            else if (outLine[lc+2] == 't' ||
                     outLine[lc+2] == 'T') {
               expandTab(lc, outLine, sizeString);
               }

            // Check the rest of the output line for another line
            // command
            else {
               lc = StrPos(Pchar(ftLineCmd), outLine+lc+2);
               continue;
               }

            lc = StrPos(Pchar(ftLineCmd), outLine);
            }

         if (!toJoin) {
            if (toMap)
               translate(outLine, map);
            _return->StrInsert(outLine);
            outLine[0] = 0;
            }
         }
      } while (_output->GotoNext());

   if (!StrEmpty(outLine)) {
      if (toMap)
         translate(outLine, map);
      _return->StrInsert(outLine);
      outLine[0] = 0;
      }

   return true;
   } // getOutput

//---------------------------------------------------------------------
//
PCStrList CXText::Resolve::GetOutput(void) {
   PCStrList result = new CStrList();
   getOutput(output, result);

   return result;
   } // GetOutput

//---------------------------------------------------------------------
//
bool CXText::Resolve::dumpSubStr(pCFile _out,
                                 Pchar  _str) {
   if (!_out)
      fwrite(_str, 1, strlen(_str), stdout);
   else
      _out->WriteSubStr(_str);

   return true;
   } // dumpSubStr

//---------------------------------------------------------------------
//
bool CXText::Resolve::dumpString(pCFile _out,
                                 Pchar  _str,
                                 int    _nr,
                                 bool   _blks) {
   bool blks = _blks;
   Trim(_str, backEnd);

   // Blank line output management
   if (_nr >= 0) {
      if      (StrEmpty(_str))
         return true;
      else if (blks) {
         blks = false;
         for (int i = 0; i < _nr; i++)
            if (!_out)
               fwrite(Pchar("\n"), 1, 1, stdout);
            else
               _out->WriteString(Pchar(""));
         }
      }

   // Output non-blank line
   if (!_out) {
      fwrite(_str, 1, strlen(_str), stdout);
      fwrite(Pchar("\n"), 1, 1, stdout);
      }
   else
      _out->WriteString(_str);

   return blks;
   } // dumpString

//---------------------------------------------------------------------
// Write the output list to the file(s), cut and joined by the embedded
// line control commands, and translated by the embedded ,translate
// lines.
//
bool CXText::Resolve::dumpOutput(PCStrList _output,
                                 Pchar     _fileName,
                                 bool      _append) {
   char fileName[sizeString];
   Strcpy(fileName, _fileName, sizeString);
   if (StrEmpty(fileName))
      strcpy(fileName, Pchar("stdout"));

   bool toMap = false;
   char map[256];

   char mssg[sizeString];

//?if (CArgs::TheArgs->FArg(ftDebugFlag))
//?   _output->Save(Pchar("output.dbg"));

   pCFile out = 0;

   // If a file name is specified and the file name is not Pchar("stdout"),
   // then output to the specified disk file. Otherwise, output to
   // stdout.
   if (stricmp(fileName, Pchar("stdout")) != 0) {
      out = new CFile(fileName);
      if (_append) {
         if (!out->OpenAppend()) {
            sprintf(message, Pchar("Could not open %s for append."),
                             out->GetName(false));
            throw(message);
            }
         }
      else {
         varHandler->Variable(fileName, -1);
         if (!out->OpenOutput()) {
            sprintf(message, Pchar("Could not open %s for output."),
                             out->GetName(false));
            throw(message);
            }
         }
      }

   // Initialize for blank line output control
   int nr;
   PCVar var = varHandler->Variable(fileName);
   if (var) {
      toString(var, mssg, sizeString);
      nr = StrToShort(mssg);
      }
   else
      nr = -1;
   
   char outLine[sizeString];
   outLine[0] = 0;

   if (_output->GotoHead()) do {
      char inpLine[sizeString];
      _output->StrItem(inpLine, sizeString);

      if (StrPos(Pchar(",translate,"), inpLine) == 0) {
         toMap = setMap(inpLine, map);
         continue;
         }

      else {
         Strcat(outLine, inpLine, sizeString);

         bool toJoin = false;
         int lc = StrPos(Pchar(ftLineCmd), outLine);
         while (lc >= 0) {
            if      (outLine[lc+2] == 'b' ||
                     outLine[lc+2] == 'B') {    // controls blank lines
               nr = -1;
               char n = outLine[lc+3];
               if (n >= '0' && n <= '9')
                  nr = n - '0';
               var->Value(nr);
               outLine[lc] = 0;
               break;
               }
            else if (outLine[lc+2] == 'c' ||
                     outLine[lc+2] == 'C') {     // outputs Pchar("<![CDATA[")
               outLine[lc] = 0;
               Strcpy(mssg, outLine, sizeString);
               Strcat(mssg, Pchar("<![CDATA["), sizeString);
               dumpSubStr(out, mssg);
               strcpy(outLine, outLine+lc+3);
               }
            else if (outLine[lc+2] == 'd' ||
                     outLine[lc+2] == 'D') {           // outputs Pchar("]]>")
               outLine[lc] = 0;
               Strcpy(mssg, outLine, sizeString);
               Strcat(mssg, Pchar("]]>"), sizeString);
               dumpSubStr(out, mssg);
               strcpy(outLine, outLine+lc+3);
               }

            // Process line join mark
            else if (outLine[lc+2] == 'j' ||
                     outLine[lc+2] == 'J') {    // joins with next line
               // Setup to append the next line
               toJoin      = true;
               outLine[lc] = 0;
               }

            // Process next line mark
            else if (outLine[lc+2] == 'n' ||
                     outLine[lc+2] == 'N') {    // outputs a line break
               // Split the line into two lines
               outLine[lc] = 0;
               if (toMap)
                  translate(outLine, map);
               didBlanks = dumpString(out, outLine, nr, didBlanks);
               strcpy(outLine, outLine+lc+3);
               }

            // Process tab mark
            else if (outLine[lc+2] == 't' ||
                     outLine[lc+2] == 'T') {    // tabs the output line
               expandTab(lc, outLine, sizeString);
               }

            // Check the rest of the output line for another line
            // command
            else {
               lc = StrPos(Pchar(ftLineCmd), outLine+lc+2);
               continue;
               }

            lc = StrPos(Pchar(ftLineCmd), outLine);
            }

         if (!toJoin) {
            if (toMap)
               translate(outLine, map);
            didBlanks = dumpString(out, outLine, nr, didBlanks);
            outLine[0] = 0;
            }
         }
      } while (_output->GotoNext());

   if (!StrEmpty(outLine)) {
      if (toMap)
         translate(outLine, map);
      didBlanks = dumpString(out, outLine, nr, didBlanks);
      outLine[0] = 0;
      }

   // Clean up if not standard output
   if (out) {
      out->Close();
      delete out;
      }

   return true;
   } // dumpOutput


